@@ -23,14 +23,24 @@
  */
 
 #include "firebird.h"
+#include "../common/classes/alloc.h"
 #include "../intl/ldcommon.h"
+#include "../jrd/CharSet.h"
+#include "../jrd/IntlUtil.h"
 #include "lc_narrow.h"
 #include "ld_proto.h"
+#include <limits.h>
 #include <math.h>
 
-/*
- * Generic base for InterBase 4.0 Language Driver
- */
+using namespace Firebird;
+
+
+static bool LC_NARROW_family2(texttype* tt, charset* cs, SSHORT country, USHORT flags,
+	const SortOrderTblEntry* noCaseOrderTbl, const BYTE* toUpperConversionTbl,
+	const BYTE* toLowerConversionTbl, const CompressPair* compressTbl, const ExpandChar* expansionTbl,
+	const ASCII* name, USHORT attributes, const UCHAR* specificAttributes, ULONG specificAttributesLength);
+static ULONG fam2_str_to_upper(texttype* obj, ULONG iLen, const BYTE* pStr, ULONG iOutLen, BYTE *pOutStr);
+static ULONG fam2_str_to_lower(texttype* obj, ULONG iLen, const BYTE* pStr, ULONG iOutLen, BYTE *pOutStr);
 
 
 const USHORT LANGFAM2_MAX_KEY	= MAX_KEY;
@@ -86,7 +96,7 @@ const UINT16 NULL_TERTIARY		= 0;
  *	The degenerate case of short strings is handled by a minimal key
  *	length.
  */
-USHORT LC_NARROW_key_length(TEXTTYPE obj, USHORT inLen)
+USHORT LC_NARROW_key_length(texttype* obj, USHORT inLen)
 {
 /* fb_assert (inLen <= LANGFAM2_MAX_KEY); *//* almost certainly an error */
 
@@ -109,7 +119,7 @@ USHORT LC_NARROW_key_length(TEXTTYPE obj, USHORT inLen)
 			bool useSecondary = false;
 			bool useTertiary = false;
 
-			for (USHORT ch = 0; ch <= 255; ++ch)
+			for (int ch = 0; ch <= 255; ++ch)
 			{
 				const SortOrderTblEntry* coll =
 					&((const SortOrderTblEntry*)obj->texttype_impl->texttype_collation_table)[ch];
@@ -156,7 +166,7 @@ static ULONG do_debug = 0;
  * RETURN:
  *		Length, in bytes, of returned key
  */
-USHORT LC_NARROW_string_to_key(TEXTTYPE obj, USHORT iInLen, const BYTE* pInChar, USHORT iOutLen, BYTE *pOutChar,
+USHORT LC_NARROW_string_to_key(texttype* obj, USHORT iInLen, const BYTE* pInChar, USHORT iOutLen, BYTE *pOutChar,
 							   USHORT key_type)
 {
 	fb_assert(pOutChar != NULL);
@@ -206,24 +216,40 @@ USHORT LC_NARROW_string_to_key(TEXTTYPE obj, USHORT iInLen, const BYTE* pInChar,
 		const SortOrderTblEntry* coll =
 			&((const SortOrderTblEntry*) obj->texttype_impl->
 			  texttype_collation_table)[*pInChar];
-		if (!(coll->IsExpand || coll->IsCompress)) {
+
+		if (coll->IsExpand && coll->IsCompress) {
+			/* Both flags set indicate a special value */
+
+			if (obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first)
+			{
+				if (coll->Primary != NULL_WEIGHT && lprimary < iOutLen)
+					outbuff[lprimary++] = coll->Primary + obj->texttype_impl->ignore_sum;
+				if (coll->Secondary != NULL_SECONDARY && lsecondary < sizeof(secondary))
+					secondary[lsecondary++] = coll->Secondary;
+				if (coll->Tertiary != NULL_TERTIARY && ltertiary < sizeof(tertiary))
+					tertiary[ltertiary++] = coll->Tertiary;
+			}
+			else
+			{
+				if ((coll->Primary != NULL_WEIGHT) &&
+					!(obj->texttype_impl->texttype_flags & TEXTTYPE_ignore_specials) &&
+					 lspecial + 1u < sizeof(special))
+				{
+					special[lspecial++] = (i + 1);	/* position */
+					special[lspecial++] = coll->Primary;
+				}
+			}
+		}
+		else if (!((coll->IsExpand && !(obj->texttype_impl->texttype_flags & TEXTTYPE_disable_expansions)) ||
+				   (coll->IsCompress && !(obj->texttype_impl->texttype_flags & TEXTTYPE_disable_compressions))))
+		{
 			if (coll->Primary != NULL_WEIGHT && lprimary < iOutLen)
-				outbuff[lprimary++] = coll->Primary;
+				outbuff[lprimary++] = coll->Primary + obj->texttype_impl->primary_sum;
 			if (coll->Secondary != NULL_SECONDARY && lsecondary < sizeof(secondary))
 				secondary[lsecondary++] = coll->Secondary;
 			if (coll->Tertiary != NULL_TERTIARY && ltertiary < sizeof(tertiary))
 				tertiary[ltertiary++] = coll->Tertiary;
 		}
-		else if (coll->IsExpand && coll->IsCompress) {
-			/* Both flags set indicate a special value */
-			if ((coll->Primary != NULL_WEIGHT) &&
-				!(obj->texttype_impl->texttype_flags & TEXTTYPE_ignore_specials) &&
-				 lspecial + 1 < sizeof(special))
-			{
-				special[lspecial++] = (i + 1);	/* position */
-				special[lspecial++] = coll->Primary;
-			}
-		}
 		else if (coll->IsExpand) {
 			const ExpandChar* exp = &((const ExpandChar*) obj->texttype_impl->texttype_expand_table)[0];
 			while (exp->Ch && exp->Ch != *pInChar)
@@ -235,7 +261,7 @@ USHORT LC_NARROW_string_to_key(TEXTTYPE obj, USHORT iInLen, const BYTE* pInChar,
 						&((const SortOrderTblEntry*) obj->texttype_impl->
 						  texttype_collation_table)[exp->ExpCh2];
 				if (coll->Primary != NULL_WEIGHT && lprimary < iOutLen)
-					outbuff[lprimary++] = coll->Primary;
+					outbuff[lprimary++] = coll->Primary + obj->texttype_impl->primary_sum;
 				if (coll->Secondary != NULL_SECONDARY && lsecondary < sizeof(secondary))
 					secondary[lsecondary++] = coll->Secondary;
 				if (coll->Tertiary != NULL_TERTIARY && ltertiary < sizeof(tertiary))
@@ -269,7 +295,7 @@ USHORT LC_NARROW_string_to_key(TEXTTYPE obj, USHORT iInLen, const BYTE* pInChar,
 			if (key_type != INTL_KEY_PARTIAL || complete)
 			{
 				if (coll->Primary != NULL_WEIGHT && lprimary < iOutLen)
-					outbuff[lprimary++] = coll->Primary;
+					outbuff[lprimary++] = coll->Primary + obj->texttype_impl->primary_sum;
 				if (coll->Secondary != NULL_SECONDARY && lsecondary < sizeof(secondary))
 					secondary[lsecondary++] = coll->Secondary;
 				if (coll->Tertiary != NULL_TERTIARY && ltertiary < sizeof(tertiary))
@@ -373,7 +399,7 @@ struct coltab_status {
 
 
 
-static SSHORT special_scan(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, const BYTE* s2)
+static SSHORT special_scan(texttype* obj, ULONG l1, const BYTE* s1, ULONG l2, const BYTE* s2)
 {
 	const SortOrderTblEntry* col1 = 0;
 	const SortOrderTblEntry* col2 = 0;
@@ -387,8 +413,13 @@ static SSHORT special_scan(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, con
 			col1 =
 				&((const SortOrderTblEntry*) obj->texttype_impl->
 				  texttype_collation_table)[*s1];
-			if (col1->IsExpand && col1->IsCompress)
+
+			if (col1->IsExpand && col1->IsCompress &&
+				!(obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first))
+			{
 				break;
+			}
+
 			l1--;
 			s1++;
 			index1++;
@@ -399,8 +430,12 @@ static SSHORT special_scan(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, con
 			col2 =
 				&((const SortOrderTblEntry*) obj->texttype_impl->
 				  texttype_collation_table)[*s2];
-			if (col2->IsExpand && col2->IsCompress)
+			if (col2->IsExpand && col2->IsCompress &&
+				!(obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first))
+			{
 				break;
+			}
+
 			l2--;
 			s2++;
 			index2++;
@@ -427,9 +462,11 @@ static SSHORT special_scan(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, con
 }
 
 
-static const SortOrderTblEntry* get_coltab_entry(TEXTTYPE obj, const UCHAR** p,
-	ULONG* l, coltab_status* stat)
+static const SortOrderTblEntry* get_coltab_entry(texttype* obj, const UCHAR** p,
+	ULONG* l, coltab_status* stat, int* sum)
 {
+	*sum = obj->texttype_impl->primary_sum;
+
 	if (stat->stat_flags & LC_HAVE_WAITING) {
 		(*l)--;
 		(*p)++;
@@ -443,20 +480,36 @@ static const SortOrderTblEntry* get_coltab_entry(TEXTTYPE obj, const UCHAR** p,
 		const SortOrderTblEntry* col =
 			&((const SortOrderTblEntry*) obj->texttype_impl->
 			  texttype_collation_table)[**p];
-		if (!(col->IsExpand || col->IsCompress)) {
+
+		if (col->IsExpand && col->IsCompress)
+		{
+			if (obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first)
+			{
+				*sum = obj->texttype_impl->ignore_sum;
+
+				/* Have col */
+				(*l)--;
+				(*p)++;
+				return col;
+			}
+			else
+			{
+				/* Both flags set indicate a special value */
+				/* Need a new col */
+				(*l)--;
+				(*p)++;
+				stat->stat_flags |= LC_HAVE_SPECIAL;
+				continue;
+			}
+		}
+		else if (!((col->IsExpand && !(obj->texttype_impl->texttype_flags & TEXTTYPE_disable_expansions)) ||
+				   (col->IsCompress && !(obj->texttype_impl->texttype_flags & TEXTTYPE_disable_compressions))))
+		{
 			/* Have col */
 			(*l)--;
 			(*p)++;
 			return col;
 		}
-		else if (col->IsExpand && col->IsCompress) {
-			/* Both flags set indicate a special value */
-			/* Need a new col */
-			(*l)--;
-			(*p)++;
-			stat->stat_flags |= LC_HAVE_SPECIAL;
-			continue;
-		}
 		else if (col->IsExpand) {
 			const ExpandChar* exp = &((const ExpandChar*) obj->texttype_impl->texttype_expand_table)[0];
 			while (exp->Ch && exp->Ch != **p)
@@ -502,7 +555,7 @@ static const SortOrderTblEntry* get_coltab_entry(TEXTTYPE obj, const UCHAR** p,
 
 #define XOR	^					/* C bitwise XOR operator - defined for readability */
 
-SSHORT LC_NARROW_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, const BYTE* s2, 
+SSHORT LC_NARROW_compare(texttype* obj, ULONG l1, const BYTE* s1, ULONG l2, const BYTE* s2, 
 	INTL_BOOL* error_flag)
 {
 	fb_assert(obj != NULL);
@@ -542,12 +595,15 @@ SSHORT LC_NARROW_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, const
 	const SortOrderTblEntry* col2 = 0;
 
 	while (true) {
-		col1 = get_coltab_entry(obj, &s1, &l1, &stat1);
-		col2 = get_coltab_entry(obj, &s2, &l2, &stat2);
+		int sum1, sum2;
+
+		col1 = get_coltab_entry(obj, &s1, &l1, &stat1, &sum1);
+		col2 = get_coltab_entry(obj, &s2, &l2, &stat2, &sum2);
+
 		if (!col1 || !col2)
 			break;
-		if (col1->Primary != col2->Primary)
-			return (col1->Primary - col2->Primary);
+		if (col1->Primary + sum1 != col2->Primary + sum2)
+			return ((col1->Primary + sum1) - (col2->Primary + sum2));
 		if ((obj->texttype_impl->texttype_flags & TEXTTYPE_secondary_insensitive) == 0 &&
 			col1->Secondary != col2->Secondary)
 		{
@@ -593,7 +649,8 @@ SSHORT LC_NARROW_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, const
 		if (
 			((stat1.stat_flags & LC_HAVE_SPECIAL)
 			 || (stat2.stat_flags & LC_HAVE_SPECIAL))
-			&& !(obj->texttype_impl->texttype_flags & TEXTTYPE_ignore_specials))
+			&& !(obj->texttype_impl->texttype_flags & TEXTTYPE_ignore_specials)
+			&& !(obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first))
 		{
 			return special_scan(obj, save_l1, save_s1, save_l2, save_s2);
 		}
@@ -614,7 +671,7 @@ SSHORT LC_NARROW_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1, ULONG l2, const
  * Routine used for comparing results from comparison algorithm
  * to results from key creation algorithm
  */
-static SSHORT old_fam2_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1,
+static SSHORT old_fam2_compare(texttype* obj, ULONG l1, const BYTE* s1,
 	ULONG l2, const BYTE* s2, INTL_BOOL* error_flag)
 {
 	BYTE key1[LANGFAM2_MAX_KEY];
@@ -653,7 +710,7 @@ static SSHORT old_fam2_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1,
  * Routine used for comparing results from comparision algorithm
  * to results from key creation algorithm
  */
-static SSHORT fam2_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1,
+static SSHORT fam2_compare(texttype* obj, ULONG l1, const BYTE* s1,
 	ULONG l2, const BYTE* s2, INTL_BOOL* error_flag)
 {
 	SSHORT res1 = old_fam2_compare(obj, l1, s1, l2, s2);
@@ -673,7 +730,7 @@ static SSHORT fam2_compare(TEXTTYPE obj, ULONG l1, const BYTE* s1,
 
 #endif	/* DEBUG_COMPARE */
 
-ULONG LC_NARROW_canonical(TEXTTYPE obj, ULONG srcLen, const UCHAR* src, ULONG dstLen, UCHAR* dst)
+ULONG LC_NARROW_canonical(texttype* obj, ULONG srcLen, const UCHAR* src, ULONG dstLen, UCHAR* dst)
 {
 	fb_assert(dst != NULL);
 	fb_assert(src != NULL);
@@ -688,10 +745,13 @@ ULONG LC_NARROW_canonical(TEXTTYPE obj, ULONG srcLen, const UCHAR* src, ULONG ds
 
 		USHORT primary = coll->Primary;
 
-		if (coll->IsExpand && coll->IsCompress)
-			primary += obj->texttype_impl->ignore_sum;
-		else
-			primary += obj->texttype_impl->primary_sum;
+		if (obj->texttype_impl->texttype_flags & TEXTTYPE_specials_first)
+		{
+			if (coll->IsExpand && coll->IsCompress)
+				primary += obj->texttype_impl->ignore_sum;
+			else
+				primary += obj->texttype_impl->primary_sum;
+		}
 
 		if ((obj->texttype_impl->texttype_flags & (TEXTTYPE_secondary_insensitive | TEXTTYPE_tertiary_insensitive)) == 0)
 		{
@@ -713,8 +773,289 @@ ULONG LC_NARROW_canonical(TEXTTYPE obj, ULONG srcLen, const UCHAR* src, ULONG ds
 }
 
 
-void LC_NARROW_destroy(TEXTTYPE obj)
+void LC_NARROW_destroy(texttype* obj)
 {
 	delete obj->texttype_impl;
 }
 
+
+
+bool LC_NARROW_family2(
+	texttype* tt,
+	charset* cs,
+	SSHORT country,
+	USHORT flags,
+	const SortOrderTblEntry* noCaseOrderTbl,
+	const BYTE* toUpperConversionTbl,
+	const BYTE* toLowerConversionTbl,
+	const CompressPair* compressTbl,
+	const ExpandChar* expansionTbl,
+	const ASCII* name,
+	USHORT attributes,
+	const UCHAR* specificAttributes,
+	ULONG specificAttributesLength)
+{
+	if (attributes & ~TEXTTYPE_ATTR_PAD_SPACE)
+		return false;
+
+	tt->texttype_version			= TEXTTYPE_VERSION_1;
+	tt->texttype_name				= name;
+	tt->texttype_country			= country;
+	tt->texttype_pad_option			= (attributes & TEXTTYPE_ATTR_PAD_SPACE) ? true : false;
+	tt->texttype_fn_key_length		= LC_NARROW_key_length;
+	tt->texttype_fn_string_to_key	= LC_NARROW_string_to_key;
+	tt->texttype_fn_compare			= LC_NARROW_compare;
+	tt->texttype_fn_str_to_upper	= fam2_str_to_upper;
+	tt->texttype_fn_str_to_lower	= fam2_str_to_lower;
+	tt->texttype_fn_destroy			= LC_NARROW_destroy;
+	tt->texttype_impl				= FB_NEW(*getDefaultMemoryPool()) TextTypeImpl;
+	tt->texttype_impl->texttype_collation_table	= (const BYTE*) noCaseOrderTbl;
+	tt->texttype_impl->texttype_toupper_table	= toUpperConversionTbl;
+	tt->texttype_impl->texttype_tolower_table	= toLowerConversionTbl;
+	tt->texttype_impl->texttype_compress_table	= (const BYTE*) compressTbl;
+	tt->texttype_impl->texttype_expand_table	= (const BYTE*) expansionTbl;
+	tt->texttype_impl->texttype_flags			= ((flags) & REVERSE) ? TEXTTYPE_reverse_secondary : 0;
+	tt->texttype_impl->texttype_bytes_per_key	= 0;
+
+	IntlUtil::SpecificAttributesMap map;
+	Jrd::CharSet* charSet = NULL;
+
+	try
+	{
+		charSet = Jrd::CharSet::createInstance(*getDefaultMemoryPool(), 0, cs);
+
+		if (!IntlUtil::parseSpecificAttributes(charSet, specificAttributesLength, specificAttributes, &map))
+		{
+			delete charSet;
+			return false;
+		}
+
+		delete charSet;
+	}
+	catch (...)
+	{
+		delete charSet;
+		return false;
+	}
+
+	int validAttributeCount = 0;
+	string value;
+
+	if (map.get("DISABLE-COMPRESSIONS", value) && (value == "0" || value == "1"))
+	{
+		++validAttributeCount;
+		if (value == "1")
+			tt->texttype_impl->texttype_flags |= TEXTTYPE_disable_compressions;
+	}
+
+	if (map.get("DISABLE-EXPANSIONS", value) && (value == "0" || value == "1"))
+	{
+		++validAttributeCount;
+		if (value == "1")
+			tt->texttype_impl->texttype_flags |= TEXTTYPE_disable_expansions;
+	}
+
+	if (map.get("SPECIALS-FIRST", value) && (value == "0" || value == "1"))
+	{
+		int maxPrimary = 0;
+		int minPrimary = INT_MAX;
+		int maxIgnore = 0;
+
+		if (!(tt->texttype_impl->texttype_flags & TEXTTYPE_disable_compressions))
+		{
+			while (compressTbl->CharPair[0])
+			{
+				if (compressTbl->NoCaseWeight.Primary > maxPrimary)
+					maxPrimary = compressTbl->NoCaseWeight.Primary;
+
+				if (compressTbl->NoCaseWeight.Primary < minPrimary)
+					minPrimary = compressTbl->NoCaseWeight.Primary;
+
+				++compressTbl;
+			}
+		}
+
+		for (int ch = 0; ch <= 255; ++ch)
+		{
+			const SortOrderTblEntry* coll =
+				&((const SortOrderTblEntry*)tt->texttype_impl->texttype_collation_table)[ch];
+
+			if (coll->IsExpand && coll->IsCompress)
+			{
+				if (coll->Primary > maxIgnore)
+					maxIgnore = coll->Primary;
+			}
+			else if (!((coll->IsExpand && !(tt->texttype_impl->texttype_flags & TEXTTYPE_disable_expansions)) ||
+					   (coll->IsCompress && !(tt->texttype_impl->texttype_flags & TEXTTYPE_disable_compressions))))
+			{
+				if (coll->Primary > maxPrimary)
+					maxPrimary = coll->Primary;
+
+				if (coll->Primary < minPrimary)
+					minPrimary = coll->Primary;
+			}
+		}
+
+		if (maxIgnore > 0 && maxPrimary + maxIgnore - 1 <= 255)
+		{
+			++validAttributeCount;
+
+			if (value == "1")
+			{
+				tt->texttype_impl->texttype_flags |= TEXTTYPE_specials_first;
+				tt->texttype_impl->ignore_sum = minPrimary - 1;
+				tt->texttype_impl->primary_sum = maxIgnore - 1;
+			}
+		}
+	}
+
+	if (map.count() - validAttributeCount != 0)
+		return false;
+
+	return true;
+}
+
+
+bool LC_NARROW_family3(
+	texttype* tt,
+	charset* cs,
+	SSHORT country,
+	USHORT flags,
+	const SortOrderTblEntry* noCaseOrderTbl,
+	const BYTE* toUpperConversionTbl,
+	const BYTE* toLowerConversionTbl,
+	const CompressPair* compressTbl,
+	const ExpandChar* expansionTbl,
+	const ASCII* name,
+	USHORT attributes,
+	const UCHAR* specificAttributes,
+	ULONG specificAttributesLength)
+{
+	bool multiLevel = false;
+
+	IntlUtil::SpecificAttributesMap map;
+	Jrd::CharSet* charSet = NULL;
+	string newSpecificAttributes;
+
+	try
+	{
+		charSet = Jrd::CharSet::createInstance(*getDefaultMemoryPool(), 0, cs);
+
+		if (!IntlUtil::parseSpecificAttributes(charSet, specificAttributesLength, specificAttributes, &map))
+		{
+			delete charSet;
+			return false;
+		}
+
+		string value;
+		if (map.get("MULTI-LEVEL", value))
+		{
+			if (value == "0")
+				multiLevel = false;
+			else if (value == "1")
+				multiLevel = true;
+			else
+				return false;
+
+			map.remove("MULTI-LEVEL");
+		}
+
+		newSpecificAttributes = IntlUtil::generateSpecificAttributes(charSet, map);
+		specificAttributes = (const UCHAR*)newSpecificAttributes.begin();
+		specificAttributesLength = newSpecificAttributes.length();
+
+		delete charSet;
+	}
+	catch (...)
+	{
+		delete charSet;
+		return false;
+	}
+
+	if (LC_NARROW_family2(tt, cs, country, flags, noCaseOrderTbl,
+				toUpperConversionTbl, toLowerConversionTbl, compressTbl, expansionTbl, name,
+				attributes & ~(TEXTTYPE_ATTR_CASE_INSENSITIVE | TEXTTYPE_ATTR_ACCENT_INSENSITIVE),
+				specificAttributes, specificAttributesLength))
+	{
+		if (!multiLevel)
+			tt->texttype_impl->texttype_flags |= TEXTTYPE_non_multi_level;
+
+		if (attributes & (TEXTTYPE_ATTR_CASE_INSENSITIVE | TEXTTYPE_ATTR_ACCENT_INSENSITIVE))
+		{
+			tt->texttype_impl->texttype_flags |= TEXTTYPE_ignore_specials;
+
+			if (multiLevel)
+			{
+				tt->texttype_flags |= TEXTTYPE_SEPARATE_UNIQUE;
+
+				if ((attributes & (TEXTTYPE_ATTR_CASE_INSENSITIVE | TEXTTYPE_ATTR_ACCENT_INSENSITIVE)) ==
+					TEXTTYPE_ATTR_ACCENT_INSENSITIVE)
+				{
+					tt->texttype_flags |= TEXTTYPE_UNSORTED_UNIQUE;
+				}
+			}
+
+			if ((attributes & (TEXTTYPE_ATTR_CASE_INSENSITIVE | TEXTTYPE_ATTR_ACCENT_INSENSITIVE)) ==
+				(TEXTTYPE_ATTR_CASE_INSENSITIVE | TEXTTYPE_ATTR_ACCENT_INSENSITIVE))
+			{
+				tt->texttype_canonical_width = 1;
+			}
+			else
+				tt->texttype_canonical_width = 2;
+
+			tt->texttype_fn_canonical = LC_NARROW_canonical;
+
+			if (attributes & TEXTTYPE_ATTR_ACCENT_INSENSITIVE)
+				tt->texttype_impl->texttype_flags |= TEXTTYPE_secondary_insensitive;
+
+			if (attributes & TEXTTYPE_ATTR_CASE_INSENSITIVE)
+				tt->texttype_impl->texttype_flags |= TEXTTYPE_tertiary_insensitive;
+		}
+
+		return true;
+	}
+	else
+		return false;
+}
+
+
+/*
+ *	Returns INTL_BAD_STR_LENGTH if output buffer was too small
+ */
+static ULONG fam2_str_to_upper(texttype* obj, ULONG iLen, const BYTE* pStr, ULONG iOutLen, BYTE *pOutStr)
+{
+	fb_assert(pStr != NULL);
+	fb_assert(pOutStr != NULL);
+	fb_assert(iOutLen >= iLen);
+	const BYTE* const p = pOutStr;
+	while (iLen && iOutLen) {
+		*pOutStr++ = (obj->texttype_impl->texttype_toupper_table[(unsigned) *pStr]);
+		pStr++;
+		iLen--;
+		iOutLen--;
+	}
+	if (iLen != 0)
+		return (INTL_BAD_STR_LENGTH);
+	return (pOutStr - p);
+}
+
+
+/*
+ *	Returns INTL_BAD_STR_LENGTH if output buffer was too small
+ */
+static ULONG fam2_str_to_lower(texttype* obj, ULONG iLen, const BYTE* pStr, ULONG iOutLen, BYTE *pOutStr)
+{
+	fb_assert(pStr != NULL);
+	fb_assert(pOutStr != NULL);
+	fb_assert(iOutLen >= iLen);
+	const BYTE* const p = pOutStr;
+	while (iLen && iOutLen) {
+		*pOutStr++ = (obj->texttype_impl->texttype_tolower_table[(unsigned) *pStr]);
+		pStr++;
+		iLen--;
+		iOutLen--;
+	}
+	if (iLen != 0)
+		return (INTL_BAD_STR_LENGTH);
+	return (pOutStr - p);
+}
