@@ -19,8 +19,6 @@
  *
  * All Rights Reserved.
  * Contributor(s): ______________________________________.
- * Added TCP_NO_DELAY option for superserver on Linux
- * FSG 16.03.2001
  *
  * 2001.07.06 Sean Leyne - Code Cleanup, removed "#ifdef READONLY_DATABASE"
  *                         conditionals, as the engine now fully supports
@@ -84,10 +82,10 @@
 # endif
 #endif
 
-#define CANCEL_OPERATION
 
 #ifdef SUPERSERVER
 #define GOVERNOR
+#define CANCEL_OPERATION
 #define FB_ARCHITECTURE isc_info_db_class_server_access
 #else
 #define FB_ARCHITECTURE isc_info_db_class_classic_access
@@ -95,7 +93,7 @@
 
 
 /*****************************************************
-* Linux for Intel platforms 
+* Linux platforms 
 *****************************************************/
 #ifdef LINUX
 #define QUADFORMAT "ll"
@@ -113,17 +111,13 @@
 //format for __LINE__
 #define LINEFORMAT "d"
 
-#ifdef SUPERSERVER
-#define SET_TCP_NO_DELAY
-#endif
-
 //#define KILLER_SIGNALS
 
 #define UNIX
 #define IEEE
 
 #ifdef AMD64
-#define IMPLEMENTATION  isc_info_db_impl_linux_amd64 /* 66  next higher unique number, See you later  */
+#define IMPLEMENTATION  isc_info_db_impl_linux_amd64 /* 66 */
 #endif
 
 #ifdef PPC
@@ -132,7 +126,7 @@
 
 #ifdef i386
 #define I386
-#define IMPLEMENTATION  isc_info_db_impl_i386 /* 60  next higher unique number, See you later  */
+#define IMPLEMENTATION  isc_info_db_impl_i386 /* 60 */
 #endif /* i386 */
 
 #ifdef sparc
@@ -140,6 +134,14 @@
 #define RISC_ALIGNMENT
 #endif /* sparc */
 
+#ifdef MIPSEL
+#define IMPLEMENTATION  isc_info_db_impl_linux_mipsel /* 71  */
+#endif /* mipsel */
+
+#ifdef MIPS
+#define IMPLEMENTATION  isc_info_db_impl_linux_mips /* 72  */
+#endif /* mips */
+
 #define MEMMOVE(from, to, length)		memmove ((void *)to, (void *)from, (size_t) length)
 #define MOVE_FAST(from, to, length)       memcpy (to, from, (int) (length))
 #define MOVE_FASTER(from, to, length)     memcpy (to, from, (int) (length))
@@ -169,12 +171,16 @@
 //#define DOUBLE_ALIGN    4
 //#define BSD_UNIX
 #define UNIX
+#ifdef powerpc 
+#define IMPLEMENTATION isc_info_db_impl_darwin_ppc /* 63 */
+#endif
 #ifdef i386
 #define I386
 #define IMPLEMENTATION isc_info_db_impl_darwin_x86 /* 70 */
 #endif
-#ifdef __ppc__
-#define IMPLEMENTATION isc_info_db_impl_darwin_ppc /* 63 */
+#ifdef __x86_64__
+#define DARWIN64
+#define IMPLEMENTATION isc_info_db_impl_darwin_x64 /* 73 */
 #endif
 #define IEEE
 #define QUADCONST(n) (n##LL)
@@ -276,40 +282,22 @@
  * in Solaris
  */
 #define SOLARIS_MT
-
-#ifdef SOLARIS_MT
-#define ANY_THREADING
 #define MULTI_THREAD
-#endif
 
 /*  Define the following only on platforms whose standard I/O
  *  implementation is so weak that we wouldn't be able to fopen
  *  a file whose underlying file descriptor would be > 255.
+ *  Hey, we're not running on PDP-11's any more: would it kill you
+ *  to use a short instead of a char to hold the fileno?  :-(
  */
  
-/* Firebird currently uses sfio downloadable from:
-   http://www.research.att.com/~gsf/cgi-bin/download.cgi?action=list&name=sfio
-   For Firebird to build correctly and use sfio properly, you need all of
-   sfio from the above download. However there are problems with it.
-   The Firebird source tree contains the corrected files where applicable.
-   1. The stdio.h created by sfio is not compatible with Firebird, so
-   you must use the amended one provided. After building sfio, copy our
-   version of stdio.h into include 
-   2. Sfio will not compile so some minor modifications need to be made
-   to the sfio (Stdio_b directory) Makefile to pull in the missing symbols 
-   for sfset.c and sfclrlock.c (SRCS and OBJS).
-   3. Finally the modified sfstrtof.h and sfvprintf.c need to be placed
-   in the src/lib/sfio directory for the build of sfio to work.
-
-   Now build SuperServer using --enable-superserver --with-sfio=yourSfioDirectory
-
-   Hopefully we will not need to use sfio for much longer as there is a fix
-   for this problem now in some versions of Solaris 10.
-   See enable_extended_FILE_stdio
-
-   Paul Beach 7th July 2008
+/* Why we (solarises) need to rewrite old BSD stdio
+   so many times I suggest stdIO from 
+   http://www.research.att.com/sw/tools/sfio/ 
 */       
-
+/* 	Need to use full sfio not just stdio emulation to fix
+	file descriptor number limit. nmcc Dec2002
+*/
 #if (!defined(SFIO) && defined(SUPERSERVER))
 #error "need to use SFIO"
 #endif
@@ -363,16 +351,16 @@
 
 #define IMPLEMENTATION  isc_info_db_impl_isc_sun4 /* 30 */
 #define RISC_ALIGNMENT
-#endif /* sparc */
+#else /* sparc */
 
 #ifdef i386
 #define I386
 #define IMPLEMENTATION  isc_info_db_impl_isc_sun_386i  /* 32 */
-#endif
+#else /* i386 */
+#define IMPLEMENTATION  isc_info_db_impl_isc_sun_68k /* 28 */
+#endif /* i386 */
 
-#ifdef AMD64
-#define IMPLEMENTATION  isc_info_db_impl_sun_amd64 /* 71 */
-#endif /* AMD64 */
+#endif /* sparc */
 
 #define MOVE_FAST(from, to, length)       memcpy (to, from, (int) (length))
 #define MOVE_FASTER(from, to, length)     memcpy (to, from, (int) (length))
@@ -525,13 +513,15 @@ typedef unsigned __int64 FB_UINT64;
    not permit the LL suffix which some other platforms require, but it
    handles numbers up to the largest 64-bit integer correctly without such
    a suffix, so the macro definition is trivial. */
-#ifdef MINGW // needed for gcc 3.3.1
+#ifdef __GNUC__ // needed for gcc 3.3.1
 #define QUADCONST(n) (n##LL)
 #else
 #define QUADCONST(n) (n)
 #endif
 
-#ifdef _X86_
+#ifdef AMD64
+#define IMPLEMENTATION  isc_info_db_impl_winnt_amd64 /* 68 */
+#else
 #ifndef I386
 #define I386
 #endif
@@ -954,6 +944,16 @@ void GDS_breakpoint(int);
 #define DOUBLE_DIVIDE(a, b)      (((double) (a)) / ((double) (b)))
 #endif
 
+#define STRINGIZE_AUX(x)	#x
+#define STRINGIZE(x)		STRINGIZE_AUX(x)
+
+#ifdef _MSC_VER
+#define CONST64(a) (a##i64)
+#else
+#define CONST64(a) (a##LL)
+#endif
+
+
 /* switch name and state table.  This structure should be used in all
  * command line tools to facilitate parsing options.*/
 struct in_sw_tab_t {
