@@ -126,8 +126,6 @@ double MTH$CVT_D_G(), MTH$CVT_G_D();
 #define SHORT_LIMIT     ((1 << 14) / 5)
 #define LONG_LIMIT      ((1L << 30) / 5)
 
-#define COMMA           ','
-
 /* NOTE: The syntax for the below line may need modification to ensure
  *	 the result of 1 << 62 is a quad
  */
@@ -282,7 +280,6 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
  *
  **************************************/
 	double value;
-	SSHORT scale;
 
 	switch (desc->dsc_dtype) {
 	case dtype_short:
@@ -336,14 +333,17 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
 								&p,
 								(vary*) buffer, sizeof(buffer), err);
 			value = 0.0;
-			scale = 0;
+			int scale = 0;
 			SSHORT sign = 0;
 			bool digit_seen = false, past_sign = false, fraction = false;
 			const char* const end = p + length;
+
+			// skip initial spaces
+			for (; p < end && *p == ' '; p++)
+				;
+
 			for (; p < end; p++) {
-				if (*p == COMMA)
-					continue;
-				else if (DIGIT(*p)) {
+				if (DIGIT(*p)) {
 					digit_seen = true;
 					past_sign = true;
 					if (fraction)
@@ -367,7 +367,17 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
 				}
 				else if (*p == 'e' || *p == 'E')
 					break;
-				else if (*p != ' ')
+				else if (*p == ' ')
+				{
+					// skip spaces
+					for (; p < end && *p == ' '; p++)
+						;
+
+					// throw if there is something after the spaces
+					if (p < end)
+						conversion_error(desc, err);
+				}
+				else
 					conversion_error(desc, err);
 			}
 
@@ -395,13 +405,23 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
 						   later in this routine. */
 
 						if (exp >= SHORT_LIMIT)
-							(*err) (isc_arith_except, isc_arg_end);
+							(*err) (isc_arith_except, 0);
 					}
 					else if (*p == '-' && !digit_seen && !sign)
 						sign = -1;
 					else if (*p == '+' && !digit_seen && !sign)
 						sign = 1;
-					else if (*p != ' ')
+					else if (*p == ' ')
+					{
+						// skip spaces
+						for (; p < end && *p == ' '; p++)
+							;
+
+						// throw if there is something after the spaces
+						if (p < end)
+							conversion_error(desc, err);
+					}
+					else
 						conversion_error(desc, err);
 				}
 				if (!digit_seen)
@@ -418,7 +438,7 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
 			   the user know...  */
 
 			if (ABSOLUT(scale) > DBL_MAX_10_EXP)
-				(*err)(isc_arith_except, isc_arg_end);
+				(*err)(isc_arith_except, 0);
 
 /*
   Repeated division is a good way to mung the least significant bits
@@ -445,26 +465,27 @@ double CVT_get_double(const dsc* desc, FPTR_ERROR err)
 		break;
 
 	default:
-		(*err) (isc_badblk, isc_arg_end);	/* internal error */
+		(*err) (isc_badblk, 0);	/* internal error */
 		break;
 	}
 
 /* Last, but not least, adjust for scale */
 
-	if ((scale = desc->dsc_scale) == 0)
+	const int dscale = desc->dsc_scale;
+	if (dscale == 0)
 		return value;
 
 /* if the scale is greater than the power of 10 representable
    in a double number, then something has gone wrong... let
    the user know... */
 
-	if (ABSOLUT(scale) > DBL_MAX_10_EXP)
-		(*err) (isc_arith_except, isc_arg_end);
+	if (ABSOLUT(dscale) > DBL_MAX_10_EXP)
+		(*err) (isc_arith_except, 0);
 
-	if (scale > 0)
-		value *= power_of_ten(scale);
-	else if (scale < 0)
-		value /= power_of_ten(-scale);
+	if (dscale > 0)
+		value *= power_of_ten(dscale);
+	else if (dscale < 0)
+		value /= power_of_ten(-dscale);
 
 	return value;
 }
@@ -483,7 +504,7 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
  *      scale.
  *
  **************************************/
-	SLONG value, high, fraction;
+	SLONG value, high;
 
 	double d, eps;
 	SINT64 val64;
@@ -510,10 +531,10 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 
 		/* adjust for scale first, *before* range-checking the value. */
 		if (scale > 0) {
-			fraction = 0;
+			SLONG fraction = 0;
 			do {
 				if (scale == 1)
-					fraction = (SLONG) (val64 % 10);
+					fraction = SLONG(val64 % 10);
 				val64 /= 10;
 			} while (--scale);
 			if (fraction > 4)
@@ -530,12 +551,12 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		else if (scale < 0)
 			do {
 				if ((val64 > INT64_LIMIT) || (val64 < -INT64_LIMIT))
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 				val64 *= 10;
 			} while (++scale);
 
 		if ((val64 > LONG_MAX_int64) || (val64 < LONG_MIN_int64))
-			(*err) (isc_arith_except, isc_arg_end);
+			(*err) (isc_arith_except, 0);
 		return (SLONG) val64;
 
 	case dtype_quad:
@@ -543,7 +564,7 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		high = ((SLONG *) p)[HIGH_WORD];
 		if ((value >= 0 && !high) || (value < 0 && high == -1))
 			break;
-		(*err) (isc_arith_except, isc_arg_end);
+		(*err) (isc_arith_except, 0);
 		break;
 
 	case dtype_real:
@@ -585,12 +606,12 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		if (d < (double) LONG_MIN_real) {
 			if (d > (double) LONG_MIN_real - 1.)
 				return LONG_MIN;
-			(*err) (isc_arith_except, isc_arg_end);
+			(*err) (isc_arith_except, 0);
 		}
 		if (d > (double) LONG_MAX_real) {
 			if (d < (double) LONG_MAX_real + 1.)
 				return LONG_MAX_int;
-			(*err) (isc_arith_except, isc_arg_end);
+			(*err) (isc_arith_except, 0);
 		}
 		return (SLONG) d;
 
@@ -614,41 +635,34 @@ SLONG CVT_get_long(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		break;
 
 	default:
-		(*err)(isc_badblk, isc_arg_end);	/* internal error */
+		(*err)(isc_badblk, 0);	/* internal error */
 		break;
 	}
 
 /* Last, but not least, adjust for scale */
 
 	if (scale > 0) {
-		if (DTYPE_IS_EXACT(desc->dsc_dtype)) {
-			fraction = 0;
-			do {
-				if (scale == 1)
-					fraction = value % 10;
-				value /= 10;
-			} while (--scale);
-			if (fraction > 4)
-				value++;
-			/*
-			 * The following 2 lines are correct for platforms where
-			 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
-			 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
-			 * we'll have to change this depending on the platform.
-			 */
-			else if (fraction < -4)
-				value--;
-		}
-		else {
-			do {
-				value /= 10;
-			} while (--scale);
-		}
+		SLONG fraction = 0;
+		do {
+			if (scale == 1)
+				fraction = value % 10;
+			value /= 10;
+		} while (--scale);
+		if (fraction > 4)
+			value++;
+		/*
+		 * The following 2 lines are correct for platforms where
+		 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
+		 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
+		 * we'll have to change this depending on the platform.
+		 */
+		else if (fraction < -4)
+			value--;
 	}
 	else if (scale < 0) {
 		do {
 			if (value > LONG_LIMIT || value < -LONG_LIMIT)
-				(*err) (isc_arith_except, isc_arg_end);
+				(*err) (isc_arith_except, 0);
 			value *= 10;
 		} while (++scale);
 	}
@@ -787,9 +801,6 @@ SQUAD CVT_get_quad(const dsc* desc, SSHORT scale, FPTR_ERROR err)
  *
  **************************************/
 	SQUAD value;
-#ifdef NATIVE_QUAD
-	SLONG fraction;
-#endif
 	double d;
 	TEXT buffer[50];			/* long enough to represent largest quad in ASCII */
 
@@ -857,7 +868,7 @@ SQUAD CVT_get_quad(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 				return QUAD_MIN_int;
 			else if (d < (double) QUAD_MAX_real + 1.)
 				return QUAD_MAX_int;
-			(*err)(isc_arith_except, isc_arg_end);
+			(*err)(isc_arith_except, 0);
 		}
 		return QUAD_FROM_DOUBLE(d, err);
 
@@ -881,7 +892,7 @@ SQUAD CVT_get_quad(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		break;
 
 	default:
-		(*err)(isc_badblk, isc_arg_end);	/* internal error */
+		(*err)(isc_badblk, 0);	/* internal error */
 		break;
 	}
 
@@ -891,39 +902,30 @@ SQUAD CVT_get_quad(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		return value;
 
 #ifndef NATIVE_QUAD
-	(*err)(isc_badblk, isc_arg_end);	/* internal error */
+	(*err)(isc_badblk, 0);	/* internal error */
 #else
 	if (scale > 0) {
-		if (desc->dsc_dtype == dtype_short ||
-			desc->dsc_dtype == dtype_long || desc->dsc_dtype == dtype_quad)
-		{
-			fraction = 0;
-			do {
-				if (scale == 1)
-					fraction = value % 10;
-				value /= 10;
-			} while (--scale);
-			if (fraction > 4)
-				value++;
-			/*
-			 * The following 2 lines are correct for platforms where
-			 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
-			 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
-			 * we'll have to change this depending on the platform.
-			 */
-			else if (fraction < -4)
-				value--;
-		}
-		else {
-			do {
-				value /= 10;
-			} while (--scale);
-		}
+		SLONG fraction = 0;
+		do {
+			if (scale == 1)
+				fraction = value % 10;
+			value /= 10;
+		} while (--scale);
+		if (fraction > 4)
+			value++;
+		/*
+		 * The following 2 lines are correct for platforms where
+		 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
+		 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
+		 * we'll have to change this depending on the platform.
+		 */
+		else if (fraction < -4)
+			value--;
 	}
 	else {
 		do {
 			if (value > QUAD_LIMIT || value < -QUAD_LIMIT)
-				(*err) (isc_arith_except, isc_arg_end);
+				(*err) (isc_arith_except, 0);
 			value *= 10;
 		} while (++scale);
 	}
@@ -947,7 +949,6 @@ SINT64 CVT_get_int64(const dsc* desc, SSHORT scale, FPTR_ERROR err)
  *
  **************************************/
 	SINT64 value;
-	SLONG fraction;
 	double d, eps;
 	TEXT buffer[50];			/* long enough to represent largest SINT64 in ASCII */
 
@@ -1019,7 +1020,7 @@ SINT64 CVT_get_int64(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		   double, and thus will have no effect on the sum. */
 
 		if (d < (double) QUAD_MIN_real || (double) QUAD_MAX_real < d)
-			(*err) (isc_arith_except, isc_arg_end);
+			(*err) (isc_arith_except, 0);
 
 		return (SINT64) d;
 
@@ -1043,43 +1044,34 @@ SINT64 CVT_get_int64(const dsc* desc, SSHORT scale, FPTR_ERROR err)
 		break;
 
 	default:
-		(*err) (isc_badblk, isc_arg_end);	/* internal error */
+		(*err) (isc_badblk, 0);	/* internal error */
 		break;
 	}
 
 /* Last, but not least, adjust for scale */
 
 	if (scale > 0) {
-		if (desc->dsc_dtype == dtype_short ||
-			desc->dsc_dtype == dtype_long || desc->dsc_dtype == dtype_int64)
-		{
-			fraction = 0;
-			do {
-				if (scale == 1)
-					fraction = (SLONG) (value % 10);
-				value /= 10;
-			} while (--scale);
-			if (fraction > 4)
-				value++;
-			/*
-			 * The following 2 lines are correct for platforms where
-			 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
-			 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
-			 * we'll have to change this depending on the platform.
-			 */
-			else if (fraction < -4)
-				value--;
-		}
-		else {
-			do {
-				value /= 10;
-			} while (--scale);
-		}
+		SLONG fraction = 0;
+		do {
+			if (scale == 1)
+				fraction = (SLONG) (value % 10);
+			value /= 10;
+		} while (--scale);
+		if (fraction > 4)
+			value++;
+		/*
+		 * The following 2 lines are correct for platforms where
+		 * ((-85 / 10 == -8) && (-85 % 10 == -5)).  If we port to
+		 * a platform where ((-85 / 10 == -9) && (-85 % 10 == 5)),
+		 * we'll have to change this depending on the platform.
+		 */
+		else if (fraction < -4)
+			value--;
 	}
 	else if (scale < 0) {
 		do {
 			if (value > INT64_LIMIT || value < -INT64_LIMIT)
-				(*err) (isc_arith_except, isc_arg_end);
+				(*err) (isc_arith_except, 0);
 			value *= 10;
 		} while (++scale);
 	}
@@ -1301,15 +1293,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
  *      Move (and possible convert) something to something else.
  *
  **************************************/
-	SSHORT fill;
 	SLONG l;
-	UCHAR *ptr;
-	USHORT strtype;
-#if !defined(REQUESTER) && !defined(SUPERCLIENT)
-	CHARSET_ID charset1, charset2;
-#endif
-	UCHAR fill_char;
-
 	SLONG length = from->dsc_length;
 	UCHAR* p = to->dsc_address;
 	const UCHAR* q = from->dsc_address;
@@ -1337,35 +1321,24 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 		case dtype_text:
 			{
 				GDS_TIMESTAMP date;
-				tm times;
-
 				string_to_datetime(from, &date, expect_timestamp, err);
-
-				isc_decode_timestamp(&date, &times);
-				if ((times.tm_year + 1900) < MIN_YEAR
-					|| (times.tm_year) + 1900 > MAX_YEAR)
-				{
-					(*err) (isc_date_range_exceeded, isc_arg_end);
-				}
-
-				((GDS_TIMESTAMP *) to->dsc_address)->timestamp_date = date.timestamp_date;
-				((GDS_TIMESTAMP *) to->dsc_address)->timestamp_time = date.timestamp_time;
+				*((GDS_TIMESTAMP *) to->dsc_address) = date;
 			}
 			return;
 
 		case dtype_sql_date:
 			((GDS_TIMESTAMP *) (to->dsc_address))->timestamp_date =
-				*(GDS_DATE *) (from->dsc_address);
+				*(GDS_DATE *) from->dsc_address;
 			((GDS_TIMESTAMP *) (to->dsc_address))->timestamp_time = 0;
 			return;
 
 		case dtype_sql_time:
 			((GDS_TIMESTAMP*) (to->dsc_address))->timestamp_date = 0;
 			((GDS_TIMESTAMP*) (to->dsc_address))->timestamp_time =
-				*(GDS_TIME*) (from->dsc_address);
+				*(GDS_TIME*) from->dsc_address;
 
 			/* Per SQL Specs, we need to set the DATE
-			   portion to the current date */
+				portion to the current date */
 			{
 				/** Cannot call JRD_get_thread_data because that macro calls
 				BUGCHECK i.e. ERR_bugcheck() which is not part of
@@ -1380,11 +1353,10 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 				   date portion of the timestamp */
 
 				SLONG cur_date;
-				if (tdbb && (tdbb->getType() == ThreadData::tddDBB) &&
-					tdbb->tdbb_request)
+				if (tdbb && (tdbb->getType() == ThreadData::tddDBB) && tdbb->getRequest())
 				{
-					fb_assert(!tdbb->tdbb_request->req_timestamp.isEmpty());
-					cur_date = tdbb->tdbb_request->req_timestamp.value().timestamp_date;
+					fb_assert(!tdbb->getRequest()->req_timestamp.isEmpty());
+					cur_date = tdbb->getRequest()->req_timestamp.value().timestamp_date;
 				}
 				else
 				{
@@ -1417,39 +1389,14 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 		case dtype_text:
 			{
 				GDS_TIMESTAMP date;
-				tm times;
-
 				string_to_datetime(from, &date, expect_sql_date, err);
-				isc_decode_timestamp(&date, &times);
-				if ((times.tm_year + 1900) < MIN_YEAR
-					|| (times.tm_year) + 1900 > MAX_YEAR)
-				{
-					(*err) (isc_date_range_exceeded, isc_arg_end);
-				}
-
 				*((GDS_DATE *) to->dsc_address) = date.timestamp_date;
 			}
 			return;
 
 		case dtype_timestamp:
-			{
-				GDS_TIMESTAMP new_date;
-				tm times;
-
-				new_date.timestamp_date =
-					((GDS_TIMESTAMP *) from->dsc_address)->timestamp_date;
-				new_date.timestamp_time = 0;
-
-				isc_decode_timestamp(&new_date, &times);
-				if ((times.tm_year + 1900) < MIN_YEAR
-					|| (times.tm_year) + 1900 > MAX_YEAR)
-				{
-					(*err) (isc_date_range_exceeded, isc_arg_end);
-				}
-
-				*((GDS_DATE *) to->dsc_address) =
-					((GDS_TIMESTAMP *) from->dsc_address)->timestamp_date;
-			}
+			*((GDS_DATE *) to->dsc_address) =
+				((GDS_TIMESTAMP *) from->dsc_address)->timestamp_date;
 			return;
 
 		default:
@@ -1477,7 +1424,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 			{
 				GDS_TIMESTAMP date;
 				string_to_datetime(from, &date, expect_sql_time, err);
-				*(GDS_TIME *) to->dsc_address = date.timestamp_time;
+				*((GDS_TIME *) to->dsc_address) = date.timestamp_time;
 			}
 			return;
 
@@ -1535,6 +1482,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 
 #ifndef REQUESTER
 #ifndef SUPERCLIENT
+			CHARSET_ID charset1, charset2;
 			if ((INTL_TTYPE(from) == ttype_dynamic) && (err == ERR_post))
 				charset1 = INTL_charset(NULL, INTL_TTYPE(from));
 			else
@@ -1562,49 +1510,50 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 				else
 #endif
 #endif
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 #ifndef REQUESTER
 #ifndef SUPERCLIENT
 			}
 #endif
 #endif
 
-			length = l =
-				CVT_get_string_ptr(from, &strtype, &ptr, NULL, 0, err);
-			q = ptr;
+			{ // scope
+				USHORT strtype_unused;
+				UCHAR *ptr;
+				length = l =
+					CVT_get_string_ptr(from, &strtype_unused, &ptr, NULL, 0, err);
+				q = ptr;
+			} // end scope
 
 			USHORT to_size = TEXT_LEN(to);
 			UCHAR* start = to->dsc_address;
+			UCHAR fill_char = ASCII_SPACE;
 #if !defined(REQUESTER) && !defined(SUPERCLIENT)
 			CharSet* toCharSet = (err != ERR_post || charset2 == ttype_dynamic || charset2 == CS_METADATA ?
 									NULL : INTL_charset_lookup(NULL, charset2));
+			if (charset2 == ttype_binary)
+				fill_char = 0x00;
+				
 			USHORT toLength;
 #endif
+			int fill;
 
 			switch (to->dsc_dtype) {
 			case dtype_text:
 				length = MIN(length, to->dsc_length);
 #if !defined(REQUESTER) && !defined(SUPERCLIENT)
 				if (toCharSet && !toCharSet->wellFormed(length, q))
-					(*err)(isc_malformed_string, isc_arg_end);
+					(*err)(isc_malformed_string, 0);
 				toLength = length;
 #endif
 
 				l -= length;
 				/* TMN: Here we should really have the following fb_assert */
 				/* fb_assert((to->dsc_length - length) <= MAX_SSHORT); */
-				fill = (SSHORT) (to->dsc_length - length);
+				fill = to->dsc_length - length;
 
 				CVT_COPY_BUFF(q, p, length);
 				if (fill > 0) {
-#ifndef REQUESTER
-#ifndef SUPERCLIENT
-					if (charset2 == ttype_binary)
-						fill_char = 0x00;
-					else
-#endif
-#endif
-						fill_char = ASCII_SPACE;
 					do {
 						*p++ = fill_char;
 					} while (--fill);
@@ -1622,7 +1571,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 				length = MIN(length, to->dsc_length - 1);
 #if !defined(REQUESTER) && !defined(SUPERCLIENT)
 				if (toCharSet && !toCharSet->wellFormed(length, q))
-					(*err)(isc_malformed_string, isc_arg_end);
+					(*err)(isc_malformed_string, 0);
 				toLength = length;
 #endif
 
@@ -1636,7 +1585,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 					MIN(length, (SLONG) (to->dsc_length - sizeof(USHORT)));
 #if !defined(REQUESTER) && !defined(SUPERCLIENT)
 				if (toCharSet && !toCharSet->wellFormed(length, q))
-					(*err)(isc_malformed_string, isc_arg_end);
+					(*err)(isc_malformed_string, 0);
 				toLength = length;
 #endif
 
@@ -1658,22 +1607,21 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 				if (toCharSet->isMultiByte() &&
 					!(toCharSet->getFlags() & CHARSET_LEGACY_SEMANTICS) &&
 					toLength != 31 &&	// allow non CHARSET_LEGACY_SEMANTICS to be used as connection charset
-					toCharSet->length(tdbb, toLength, start, false) > to_size / toCharSet->maxBytesPerChar())
+					toCharSet->length(toLength, start, false) > to_size / toCharSet->maxBytesPerChar())
 				{
-					(*err)(isc_arith_except, isc_arg_end);
+					(*err)(isc_arith_except, 0);
 				}
 			}
 #endif
 
 			if (l) {
-				/* scan the truncated string to ensure only spaces lost */
-				/* Note: it  is correct only for narrow and multi-byte
-				   character sets which use ASCII for the SPACE
-				   character. */
+				// Scan the truncated string to ensure only spaces lost
+				// Warning: it is correct only for narrow and multi-byte
+				// character sets which use ASCII or NULL for the SPACE character
 
 				do {
-					if (*q++ != ASCII_SPACE)
-						(*err) (isc_arith_except, isc_arg_end);
+					if (*q++ != fill_char)
+						(*err) (isc_arith_except, 0);
 				} while (--l);
 			}
 			return;
@@ -1718,7 +1666,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 
 		if (to->dsc_dtype != from->dsc_dtype)
 			(*err) (isc_wish_list, isc_arg_gds, isc_blobnotsup,
-					isc_arg_string, "move", isc_arg_end);
+					isc_arg_string, "move", 0);
 
 		/* Note: DSC_EQUIV failed above as the blob sub_types were different,
 		 * or their character sets were different.  In V4 we aren't trying
@@ -1735,7 +1683,7 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 		/* fb_assert(l <= MAX_SSHORT); */
 		*(SSHORT *) p = (SSHORT) l;
 		if (*(SSHORT *) p != l)
-			(*err) (isc_arith_except, isc_arg_end);
+			(*err) (isc_arith_except, 0);
 		return;
 
 	case dtype_long:
@@ -1757,10 +1705,9 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 
 	case dtype_real:
 		{
-			double d_value;
-			d_value = CVT_get_double(from, err);
+			double d_value = CVT_get_double(from, err);
 			if (ABSOLUT(d_value) > FLOAT_MAX)
-				(*err) (isc_arith_except, isc_arg_end);
+				(*err) (isc_arith_except, 0);
 			*(float*) p = (float) d_value;
 		}
 		return;
@@ -1785,10 +1732,10 @@ void CVT_move(const dsc* from, dsc* to, FPTR_ERROR err)
 	if (from->dsc_dtype == dtype_array || from->dsc_dtype == dtype_blob)
 	{
 		(*err) (isc_wish_list, isc_arg_gds, isc_blobnotsup,
-		        isc_arg_string, "move", isc_arg_end);
+		        isc_arg_string, "move", 0);
 	}
 
-	(*err) (isc_badblk, isc_arg_end);	/* internal error */
+	(*err) (isc_badblk, 0);	/* internal error */
 }
 
 
@@ -1823,7 +1770,7 @@ static void conversion_error(const dsc* desc, FPTR_ERROR err)
 #endif
 	}
 
-	(*err)(isc_convert_error, isc_arg_string, p, isc_arg_end);
+	(*err)(isc_convert_error, isc_arg_string, p, 0);
 }
 
 
@@ -1846,40 +1793,34 @@ static void datetime_to_text(const dsc* from, dsc* to, FPTR_ERROR err)
 
 /* Convert a date or time value into a timestamp for manipulation */
 
-	GDS_TIMESTAMP date;
-	
+	Firebird::TimeStamp ts(true);
+
 	switch (from->dsc_dtype) {
 	case dtype_sql_time:
-		date.timestamp_date = 0;
-		date.timestamp_time = *(GDS_TIME *) from->dsc_address;
+		ts.value().timestamp_time = *(GDS_TIME *) from->dsc_address;
 		break;
 	case dtype_sql_date:
-		date.timestamp_date = *(GDS_DATE *) from->dsc_address;
-		date.timestamp_time = 0;
+		ts.value().timestamp_date = *(GDS_DATE *) from->dsc_address;
 		break;
 	case dtype_timestamp:
 		/** Cannot call JRD_get_thread_data because that macro calls 
 			BUGCHECK i.e. ERR_bugcheck() which is not part of 
 			client library **/
 		tdbb = (thread_db*) ThreadData::getSpecific();
-		if (tdbb && (tdbb->getType() == ThreadData::tddDBB) &&
-			tdbb->tdbb_request)
-		{
-			version4 = (tdbb->tdbb_request->req_flags & req_blr_version4) ?
-				true : false;
-		}
-		date = *(GDS_TIMESTAMP *) from->dsc_address;
+		if (tdbb && (tdbb->getType() == ThreadData::tddDBB) && tdbb->getRequest())
+			version4 = (tdbb->getRequest()->req_flags & req_blr_version4) ? true : false;
+		ts.value() = *(GDS_TIMESTAMP *) from->dsc_address;
 		break;
 	default:
 		fb_assert(false);
-		(*err) (isc_badblk, isc_arg_end);	/* internal error */
+		(*err) (isc_badblk, 0);	/* internal error */
 		break;
 	}
 
 /* Decode the timestamp into human readable terms */
 
 	tm times;
-	isc_decode_timestamp(&date, &times);
+	ts.decode(&times);
 	
 	TEXT temp[30];			/* yyyy-mm-dd hh:mm:ss.tttt  OR
 							dd-MMM-yyyy hh:mm:ss.tttt */
@@ -1915,17 +1856,17 @@ static void datetime_to_text(const dsc* from, dsc* to, FPTR_ERROR err)
 		if (from->dsc_dtype == dtype_sql_time || !version4) {
 			sprintf(p, "%2.2d:%2.2d:%2.2d.%4.4d",
 					times.tm_hour, times.tm_min, times.tm_sec,
-					(USHORT) (date.timestamp_time %
+					(USHORT) (ts.value().timestamp_time %
 							  ISC_TIME_SECONDS_PRECISION));
 		}
 		else if (times.tm_hour || times.tm_min || times.tm_sec
-				 || date.timestamp_time)
+				 || ts.value().timestamp_time)
 		{
 			/* Timestamp formating prior to BLR Version 5 is slightly
 			   different */
 			sprintf(p, " %d:%.2d:%.2d.%.4d",
 					times.tm_hour, times.tm_min, times.tm_sec,
-					(USHORT) (date.timestamp_time %
+					(USHORT) (ts.value().timestamp_time %
 							  ISC_TIME_SECONDS_PRECISION));
 		}
 		while (*p)
@@ -1976,7 +1917,7 @@ static SSHORT decompose(const char* string,
    supported by the platform as a native datatype. */
 
 	if (dtype == dtype_quad)
-		(*err) (isc_badblk, isc_arg_end);	/* internal error */
+		(*err) (isc_badblk, 0);	/* internal error */
 #endif
 
 	dsc errd;
@@ -1987,7 +1928,8 @@ static SSHORT decompose(const char* string,
 	errd.dsc_address = reinterpret_cast<UCHAR*>(const_cast<char*>(string));
 
 	SINT64 value = 0;
-	SSHORT scale = 0, sign = 0;
+	SSHORT scale = 0;
+	int sign = 0;
 	bool digit_seen = false, fraction = false;
 	const SINT64 lower_limit = (dtype == dtype_long) ? MIN_SLONG : MIN_SINT64;
 	const SINT64 upper_limit = (dtype == dtype_long) ? MAX_SLONG : MAX_SINT64;
@@ -1996,11 +1938,14 @@ static SSHORT decompose(const char* string,
 
 	const char* p = string;
 	const char* const end = p + length;
+
+	// skip initial spaces
+	for (; p < end && *p == ' '; p++)
+		;
+
 	for (; p < end; p++)
 	{
-		if (*p == ',')
-			continue;
-		else if (DIGIT(*p)) {
+		if (DIGIT(*p)) {
 			digit_seen = true;
 
 			/* Before computing the next value, make sure there will be
@@ -2011,11 +1956,11 @@ static SSHORT decompose(const char* string,
 			if (value >= limit_by_10) {
 				/* possibility of an overflow */
 				if (value > limit_by_10)
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 				else if (((*p > '8') && (sign == -1))
 						 || ((*p > '7') && (sign != -1)))
 				{
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 				}
 			}
 
@@ -2035,11 +1980,20 @@ static SSHORT decompose(const char* string,
 			sign = 1;
 		else if (*p == 'e' || *p == 'E')
 			break;
-		else if (*p != ' ')
+		else if (*p == ' ')
+		{
+			// skip spaces
+			for (; p < end && *p == ' '; p++)
+				;
+
+			// throw if there is something after the spaces
+			if (p < end)
+				conversion_error(&errd, err);
+		}
+		else
 			conversion_error(&errd, err);
 	}
 
-
 	if (!digit_seen)
 		conversion_error(&errd, err);
 
@@ -2062,13 +2016,23 @@ static SSHORT decompose(const char* string,
 				   applied to the value. */
 
 				if (exp >= SHORT_LIMIT)
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 			}
 			else if (*p == '-' && !digit_seen && !sign)
 				sign = -1;
 			else if (*p == '+' && !digit_seen && !sign)
 				sign = 1;
-			else if (*p != ' ')
+			else if (*p == ' ')
+			{
+				// skip spaces
+				for (; p < end && *p == ' '; p++)
+					;
+
+				// throw if there is something after the spaces
+				if (p < end)
+					conversion_error(&errd, err);
+			}
+			else
 				conversion_error(&errd, err);
 		}
 		if (sign == -1)
@@ -2198,7 +2162,7 @@ static void float_to_text(const dsc* from, dsc* to, FPTR_ERROR err)
 			/* If we cannot print at least two digits, one on each side of the
 			   ".", report an overflow exception. */
 			if (precision < 2)
-				(*err) (isc_arith_except, isc_arg_end);
+				(*err) (isc_arith_except, 0);
 
 			chars_printed = sprintf(temp, num_format, width, precision, d);
 
@@ -2210,14 +2174,19 @@ static void float_to_text(const dsc* from, dsc* to, FPTR_ERROR err)
 			if (chars_printed > width) {
 				precision -= (chars_printed - width);
 				if (precision < 2)
-					(*err) (isc_arith_except, isc_arg_end);
+					(*err) (isc_arith_except, 0);
 			    chars_printed = sprintf(temp, num_format, width, precision, d);
 			}
 		}
 	}
 	fb_assert(chars_printed <= width);
 
-/* Now move the result to the destination array. */
+	// trim trailing spaces
+	const char* p = strchr(temp + 1, ' ');
+	if (p)
+		chars_printed = p - temp;
+
+	// Now move the result to the destination array.
 
 	dsc intermediate;
 	intermediate.dsc_dtype = dtype_text;
@@ -2261,7 +2230,7 @@ static void integer_to_text(const dsc* from, dsc* to, FPTR_ERROR err)
    supported by the platform as a native datatype. */
 
 	if (from->dsc_dtype == dtype_quad)
-		(*err) (isc_badblk, isc_arg_end);	/* internal error */
+		(*err) (isc_badblk, 0);	/* internal error */
 #endif
 
 	SSHORT pad_count = 0, decimal = 0, neg = 0;
@@ -2457,53 +2426,54 @@ static void string_to_datetime(
  *
  **************************************/
 
-/* Values inside of description
-     > 0 is number of digits 
-       0 means missing 
-     ENGLISH_MONTH for the presence of an English month name
-     SPECIAL       for a special date verb */
-#define		ENGLISH_MONTH	-1
-#define		SPECIAL		-2
-	USHORT position_year = 0;
-	USHORT position_month = 1;
-	USHORT position_day = 2;
+	// Values inside of description
+	// > 0 is number of digits
+	//   0 means missing
+	// ENGLISH_MONTH for the presence of an English month name
+	// SPECIAL       for a special date verb
+	const int ENGLISH_MONTH	= -1;
+	const int SPECIAL		= -2; // CVC: I see it set, but never tested.
+
+	unsigned int position_year = 0;
+	unsigned int position_month = 1;
+	unsigned int position_day = 2;
 	bool have_english_month = false;
 	bool dot_separator_seen = false;
-	TEXT buffer[100];			/* arbitrarily large */
+	TEXT buffer[100];			// arbitrarily large
 
-	const char* string;
+	const char* p = NULL;
 	const USHORT length =
-		CVT_make_string(desc, ttype_ascii, &string,
-						(vary*) buffer, sizeof(buffer), err);
+		CVT_make_string(desc, ttype_ascii, &p, (vary*) buffer, sizeof(buffer), err);
 						
-	const char* p = string;
 	const char* const end = p + length;
 
 	USHORT n, components[7];
-	SSHORT description[7];
+	int description[7];
 	memset(components, 0, sizeof(components));
 	memset(description, 0, sizeof(description));
 
-/* Parse components */
-/* The 7 components are Year, Month, Day, Hours, Minutes, Seconds, Thou */
-/* The first 3 can be in any order */
+	// Parse components
+	// The 7 components are Year, Month, Day, Hours, Minutes, Seconds, Thou
+	// The first 3 can be in any order
 
 	const int start_component = (expect_type == expect_sql_time) ? 3 : 0;
 	int i;
-	for (i = start_component; i < 7; i++) {
+	for (i = start_component; i < 7; i++) 
+	{
 
-		/* Skip leading blanks.  If we run out of characters, we're done
-		   with parse.  */
+		// Skip leading blanks.  If we run out of characters, we're done
+		// with parse.
 
 		while (p < end && (*p == ' ' || *p == '\t'))
 			p++;
 		if (p == end)
 			break;
 
-		/* Handle digit or character strings */
+		// Handle digit or character strings
 
 		TEXT c = UPPER7(*p);
-		if (DIGIT(c)) {
+		if (DIGIT(c))
+		{
 			USHORT precision = 0;
 			n = 0;
 			while (p < end && DIGIT(*p)) {
@@ -2512,7 +2482,8 @@ static void string_to_datetime(
 			}
 			description[i] = precision;
 		}
-		else if (LETTER7(c) && !have_english_month) {
+		else if (LETTER7(c) && !have_english_month)
+		{
 			TEXT temp[sizeof(YESTERDAY) + 1];
 
 			TEXT* t = temp;
@@ -2525,7 +2496,7 @@ static void string_to_datetime(
 			}
 			*t = 0;
 
-			/* Insist on at least 3 characters for month names */
+			// Insist on at least 3 characters for month names
 			if (t - temp < 3) {
 				conversion_error(desc, err);
 				return;
@@ -2533,7 +2504,7 @@ static void string_to_datetime(
 
 			const TEXT* const* month_ptr = FB_LONG_MONTHS_UPPER;
 			while (true) {
-				/* Month names are only allowed in first 2 positions */
+				// Month names are only allowed in first 2 positions
 				if (*month_ptr && i < 2) {
 					t = temp;
 					const TEXT* m = *month_ptr++;
@@ -2545,8 +2516,8 @@ static void string_to_datetime(
 						break;
 				}
 				else {
-					/* it's not a month name, so it's either a magic word or
-					   a non-date string.  If there are more characters, it's bad */
+					// it's not a month name, so it's either a magic word or
+					// a non-date string.  If there are more characters, it's bad
 
 					description[i] = SPECIAL;
 
@@ -2554,7 +2525,7 @@ static void string_to_datetime(
 						if (*p != ' ' && *p != '\t' && *p != 0)
 							conversion_error(desc, err);
 
-					/* fetch the current time */
+					// fetch the current datetime
 					*date = Firebird::TimeStamp().value();
 
 					if (strcmp(temp, NOW) == 0)
@@ -2583,22 +2554,23 @@ static void string_to_datetime(
 			description[i] = ENGLISH_MONTH;
 			have_english_month = true;
 		}
-		else {					/* Not a digit and not a letter - must be punctuation */
-
+		else
+		{
+			// Not a digit and not a letter - must be punctuation
 			conversion_error(desc, err);
 			return;
 		}
 
 		components[i] = n;
 
-		/* Grab whitespace following the number */
+		// Grab whitespace following the number
 		while (p < end && (*p == ' ' || *p == '\t'))
 			p++;
 
 		if (p == end)
 			break;
 
-		/* Grab a separator character */
+		// Grab a separator character
 		if (*p == '/' || *p == '-' || *p == ',' || *p == ':') {
 			p++;
 			continue;
@@ -2607,66 +2579,78 @@ static void string_to_datetime(
 			if (i <= 1)
 				dot_separator_seen = true;
 			p++;
-			continue;
+			//continue;
 		}
 	}
 
-/* User must provide at least 2 components */
+	// User must provide at least 2 components
 	if (i - start_component < 1) {
 		conversion_error(desc, err);
 		return;
 	}
 
-/* Dates cannot have a Time portion */
+	// Dates cannot have a Time portion
 	if (expect_type == expect_sql_date && i > 2) {
 		conversion_error(desc, err);
 		return;
 	}
 
-	tm times, times2;
+	// We won't allow random trash after the recognized string
+	while (p < end)
+	{
+		if (*p != ' ' && *p != '\t' && p != 0)
+		{
+			conversion_error(desc, err);
+			return;
+		}
+		++p;
+	}
+
+	tm times;
 	memset(&times, 0, sizeof(times));
 
-	if (expect_type != expect_sql_time) {
-		/* Figure out what format the user typed the date in */
+	if (expect_type != expect_sql_time)
+	{
+		// Figure out what format the user typed the date in
 
-		/* A 4 digit number to start implies YYYY-MM-DD */
+		// A 4 digit number to start implies YYYY-MM-DD
 		if (description[0] >= 3) {
 			position_year = 0;
 			position_month = 1;
 			position_day = 2;
 		}
 
-		/* An English month to start implies MM-DD-YYYY */
+		// An English month to start implies MM-DD-YYYY
 		else if (description[0] == ENGLISH_MONTH) {
 			position_year = 2;
 			position_month = 0;
 			position_day = 1;
 		}
 
-		/* An English month in the middle implies DD-MM-YYYY */
+		// An English month in the middle implies DD-MM-YYYY
 		else if (description[1] == ENGLISH_MONTH) {
 			position_year = 2;
 			position_month = 1;
 			position_day = 0;
 		}
 
-		/* A period as a separator implies DD.MM.YYYY */
+		// A period as a separator implies DD.MM.YYYY
 		else if (dot_separator_seen) {
 			position_year = 2;
 			position_month = 1;
 			position_day = 0;
 		}
 
-		/* Otherwise assume MM-DD-YYYY */
+		// Otherwise assume MM-DD-YYYY
 		else {
 			position_year = 2;
 			position_month = 0;
 			position_day = 1;
 		}
 
-		/* Forbid years with more than 4 digits */
-		/* Forbid months or days with more than 2 digits */
-		/* Forbid months or days being missing */
+		// Forbid years with more than 4 digits
+		// Forbid months or days with more than 2 digits
+		// Forbid months or days being missing
 		if (description[position_year] > 4 ||
 			description[position_month] > 2
 			|| description[position_month] == 0
@@ -2677,22 +2661,23 @@ static void string_to_datetime(
 			return;
 		}
 
-		/* Slide things into day, month, year form */
+		// Slide things into day, month, year form
 
 		times.tm_year = components[position_year];
 		times.tm_mon = components[position_month];
 		times.tm_mday = components[position_day];
 
 		// Fetch current date/time
+		tm times2;
 		Firebird::TimeStamp().decode(&times2);
 
-		/* Handle defaulting of year */
+		// Handle defaulting of year
 
 		if (description[position_year] == 0) {
 			times.tm_year = times2.tm_year + 1900;
 		}
 
-		/* Handle conversion of 2-digit years */
+		// Handle conversion of 2-digit years
 
 		else if (description[position_year] <= 2) {
 			if (times.tm_year < (times2.tm_year - 50) % 100)
@@ -2705,49 +2690,56 @@ static void string_to_datetime(
 		times.tm_mon -= 1;
 	}
 	else {
-		/* The date portion isn't needed for time - but to
-		   keep the conversion in/out of isc_time clean lets
-		   intialize it properly anyway */
+		// The date portion isn't needed for time - but to
+		// keep the conversion in/out of isc_time clean lets
+		// initialize it properly anyway
 		times.tm_year = 0;
 		times.tm_mon = 0;
 		times.tm_mday = 1;
 	}
 
-/* Handle time values out of range - note possibility of 60 for
- * seconds value due to LEAP second (Not supported in V4.0).
- */
-	if (((times.tm_hour = components[3]) > 23) ||
-		((times.tm_min = components[4]) > 59) ||
-		((times.tm_sec = components[5]) > 59) ||
-		(description[6] > -ISC_TIME_SECONDS_PRECISION_SCALE))
+	// Handle time values out of range - note possibility of 60 for
+	// seconds value due to LEAP second (Not supported in V4.0).
+	if (i > 2 &&
+		(((times.tm_hour = components[3]) > 23) ||
+		 ((times.tm_min = components[4]) > 59) ||
+		 ((times.tm_sec = components[5]) > 59) ||
+		 description[3] > 2 || description[3] == 0 ||
+		 description[4] > 2 || description[4] == 0 ||
+		 description[5] > 2 ||
+		 description[6] > -ISC_TIME_SECONDS_PRECISION_SCALE))
 	{
 		conversion_error(desc, err);
 	}
 
-/* convert day/month/year to Julian and validate result
-   This catches things like 29-Feb-1995 (not a leap year) */
+	// convert day/month/year to Julian and validate result
+	// This catches things like 29-Feb-1995 (not a leap year)
+
+	Firebird::TimeStamp ts(true);
+	ts.encode(&times);
+
+	if (!ts.isRangeValid()) {
+		(*err) (isc_date_range_exceeded, 0);
+	}
 
-	isc_encode_timestamp(&times, date);
 	if (expect_type != expect_sql_time) {
-		isc_decode_timestamp(date, &times2);
-
-		if ((times.tm_year + 1900) < MIN_YEAR
-			|| (times.tm_year) + 1900 > MAX_YEAR)
-		{
-			(*err) (isc_date_range_exceeded, isc_arg_end);
-		}
+		tm times2;
+		ts.decode(&times2);
 
 		if (times.tm_year != times2.tm_year ||
 			times.tm_mon != times2.tm_mon ||
 			times.tm_mday != times2.tm_mday ||
 			times.tm_hour != times2.tm_hour ||
-			times.tm_min != times2.tm_min || times.tm_sec != times2.tm_sec)
+			times.tm_min != times2.tm_min ||
+			times.tm_sec != times2.tm_sec)
 		{
 			conversion_error(desc, err);
 		}
 	}
 
-/* Convert fraction of seconds */
+	*date = ts.value();
+
+	// Convert fraction of seconds
 	while (description[6]++ < -ISC_TIME_SECONDS_PRECISION_SCALE)
 		components[6] *= 10;
 
