@@ -65,10 +65,6 @@
 #define MUTEX		event_mutex
 #endif
 
-#ifdef SERVER
-#undef MULTI_THREAD
-#endif
-
 #ifndef AST_TYPE
 #define AST_TYPE	void
 #endif
@@ -139,8 +135,6 @@ void EVENT_cancel(SLONG request_id)
  *	Cancel an outstanding event.
  *
  **************************************/
-	srq *event_srq, *que2;
-
 	if (!EVENT_header)
 		return;
 
@@ -148,8 +142,10 @@ void EVENT_cancel(SLONG request_id)
 
 	PRB process = (PRB) SRQ_ABS_PTR(EVENT_process_offset);
 
+	srq* que2;
 	SRQ_LOOP(process->prb_sessions, que2) {
 		SES session = (SES) ((UCHAR *) que2 - OFFSET(SES, ses_sessions));
+		srq* event_srq;
 		SRQ_LOOP(session->ses_requests, event_srq) {
 			EVT_REQ request = (EVT_REQ) ((UCHAR *) event_srq - OFFSET(EVT_REQ, req_requests));
 			if (request->req_request_id == request_id) {
@@ -164,7 +160,8 @@ void EVENT_cancel(SLONG request_id)
 }
 
 
-SLONG EVENT_create_session(ISC_STATUS * status_vector)
+#ifndef SUPERCLIENT
+SLONG EVENT_create_session(ISC_STATUS* status_vector)
 {
 /**************************************
  *
@@ -178,7 +175,7 @@ SLONG EVENT_create_session(ISC_STATUS * status_vector)
  **************************************/
 // If we're not initialized, do so now.
 
-	if (!EVENT_header && !EVENT_init(status_vector, true))
+	if (!EVENT_header && !EVENT_init(status_vector))
 		return 0;
 
 	if (!EVENT_process_offset)
@@ -200,6 +197,7 @@ SLONG EVENT_create_session(ISC_STATUS * status_vector)
 
 	return id;
 }
+#endif
 
 
 void EVENT_delete_session(SLONG session_id)
@@ -242,7 +240,6 @@ void EVENT_deliver()
  *  for delivery with EVENT_post.
  *
  **************************************/
-	srq	*event_srq;
 
 	/* If we're not initialized, do so now */
 
@@ -257,6 +254,7 @@ void EVENT_deliver()
 
 	while (flag) {
 		flag = false;
+		srq* event_srq;
 		SRQ_LOOP (EVENT_header->evh_processes, event_srq)	{
 			PRB process = (PRB) ((UCHAR*) event_srq - OFFSET (PRB, prb_processes));
 			if (process->prb_flags & PRB_wakeup) {
@@ -271,9 +269,7 @@ void EVENT_deliver()
 }
 
 
-// CVC: Contrary to the explanation, server_flag is not used!
-#pragma FB_COMPILER_MESSAGE("server_flag var is not honored: bug or deprecated?")
-EVH EVENT_init(ISC_STATUS* status_vector, bool server_flag)
+EVH EVENT_init(ISC_STATUS* status_vector)
 {
 /**************************************
  *
@@ -282,9 +278,8 @@ EVH EVENT_init(ISC_STATUS* status_vector, bool server_flag)
  **************************************
  *
  * Functional description
- *	Initialize for access to shared global region.  If "server_flag" is true,
- *	create region if it doesn't exist.  Return address of header if region
- *	exits, otherwise return NULL.
+ *	Initialize for access to shared global region.
+ *	Return address of header.
  *
  **************************************/
 	TEXT buffer[MAXPATHLEN];
@@ -296,28 +291,7 @@ EVH EVENT_init(ISC_STATUS* status_vector, bool server_flag)
 
 	EVENT_default_size = Config::getEventMemSize();
 
-#ifdef SERVER
-	EVENT_data.sh_mem_address = EVENT_header =
-		(EVH) gds__alloc((SLONG) EVENT_default_size);
-/* FREE: apparently only freed at process exit */
-	if (!EVENT_header) {		/* NOMEM: */
-		status_vector[0] = isc_arg_gds;
-		status_vector[1] = isc_virmemexh;
-		status_vector[2] = isc_arg_end;
-		return NULL;
-	}
-#ifdef DEBUG_GDS_ALLOC
-/* This structure is apparently only freed when the process exits */
-/* 1994-October-25 David Schnepper */
-	gds_alloc_flag_unfreed((void *) EVENT_header);
-#endif /* DEBUG_GDS_ALLOC */
-
-	EVENT_data.sh_mem_length_mapped = EVENT_default_size;
-	EVENT_data.sh_mem_mutex_arg = 0;
-	init((SLONG) 0, &EVENT_data, true);
-#else
-
-#if (defined UNIX)
+#ifdef UNIX
 	EVENT_data.sh_mem_semaphores = SEMAPHORES;
 #endif
 
@@ -329,14 +303,9 @@ EVH EVENT_init(ISC_STATUS* status_vector, bool server_flag)
 											init, 0, EVENT_default_size,
 											&EVENT_data))) 
 	{
-#ifdef SERVER
-		gds__free(EVENT_data.sh_mem_address);
-#endif /* SERVER */
 		return NULL;
 	}
 
-#endif
-
 	gds__register_cleanup(exit_handler, 0);
 
 #ifdef UNIX
@@ -366,11 +335,10 @@ int EVENT_post(ISC_STATUS * status_vector,
  *	Post an event.
  *
  **************************************/
-	srq *event_srq;
 
 /* If we're not initialized, do so now */
 
-	if (!EVENT_header && !EVENT_init(status_vector, false))
+	if (!EVENT_header && !EVENT_init(status_vector))
 		return status_vector[1];
 
 	acquire();
@@ -381,6 +349,7 @@ int EVENT_post(ISC_STATUS * status_vector,
 		(event = find_event(minor_length, minor_code, parent)))
 	{
 		event->evnt_count += count;
+		srq* event_srq;
 		SRQ_LOOP(event->evnt_interests, event_srq) {
 			RINT interest = (RINT) ((UCHAR *) event_srq - OFFSET(RINT, rint_interests));
 			if (interest->rint_request) {
@@ -536,10 +505,8 @@ static EVH acquire(void)
 	EVENT_header->evh_current_process = EVENT_process_offset;
 #else
 	if (++acquire_count == 1) {
-#ifndef SERVER
 		if (mutex_state = ISC_mutex_lock(MUTEX))
 			mutex_bugcheck("mutex lock", mutex_state);
-#endif
 		EVENT_header->evh_current_process = EVENT_process_offset;
 	}
 #endif
@@ -730,16 +697,13 @@ static SLONG create_process(void)
 	EVENT_process_offset = SRQ_REL_PTR(process);
 
 #ifdef MULTI_THREAD
-
 #ifdef SOLARIS_MT
 	ISC_STATUS_ARRAY local_status;
 	ISC_event_init(process->prb_event, 0, EVENT_SIGNAL);
 	EVENT_process = (PRB) ISC_map_object(local_status, &EVENT_data,
 										 EVENT_process_offset,
 										 sizeof(prb));
-#endif
-
-#if !(defined SOLARIS_MT)
+#else
 	ISC_event_init(process->prb_event, EVENT_SIGNAL, 0);
 #endif
 #endif
@@ -750,9 +714,7 @@ static SLONG create_process(void)
 	release();
 
 #ifdef MULTI_THREAD
-	if (gds__thread_start
-		(watcher_thread, NULL,
-		 THREAD_medium, THREAD_blast, 0))
+	if (gds__thread_start(watcher_thread, NULL, THREAD_medium, THREAD_blast, 0))
 		ERR_bugcheck_msg("cannot start thread");
 #endif
 
@@ -890,7 +852,7 @@ static void delete_session(SLONG session_id)
 	SES session = (SES) SRQ_ABS_PTR(session_id);
 
 #ifdef MULTI_THREAD
-	// if session currently delivered events delay its deletion until deliver ends
+	// if session currently delivered events, delay its deletion until deliver ends
 	if (session->ses_flags & SES_delivering)
 	{
 		session->ses_flags |= SES_purge;
@@ -944,8 +906,6 @@ static AST_TYPE deliver(void* arg)
  *	We've been poked -- deliver any satisfying requests.
  *
  **************************************/
-	srq *event_srq, *que2;
-
 #ifdef UNIX
 	if (acquire_count)
 		return;
@@ -955,7 +915,7 @@ static AST_TYPE deliver(void* arg)
 	PRB process = (PRB) SRQ_ABS_PTR(EVENT_process_offset);
 	process->prb_flags &= ~PRB_pending;
 
-	que2 = SRQ_NEXT(process->prb_sessions);
+	srq* que2 = SRQ_NEXT(process->prb_sessions);
 	while (que2 != &process->prb_sessions)
 	{
 		SES session = (SES) ((UCHAR *) que2 - OFFSET(SES, ses_sessions));
@@ -966,6 +926,7 @@ static AST_TYPE deliver(void* arg)
 		const SLONG que2_offset = SRQ_REL_PTR(que2);
 		for (bool flag = true; flag;) {
 			flag = false;
+			srq* event_srq;
 			SRQ_LOOP(session->ses_requests, event_srq) {
 				EVT_REQ request = (EVT_REQ) ((UCHAR *) event_srq - OFFSET(EVT_REQ, req_requests));
 				if (request_completed(request)) {
@@ -975,12 +936,13 @@ static AST_TYPE deliver(void* arg)
 					que2 = (srq *) SRQ_ABS_PTR(que2_offset);
 #ifdef MULTI_THREAD
 					flag = !(session->ses_flags & SES_purge);
+#else
+					flag = true;
 #endif
 					break;
 				}
 			}
 		}
-
 #ifdef MULTI_THREAD
 		session->ses_flags &= ~SES_delivering;
 		if (session->ses_flags & SES_purge)
@@ -1096,13 +1058,11 @@ static void exit_handler(void* arg)
 
 	ISC_STATUS_ARRAY local_status;
 
-#ifndef SERVER
 #ifdef SOLARIS_MT
 	ISC_unmap_object(local_status, &EVENT_data, (UCHAR**)&EVENT_process,
 					 sizeof(prb));
 #endif
 	ISC_unmap_file(local_status, &EVENT_data, 0);
-#endif
 
 	EVENT_header = NULL;
 }
@@ -1120,10 +1080,9 @@ static EVNT find_event(USHORT length, const TEXT* string, EVNT parent)
  *	Lookup an event.
  *
  **************************************/
-	srq* event_srq;
-
 	SRQ_PTR parent_offset = (parent) ? SRQ_REL_PTR(parent) : 0;
 
+	srq* event_srq;
 	SRQ_LOOP(EVENT_header->evh_events, event_srq) {
 		EVNT event = (EVNT) ((UCHAR *) event_srq - OFFSET(EVNT, evnt_events));
 		if (event->evnt_parent == parent_offset &&
@@ -1235,7 +1194,9 @@ static void init(void* arg, SH_MEM shmem_data, bool initialize)
 	int mutex_state;
 
 #if defined(WIN_NT)
-	if (mutex_state = ISC_mutex_init(MUTEX, EVENT_FILE))
+	char buffer[MAXPATHLEN];
+	gds__prefix_lock(buffer, EVENT_FILE);
+	if (mutex_state = ISC_mutex_init(MUTEX, buffer))
 		mutex_bugcheck("mutex init", mutex_state);
 #endif
 
@@ -1249,11 +1210,9 @@ static void init(void* arg, SH_MEM shmem_data, bool initialize)
 	SRQ_INIT(EVENT_header->evh_processes);
 	SRQ_INIT(EVENT_header->evh_events);
 
-#ifndef SERVER
 #if !defined(WIN_NT)
 	if (mutex_state = ISC_mutex_init(MUTEX, shmem_data->sh_mem_mutex_arg))
 		mutex_bugcheck("mutex init", mutex_state);
-#endif
 #endif
 
 	FRB free = (FRB) ((UCHAR*) EVENT_header + sizeof(evh));
@@ -1357,10 +1316,6 @@ static void post_process(PRB process)
 	process->prb_flags |= PRB_pending;
 	release();
 
-#ifdef SERVER
-	deliver();
-#else
-
 #ifdef MULTI_THREAD
 	ISC_event_post(process->prb_event);
 #else
@@ -1372,7 +1327,6 @@ static void post_process(PRB process)
 
 #endif
 	acquire();
-#endif
 }
 
 
@@ -1446,10 +1400,8 @@ static void release(void)
 #else
 	if (!--acquire_count) {
 		EVENT_header->evh_current_process = 0;
-#ifndef SERVER
 		if (mutex_state = ISC_mutex_unlock(MUTEX))
 			mutex_bugcheck("mutex lock", mutex_state);
-#endif
 #ifdef UNIX
 		if (EVENT_process_offset) {
 			PRB process = (PRB) SRQ_ABS_PTR(EVENT_process_offset);
