@@ -54,7 +54,6 @@
 #include "../jrd/cch.h"
 #include "../jrd/ibase.h"
 #include "gen/iberror.h"
-#include "../jrd/all_proto.h"
 #include "../jrd/cch_proto.h"
 #include "../jrd/err_proto.h"
 #include "../jrd/gds_proto.h"
@@ -66,9 +65,14 @@
 #include "../jrd/ods_proto.h"
 #include "../jrd/os/pio_proto.h"
 #include "../jrd/thd.h"
+#include "../common/classes/init.h"
 
 using namespace Jrd;
 
+#ifdef DARWIN
+#define O_DIRECT F_NOCACHE
+#endif
+
 // Some operating systems have problems with use of write/read with 
 // big (>2Gb) files. On the other hand, pwrite/pread works fine for them.
 // Therefore:
@@ -88,15 +92,11 @@ using namespace Jrd;
 #endif
 
 #ifdef SUPERSERVER
-#define THD_IO_MUTEX_INIT(mutx)
-#define THD_IO_MUTEX_LOCK(mutx)		mutx->enter()
-#define THD_IO_MUTEX_UNLOCK(mutx)	mutx->leave()
-#define THD_IO_MUTEX_DESTROY(mutx)
+#define THD_IO_MUTEX_LOCK(mutx)		mutx.enter()
+#define THD_IO_MUTEX_UNLOCK(mutx)	mutx.leave()
 #else
-#define THD_IO_MUTEX_INIT(mutx)
 #define THD_IO_MUTEX_LOCK(mutx)
 #define THD_IO_MUTEX_UNLOCK(mutx)
-#define THD_IO_MUTEX_DESTROY(mutx)
 #endif
 
 #define IO_RETRY	20
@@ -127,16 +127,16 @@ using namespace Jrd;
 #ifdef SUPERSERVER
 #define MASK		0600
 #else
-#define MASK		0660
+#define MASK		0666
 #endif
 
 #define FCNTL_BROKEN
 // please undefine FCNTL_BROKEN for operating systems,
 // that can successfully change BOTH O_DIRECT and O_SYNC using fcntl()
 
-static jrd_file* seek_file(jrd_file*, BufferDesc*, FB_UINT64 *, ISC_STATUS *);
+static jrd_file* seek_file(jrd_file*, BufferDesc*, FB_UINT64*, ISC_STATUS*);
 static jrd_file* setup_file(Database*, const Firebird::PathName&, int);
-static bool unix_error(TEXT*, jrd_file*, ISC_STATUS, ISC_STATUS*);
+static bool unix_error(TEXT*, const jrd_file*, ISC_STATUS, ISC_STATUS*);
 #if defined PREAD_PWRITE && !(defined HAVE_PREAD && defined HAVE_PWRITE)
 static SLONG pread(int, SCHAR *, SLONG, SLONG);
 static SLONG pwrite(int, SCHAR *, SLONG, SLONG);
@@ -145,8 +145,8 @@ static SLONG pwrite(int, SCHAR *, SLONG, SLONG);
 static bool raw_devices_validate_database (int, const Firebird::PathName&);
 static int  raw_devices_unlink_database (const Firebird::PathName&);
 #endif
-static int openFile(const char*, bool, bool);
-static bool maybe_close_file(int&);
+static int	openFile(const char*, bool, bool, bool);
+static void	maybeCloseFile(int&);
 
 
 int PIO_add_file(Database* dbb, jrd_file* main_file, const Firebird::PathName& file_name, SLONG start)
@@ -166,7 +166,7 @@ int PIO_add_file(Database* dbb, jrd_file* main_file, const Firebird::PathName& f
  *	have been locked before entry.
  *
  **************************************/
-	jrd_file* new_file = PIO_create(dbb, file_name, false, false);
+	jrd_file* new_file = PIO_create(dbb, file_name, false, false, false);
 	if (!new_file)
 		return 0;
 
@@ -203,36 +203,12 @@ void PIO_close(jrd_file* main_file)
 		if (file->fil_desc && file->fil_desc != -1) {
 			close(file->fil_desc);
 			file->fil_desc = -1;
-#ifndef PREAD_PWRITE
-			THD_IO_MUTEX_DESTROY(file->fil_mutex);
-#endif
 		}
 	}
 }
 
 
-int PIO_connection(const Firebird::PathName& file_name)
-{
-/**************************************
- *
- *	P I O _ c o n n e c t i o n
- *
- **************************************
- *
- * Functional description
- *	Analyze a file specification and determine whether a page/lock
- *	server is required and available.  If so, return a "connection"
- *	block.  If not, return NULL.
- *
- *	Note: The file name must have been expanded prior to this call.
- *
- **************************************/
-
-	return 0;
-}
-
-
-jrd_file* PIO_create(Database* dbb, const Firebird::PathName& file_name, bool overwrite, bool share_delete)
+jrd_file* PIO_create(Database* dbb, const Firebird::PathName& file_name, bool overwrite, bool temporary, bool /*share_delete*/)
 {
 /**************************************
  *
@@ -261,26 +237,36 @@ jrd_file* PIO_create(Database* dbb, const Firebird::PathName& file_name, bool ov
 #endif
 #endif
 
-	const int desc = open(file_name.c_str(), flag, 0666);
+	const int desc = open(file_name.c_str(), flag, MASK);
 	if (desc == -1) 
 	{
 		ERR_post(isc_io_error,
 				 isc_arg_string, "open O_CREAT",
-				 isc_arg_cstring, file_name.length(), ERR_string(file_name),
-				 isc_arg_gds, isc_io_create_err, isc_arg_unix, errno, isc_arg_end);
+				 isc_arg_string, ERR_string(file_name),
+				 isc_arg_gds, isc_io_create_err, isc_arg_unix, errno, 0);
 	}
-#ifdef HAVE_FCHMOD
-	if (fchmod(desc, MASK) < 0)
-#else
-	if (chmod(file_name.c_str(), MASK) < 0)
+
+	if (temporary
+#ifdef SUPPORT_RAW_DEVICES
+		&& (!PIO_on_raw_device(file_name))
 #endif
+				 )
 	{
-		ERR_post(isc_io_error,
-				 isc_arg_string, "chmod",
-				 isc_arg_cstring, file_name.length(), ERR_string(file_name),
-				 isc_arg_gds, isc_io_create_err, isc_arg_unix, errno, isc_arg_end);
+		int rc = unlink(file_name.c_str());	
+		// it's no use throwing an error if unlink() failed for temp file in release build
+#ifdef DEV_BUILD
+		if (rc < 0)
+		{
+			ERR_post(isc_io_error,
+					 isc_arg_string, "unlink",
+					 isc_arg_string, ERR_string(file_name),
+					 isc_arg_gds, isc_io_create_err, isc_arg_unix, errno, 0);
+		}
+#endif
 	}
 
+	// posix_fadvise(desc, 0, 0, POSIX_FADV_RANDOM);
+
 /* File open succeeded.  Now expand the file name. */
 
 	Firebird::PathName expanded_name(file_name);
@@ -290,7 +276,7 @@ jrd_file* PIO_create(Database* dbb, const Firebird::PathName& file_name, bool ov
 	{
 		file = setup_file(dbb, expanded_name, desc);
 	} 
-	catch(const std::exception&) 
+	catch (const Firebird::Exception&) 
 	{
 		close(desc);
 		throw;
@@ -299,7 +285,7 @@ jrd_file* PIO_create(Database* dbb, const Firebird::PathName& file_name, bool ov
 }
 
 
-int PIO_expand(const TEXT* file_name, USHORT file_length, TEXT* expanded_name, size_t len_expanded)
+bool PIO_expand(const TEXT* file_name, USHORT file_length, TEXT* expanded_name, size_t len_expanded)
 {
 /**************************************
  *
@@ -318,6 +304,23 @@ int PIO_expand(const TEXT* file_name, USHORT file_length, TEXT* expanded_name, s
 }
 
 
+void PIO_extend(jrd_file* /*main_file*/, const ULONG /*extPages*/, const USHORT /*pageSize*/)
+{
+/**************************************
+ *
+ *	P I O _ e x t e n d
+ *
+ **************************************
+ *
+ * Functional description
+ *	Extend file by extPages pages of pageSize size. 
+ *
+ **************************************/
+	// not implemented
+	return;
+}
+
+
 void PIO_flush(jrd_file* main_file)
 {
 /**************************************
@@ -346,65 +349,7 @@ void PIO_flush(jrd_file* main_file)
 }
 
 
-static int openFile(const char* name, bool forcedWrites, bool readOnly)
-{
-/**************************************
- *
- *	o p e n F i l e
- *
- **************************************
- *
- * Functional description
- *	Open a file with appropriate flags.
- *
- **************************************/
-
-	int flag = O_BINARY | (readOnly ? O_RDONLY : O_RDWR);
-#ifdef SUPERSERVER_V2
-	flag |= SYNC;
-	// what to do with O_DIRECT here ?
-#else
-	if (forcedWrites)
-		flag |= SYNC;
-#endif
-
-	for (int i = 0; i < IO_RETRY; i++)
-	{
-		int desc = open(name, flag);
-		if (desc != -1)
-			return desc;
-		if (!SYSCALL_INTERRUPTED(errno))
-			break;
-	}
-
-	return -1;
-}
-
-
-static bool maybe_close_file(int& desc)
-{
-/**************************************
- *
- * M a y b e C l o s e F i l e
- *
- **************************************
- *
- * Functional description
- * If the file is open, close it.
- *
- **************************************/
-
-	if (desc >= 0)
-	{
-		close(desc);
-		desc = -1;
-		return true;
-	}
-	return false;
-}
-
-
-void PIO_force_write(jrd_file* file, bool forcedWrites)
+void PIO_force_write(jrd_file* file, bool forcedWrites, bool notUseFSCache)
 {
 /**************************************
  *
@@ -417,14 +362,17 @@ void PIO_force_write(jrd_file* file, bool forcedWrites)
  *
  **************************************/
 
-// Since all SUPERSERVER_V2 database and shadow I/O is synchronous, this is a no-op. 
+/* Since all SUPERSERVER_V2 database and shadow I/O is synchronous, this
+   is a no-op. */
 
 #ifndef SUPERSERVER_V2
 	const bool oldForce = (file->fil_flags & FIL_force_write) != 0;
+	const bool oldNotUseCache = (file->fil_flags & FIL_no_fs_cache) != 0;
 
-	if (forcedWrites != oldForce)
+	if (forcedWrites != oldForce || notUseFSCache != oldNotUseCache)
 	{
-		const int control = forcedWrites ? SYNC : 0;
+
+		const int control = (forcedWrites ? SYNC : 0) | (notUseFSCache ? O_DIRECT : 0);
 
 #ifndef FCNTL_BROKEN
 		if (fcntl(file->fil_desc, F_SETFL, control) == -1)
@@ -433,26 +381,65 @@ void PIO_force_write(jrd_file* file, bool forcedWrites)
 					 isc_arg_string, "fcntl() SYNC/DIRECT",
 					 isc_arg_cstring, file->fil_length,
 					 ERR_string(file->fil_string, file->fil_length), isc_arg_gds,
-					 isc_io_access_err, isc_arg_unix, errno, isc_arg_end);
+					 isc_io_access_err, isc_arg_unix, errno, 0);
 		}
 #else //FCNTL_BROKEN
-		maybe_close_file(file->fil_desc);
-		file->fil_desc = openFile(file->fil_string, forcedWrites, file->fil_flags & FIL_readonly);
+		maybeCloseFile(file->fil_desc);
+		file->fil_desc = openFile(file->fil_string, forcedWrites, 
+								  notUseFSCache, file->fil_flags & FIL_readonly);
 		if (file->fil_desc == -1)
 		{
 			ERR_post(isc_io_error, isc_arg_string, "re open() for SYNC/DIRECT",
 					 isc_arg_cstring, file->fil_length, ERR_string(file->fil_string, file->fil_length),
-					 isc_arg_gds, isc_io_open_err, isc_arg_unix, errno, isc_arg_end);
+					 isc_arg_gds, isc_io_open_err, isc_arg_unix, errno, 0);
 		}
 #endif //FCNTL_BROKEN
 
-		file->fil_flags &= ~FIL_force_write;
-		file->fil_flags |= (forcedWrites ? FIL_force_write : 0);
+		file->fil_flags &= ~(FIL_force_write | FIL_no_fs_cache);
+		file->fil_flags |= (forcedWrites ? FIL_force_write : 0) | 
+						   (notUseFSCache ? FIL_no_fs_cache : 0);
 	}
 #endif
 }
 
 
+ULONG PIO_get_number_of_pages(const jrd_file* file, const USHORT pagesize)
+{
+/**************************************
+ *
+ *	P I O _ g e t _ n u m b e r _ o f _ p a g e s
+ *
+ **************************************
+ *
+ * Functional description
+ *	Compute number of pages in file, based only on file size.
+ *
+ **************************************/
+
+	if (file->fil_desc == -1) {
+		unix_error("fstat", file, isc_io_access_err, 0);
+		return (0);
+	}
+
+	struct stat statistics;
+	if (fstat(file->fil_desc, &statistics)) {
+		unix_error("fstat", file, isc_io_access_err, 0);
+	}
+
+	const FB_UINT64 length = statistics.st_size;
+
+/****
+#ifndef sun
+length = statistics.st_size;
+#else
+length = statistics.st_blocks * statistics.st_blksize;
+#endif
+****/
+
+	return (length + pagesize - 1) / pagesize;
+}
+
+
 void PIO_header(Database* dbb, SCHAR * address, int length)
 {
 /**************************************
@@ -469,7 +456,8 @@ void PIO_header(Database* dbb, SCHAR * address, int length)
 	int i;
 	FB_UINT64 bytes;
 
-	jrd_file* file = dbb->dbb_file;
+	PageSpace* pageSpace = dbb->dbb_page_manager.findPageSpace(DB_PAGE_SPACE);
+	jrd_file* file = pageSpace->file;
 
 	SignalInhibit siHolder;
 
@@ -490,11 +478,9 @@ void PIO_header(Database* dbb, SCHAR * address, int length)
 			SLONG spare_buffer[MAX_PAGE_SIZE / sizeof(SLONG)];
 
 #ifdef PREAD_PWRITE
-			if ((bytes = pread(file->fil_desc, spare_buffer, length, 0)) ==
-				(FB_UINT64) -1) {
+			if ((bytes = pread(file->fil_desc, spare_buffer, length, 0)) == (FB_UINT64) -1) {
 #else
-			if ((bytes = read(file->fil_desc, spare_buffer, length)) == 
-				(FB_UINT64) -1) {
+			if ((bytes = read(file->fil_desc, spare_buffer, length)) == (FB_UINT64) -1) {
 				THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 				if (SYSCALL_INTERRUPTED(errno))
@@ -541,94 +527,110 @@ void PIO_header(Database* dbb, SCHAR * address, int length)
 #endif
 }
 
+namespace {
+	// we need a class here only to return memory on shutdown and avoid
+	// false memory leak reports
+	static const int ZERO_BUF_SIZE = 1024 * 128;
 
-SLONG PIO_max_alloc(Database* dbb)
+	class HugeStaticBuffer 
+	{
+	public:
+		HugeStaticBuffer(MemoryPool& p)
+			: zeroArray(p), 
+			  zeroBuff(zeroArray.getBuffer(ZERO_BUF_SIZE)) 
+		{
+			memset(zeroBuff, 0, ZERO_BUF_SIZE);
+		}
+
+		const char* get() { return zeroBuff; }
+
+	private:
+		Firebird::Array<char> zeroArray;
+		char* zeroBuff;
+	};
+
+	static Firebird::InitInstance<HugeStaticBuffer> zeros;
+}
+
+
+USHORT PIO_init_data(Database* dbb, jrd_file* main_file, ISC_STATUS* status_vector, 
+					 ULONG startPage, USHORT initPages)
 {
 /**************************************
  *
- *	P I O _ m a x _ a l l o c
+ *	P I O _ i n i t _ d a t a
  *
  **************************************
  *
  * Functional description
- *	Compute last physically allocated page of database.
+ *	Initialize tail of file with zeros
  *
  **************************************/
-	jrd_file* file;
 
-	for (file = dbb->dbb_file; file->fil_next; file = file->fil_next);
+	// Fake buffer, used in seek_file. Page space ID have no matter there
+	// as we already know file to work with
+	BufferDesc bdb;
+	bdb.bdb_dbb = dbb;
+	bdb.bdb_page = PageNumber(0, startPage);
 
-	if (file->fil_desc == -1) {
-		unix_error("fstat", file, isc_io_access_err, 0);
-		return (0);
-	}
+	FB_UINT64 bytes, offset;
 
-	struct stat statistics;
-	if (fstat(file->fil_desc, &statistics)) {
-		unix_error("fstat", file, isc_io_access_err, 0);
-	}
+	ThreadExit teHolder;
+	SignalInhibit siHolder;
 
-	FB_UINT64 length = statistics.st_size;
+	jrd_file* file = seek_file(main_file, &bdb, &offset, status_vector);
 
-/****
-#ifndef sun
-length = statistics.st_size;
+	if (!file)
+		return 0;
+
+	if (file->fil_min_page + 8 > startPage)
+		return 0;
+
+	USHORT leftPages = initPages;
+	const ULONG initBy = MIN(file->fil_max_page - startPage, leftPages);
+	for (ULONG i = startPage; i < startPage + initBy; )
+	{
+		bdb.bdb_page = PageNumber(0, i);
+		USHORT write_pages = ZERO_BUF_SIZE / dbb->dbb_page_size;
+		if (write_pages > leftPages)
+			write_pages = leftPages;
+
+		SLONG to_write = write_pages * dbb->dbb_page_size;
+		SLONG written;
+
+		for (int r = 0; r < IO_RETRY; r++) 
+		{
+			if (!(file = seek_file(file, &bdb, &offset, status_vector)))
+				return false;
+#ifdef PREAD_PWRITE
+			if ((written = pwrite(file->fil_desc, zeros().get(), to_write, LSEEK_OFFSET_CAST offset)) == to_write)
+				break;
 #else
-length = statistics.st_blocks * statistics.st_blksize;
+			if ((written = write(file->fil_desc, zeros().get(), to_write)) == to_write)
+				break;
+			THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
-****/
-
-	return file->fil_min_page - file->fil_fudge +
-		(length + dbb->dbb_page_size - 1) / dbb->dbb_page_size;
-}
-
-
-SLONG PIO_act_alloc(Database* dbb)
-{
-/**************************************
- *
- *	P I O _ a c t  _ a l l o c
- *
- **************************************
- *
- * Functional description
- *  Compute actual number of physically allocated pages of database.
- *
- **************************************/
-	struct stat statistics;
-	ULONG tot_pages = 0;
-
-/**
- **  Traverse the linked list of files and add up the number of pages
- **  in each file
- **/
-	for (jrd_file* file = dbb->dbb_file; file != NULL; file = file->fil_next) {
-		if (file->fil_desc == -1) {
-			unix_error("fstat", file, isc_io_access_err, 0);
-			return (0);
+			if (written == (SLONG) -1 && !SYSCALL_INTERRUPTED(errno))
+				return unix_error("write", file, isc_io_write_err, status_vector);
 		}
 
-		if (fstat(file->fil_desc, &statistics)) {
-			unix_error("fstat", file, isc_io_access_err, 0);
-		}
+#ifndef PREAD_PWRITE
+		THD_IO_MUTEX_UNLOCK(file->fil_mutex);
+#endif
 
-		FB_UINT64 length = statistics.st_size;
-
-		tot_pages += (length + dbb->dbb_page_size - 1) / dbb->dbb_page_size;
+		leftPages -= write_pages;
+		i += write_pages;
 	}
 
-	return tot_pages;
+	return (initPages - leftPages);
 }
 
 
-
-
 jrd_file* PIO_open(Database* dbb,
 			 const Firebird::PathName& string,
 			 bool trace_flag,
-			 blk* connection, 
 			 const Firebird::PathName& file_name,
-			 bool share_delete)
+			 bool /*share_delete*/)
 {
 /**************************************
  *
@@ -637,24 +639,23 @@ jrd_file* PIO_open(Database* dbb,
  **************************************
  *
  * Functional description
- *	Open a database file.  If a "connection" block is provided, use
- *	the connection to communication with a page/lock server.
+ *	Open a database file.
  *
  **************************************/
 	bool readOnly = false;
 	const TEXT* ptr = (string.hasData() ? string : file_name).c_str();
-	int desc = openFile(ptr, false, false);
+	int desc = openFile(ptr, false, false, false);
 
 	if (desc == -1) {
 		/* Try opening the database file in ReadOnly mode. The database file could
 		 * be on a RO medium (CD-ROM etc.). If this fileopen fails, return error.
 		 */
-		desc = openFile(ptr, false, true);
+		desc = openFile(ptr, false, false, true);
 		if (desc == -1) {
 			ERR_post(isc_io_error,
 					 isc_arg_string, "open",
 					 isc_arg_cstring, file_name.length(), ERR_cstring(file_name),
-					 isc_arg_gds, isc_io_open_err, isc_arg_unix, errno, isc_arg_end);
+					 isc_arg_gds, isc_io_open_err, isc_arg_unix, errno, 0);
 		}
 		else {
 			/* If this is the primary file, set Database flag to indicate that it is
@@ -662,12 +663,15 @@ jrd_file* PIO_open(Database* dbb,
 			 * the Header Page flag setting to make sure that the database is set
 			 * ReadOnly.
 			 */
-			if (!dbb->dbb_file)
+			PageSpace* pageSpace = dbb->dbb_page_manager.findPageSpace(DB_PAGE_SPACE);
+			if (!pageSpace->file)
 				dbb->dbb_flags |= DBB_being_opened_read_only;
 			readOnly = true;
 		}
 	}
 
+	// posix_fadvise(desc, 0, 0, POSIX_FADV_RANDOM);
+
 #ifdef SUPPORT_RAW_DEVICES
 	/* At this point the file has successfully been opened in either RW or RO
 	 * mode. Check if it is a special file (i.e. raw block device) and if a
@@ -681,7 +685,7 @@ jrd_file* PIO_open(Database* dbb,
 					isc_arg_cstring, file_name.length(),
 						ERR_cstring (file_name),
 					isc_arg_gds, isc_io_open_err,
-					isc_arg_unix, ENOENT, isc_arg_end);
+					isc_arg_unix, ENOENT, 0);
 	}
 #endif /* SUPPORT_RAW_DEVICES */
 
@@ -690,7 +694,8 @@ jrd_file* PIO_open(Database* dbb,
 		file = setup_file(dbb, string, desc);
 		if (readOnly)
 			file->fil_flags |= FIL_readonly;
-	} catch(const std::exception&) {
+	}
+	catch (const Firebird::Exception&) {
 		close(desc);
 		throw;
 	}
@@ -713,10 +718,12 @@ bool PIO_read(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* statu
 	int i;
 	FB_UINT64 bytes, offset;
 
-	SignalInhibit siHolder;
-
-	if (file->fil_desc == -1)
+	if (file->fil_desc == -1) {
 		return unix_error("read", file, isc_io_read_err, status_vector);
+	}
+
+	ThreadExit teHolder;
+	SignalInhibit siHolder;
 
 	Database* dbb = bdb->bdb_dbb;
 	const FB_UINT64 size = dbb->dbb_page_size;
@@ -742,8 +749,7 @@ bool PIO_read(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* statu
 			THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 			if (bytes == -1U && !SYSCALL_INTERRUPTED(errno))
-				return unix_error("read", file, isc_io_read_err,
-								  status_vector);
+				return unix_error("read", file, isc_io_read_err, status_vector);
 		}
 	}
 	else
@@ -761,8 +767,7 @@ bool PIO_read(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* statu
 			THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 			if (bytes == -1U && !SYSCALL_INTERRUPTED(errno))
-				return unix_error("read", file, isc_io_read_err,
-								  status_vector);
+				return unix_error("read", file, isc_io_read_err, status_vector);
 		}
 	}
 
@@ -786,6 +791,7 @@ bool PIO_read(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* statu
 		}
 	}
 
+	// posix_fadvise(file->desc, offset, size, POSIX_FADV_NOREUSE);
 	return true;
 }
 
@@ -806,11 +812,12 @@ bool PIO_write(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* stat
 	SLONG bytes;
     FB_UINT64 offset;
 
-	SignalInhibit siHolder;
-
 	if (file->fil_desc == -1)
 		return unix_error("write", file, isc_io_write_err, status_vector);
 
+	ThreadExit teHolder;
+	SignalInhibit siHolder;
+
 	Database* dbb = bdb->bdb_dbb;
 	const SLONG size = dbb->dbb_page_size;
 
@@ -833,8 +840,7 @@ bool PIO_write(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* stat
 			THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 			if (bytes == -1U && !SYSCALL_INTERRUPTED(errno))
-				return unix_error("write", file, isc_io_write_err,
-								  status_vector);
+				return unix_error("write", file, isc_io_write_err, status_vector);
 		}
 	}
 	else
@@ -852,8 +858,7 @@ bool PIO_write(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* stat
 			THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 			if (bytes == (SLONG) -1 && !SYSCALL_INTERRUPTED(errno))
-				return unix_error("write", file, isc_io_write_err,
-								  status_vector);
+				return unix_error("write", file, isc_io_write_err, status_vector);
 		}
 	}
 
@@ -861,6 +866,7 @@ bool PIO_write(jrd_file* file, BufferDesc* bdb, Ods::pag* page, ISC_STATUS* stat
 	THD_IO_MUTEX_UNLOCK(file->fil_mutex);
 #endif
 
+	// posix_fadvise(file->desc, offset, size, POSIX_FADV_DONTNEED);
 	return true;
 }
 
@@ -880,7 +886,7 @@ static jrd_file* seek_file(jrd_file* file, BufferDesc* bdb, FB_UINT64* offset,
  *
  **************************************/
 	Database* dbb = bdb->bdb_dbb;
-	ULONG page = bdb->bdb_page;
+	ULONG page = bdb->bdb_page.getPageNum();
 
 	for (;; file = file->fil_next)
 	{
@@ -925,6 +931,64 @@ static jrd_file* seek_file(jrd_file* file, BufferDesc* bdb, FB_UINT64* offset,
 }
 
 
+static int openFile(const char* name, bool forcedWrites, bool notUseFSCache, bool readOnly)
+{
+/**************************************
+ *
+ *	o p e n F i l e
+ *
+ **************************************
+ *
+ * Functional description
+ *	Open a file with appropriate flags.
+ *
+ **************************************/
+
+	int flag = O_BINARY | (readOnly ? O_RDONLY : O_RDWR);
+#ifdef SUPERSERVER_V2
+	flag |= SYNC;
+	// what to do with O_DIRECT here ?
+#else
+	if (forcedWrites)
+		flag |= SYNC;
+	if (notUseFSCache)
+		flag |= O_DIRECT;
+#endif
+
+	for (int i = 0; i < IO_RETRY; i++)
+	{
+		int desc = open(name, flag);
+		if (desc != -1)
+			return desc;
+		if (!SYSCALL_INTERRUPTED(errno))
+			break;
+	}
+
+	return -1;
+}
+
+
+static void maybeCloseFile(int& desc)
+{
+/**************************************
+ *
+ *	m a y b e C l o s e F i l e
+ *
+ **************************************
+ *
+ * Functional description
+ *	If the file is open, close it.
+ *
+ **************************************/
+
+	if (desc >= 0)
+	{
+		close(desc);
+		desc = -1;
+	}
+}
+
+
 static jrd_file* setup_file(Database* dbb, const Firebird::PathName& file_name, int desc)
 {
 /**************************************
@@ -946,13 +1010,11 @@ static jrd_file* setup_file(Database* dbb, const Firebird::PathName& file_name,
 	file->fil_max_page = -1UL;
 	MOVE_FAST(file_name.c_str(), file->fil_string, file_name.length());
 	file->fil_string[file_name.length()] = '\0';
-#ifndef PREAD_PWRITE
-	THD_IO_MUTEX_INIT(file->fil_mutex);
-#endif
 
 /* If this isn't the primary file, we're done */
 
-	if (dbb->dbb_file)
+	PageSpace* pageSpace = dbb->dbb_page_manager.findPageSpace(DB_PAGE_SPACE);
+	if (pageSpace && pageSpace->file)
 		return file;
 
 /* Build unique lock string for file and construct lock block */
@@ -997,8 +1059,9 @@ static jrd_file* setup_file(Database* dbb, const Firebird::PathName& file_name,
 			SCHAR spare_memory[MIN_PAGE_SIZE * 2];
 			SCHAR *header_page_buffer = (SCHAR*) FB_ALIGN((IPTR)spare_memory, MIN_PAGE_SIZE);
 		
+			PageSpace* pageSpace = dbb->dbb_page_manager.findPageSpace(DB_PAGE_SPACE);
 			try {
-				dbb->dbb_file = file;
+				pageSpace->file = file;
 				PIO_header(dbb, header_page_buffer, MIN_PAGE_SIZE);
 				/* Rewind file pointer */
 				if (lseek (file->fil_desc, LSEEK_OFFSET_CAST 0, 0) == (off_t)-1)
@@ -1006,15 +1069,16 @@ static jrd_file* setup_file(Database* dbb, const Firebird::PathName& file_name,
 						isc_arg_string, "lseek",
 						isc_arg_cstring, file_name.length(), ERR_cstring (file_name),
 						isc_arg_gds, isc_io_read_err,
-						isc_arg_unix, errno, isc_arg_end);
+						isc_arg_unix, errno, 0);
 				if ((reinterpret_cast<Ods::header_page*>(header_page_buffer)->hdr_flags & Ods::hdr_shutdown_mask) == Ods::hdr_shutdown_single)
-					ERR_post(isc_shutdown, isc_arg_cstring, file_name.length(), ERR_cstring(file_name), isc_arg_end);
-				dbb->dbb_file = NULL; // Will be set again later by the caller				
-			} catch(const std::exception&) {
+					ERR_post(isc_shutdown, isc_arg_cstring, file_name.length(), ERR_cstring(file_name), 0);
+				pageSpace->file = NULL; // Will be set again later by the caller				
+			}
+			catch (const Firebird::Exception&) {
 				delete dbb->dbb_lock;
 				dbb->dbb_lock = NULL;
 				delete file;
-				dbb->dbb_file = NULL; // Will be set again later by the caller
+				pageSpace->file = NULL; // Will be set again later by the caller
 				throw;
 			}
 		}
@@ -1026,7 +1090,7 @@ static jrd_file* setup_file(Database* dbb, const Firebird::PathName& file_name,
 
 static bool unix_error(
 						  TEXT* string,
-						  jrd_file* file, ISC_STATUS operation,
+						  const jrd_file* file, ISC_STATUS operation,
 						  ISC_STATUS* status_vector)
 {
 /**************************************
@@ -1062,7 +1126,7 @@ static bool unix_error(
 				 isc_arg_string, ERR_string(file->fil_string,
 											file->fil_length),
 				 isc_arg_gds,
-				 operation, isc_arg_unix, errno, isc_arg_end);
+				 operation, isc_arg_unix, errno, 0);
 
 
     // Added a false for final return - which seems to be the answer,
@@ -1177,7 +1241,7 @@ int PIO_unlink (const Firebird::PathName& file_name)
 }
 
 
-bool PIO_on_raw_device(
+bool PIO_on_raw_device (
 	const Firebird::PathName& file_name)
 {
 /**************************************
@@ -1221,7 +1285,7 @@ static bool raw_devices_validate_database (
 					isc_arg_string, "raw_devices_validate_database",
 					isc_arg_string, file_name.length(), ERR_cstring (file_name),
 					isc_arg_gds, isc_io_read_err,
-					isc_arg_unix, errno, isc_arg_end);
+					isc_arg_unix, errno, 0);
 
 	for (int i = 0; i < IO_RETRY; i++)
 	{
@@ -1230,7 +1294,7 @@ static bool raw_devices_validate_database (
 						isc_arg_string, "lseek",
 						isc_arg_string, file_name.length(), ERR_cstring (file_name),
 						isc_arg_gds, isc_io_read_err,
-						isc_arg_unix, errno, isc_arg_end);
+						isc_arg_unix, errno, 0);
 		const ssize_t bytes = read (desc, header, sizeof(header));
 		if (bytes == sizeof(header))
 			goto read_finished;
@@ -1239,14 +1303,14 @@ static bool raw_devices_validate_database (
 						isc_arg_string, "read",
 						isc_arg_string, file_name.length(), ERR_cstring (file_name),
 						isc_arg_gds, isc_io_read_err,
-						isc_arg_unix, errno, isc_arg_end);
+						isc_arg_unix, errno, 0);
 	}
 
 	ERR_post (isc_io_error,
 				isc_arg_string, "read_retry",
 				isc_arg_string, file_name.length(), ERR_cstring (file_name),
 				isc_arg_gds, isc_io_read_err,
-				isc_arg_unix, errno, isc_arg_end);
+				isc_arg_unix, errno, 0);
 
   read_finished:
 	/* Rewind file pointer */
@@ -1255,7 +1319,7 @@ static bool raw_devices_validate_database (
 					isc_arg_string, "lseek",
 					isc_arg_string, file_name.length(), ERR_cstring (file_name),
 					isc_arg_gds, isc_io_read_err,
-					isc_arg_unix, errno, isc_arg_end);
+					isc_arg_unix, errno, 0);
 
 	/* Validate database header. Code lifted from PAG_header. */
 	if (hp->hdr_header.pag_type != pag_header /*|| hp->hdr_sequence*/)
@@ -1298,7 +1362,7 @@ static int raw_devices_unlink_database (
 						isc_arg_string, "open",
 						isc_arg_string, file_name.length(), ERR_cstring (file_name),
 						isc_arg_gds, isc_io_open_err,
-						isc_arg_unix, errno, isc_arg_end);
+						isc_arg_unix, errno, 0);
 	}
 
 	memset(header, 0xa5, sizeof(header));
@@ -1314,7 +1378,7 @@ static int raw_devices_unlink_database (
 			isc_arg_string, "write",
 			isc_arg_string, file_name.length(), ERR_cstring (file_name),
 			isc_arg_gds, isc_io_write_err,
-			isc_arg_unix, errno, isc_arg_end);
+			isc_arg_unix, errno, 0);
 	}
 
 	//if (desc != -1) perhaps it's better to check this???
@@ -1328,4 +1392,3 @@ static int raw_devices_unlink_database (
 	return 0;
 }
 #endif // SUPPORT_RAW_DEVICES
-
