@@ -33,6 +33,7 @@
  * 2002.10.29 Sean Leyne - Removed obsolete "Netware" port
  * 2003.10.05 Dmitry Yemanov: Added support for explicit cursors in PSQL
  * 2004.01.16 Vlad Horsun: Added support for default parameters 
+ * Adriano dos Santos Fernandes
  */
 
 #include "firebird.h"
@@ -42,6 +43,7 @@
 #include <stdarg.h>
 #include "../jrd/jrd.h"
 #include "../jrd/ibase.h"
+#include "../jrd/ini.h"
 #include "../jrd/val.h"
 #include "../jrd/align.h"
 #include "../jrd/exe.h"
@@ -54,7 +56,6 @@
 #include "../jrd/blb.h"
 #include "../jrd/intl.h"
 #include "../jrd/met.h"
-#include "../jrd/all_proto.h"
 #include "../jrd/cmp_proto.h"
 #include "../jrd/cvt_proto.h"
 #include "../jrd/err_proto.h"
@@ -64,6 +65,7 @@
 #include "../jrd/par_proto.h"
 #include "../jrd/thd.h"
 #include "../common/utils_proto.h"
+#include "../jrd/SysFunction.h"
 
 
 /* blr type classes */
@@ -103,8 +105,9 @@ static jrd_nod* par_function(thread_db*, CompilerScratch*);
 static jrd_nod* par_literal(thread_db*, CompilerScratch*);
 static jrd_nod* par_map(thread_db*, CompilerScratch*, USHORT);
 static jrd_nod* par_message(thread_db*, CompilerScratch*);
-static jrd_nod* par_modify(thread_db*, CompilerScratch*);
+static jrd_nod* par_modify(thread_db*, CompilerScratch*, SSHORT);
 static USHORT par_name(CompilerScratch*, Firebird::MetaName&);
+static size_t par_name(CompilerScratch* csb, Firebird::string& name);
 static jrd_nod* par_plan(thread_db*, CompilerScratch*);
 static jrd_nod* par_procedure(thread_db*, CompilerScratch*, SSHORT);
 static void par_procedure_parms(thread_db*, CompilerScratch*, jrd_prc*, jrd_nod**,
@@ -112,8 +115,11 @@ static void par_procedure_parms(thread_db*, CompilerScratch*, jrd_prc*, jrd_nod*
 static jrd_nod* par_relation(thread_db*, CompilerScratch*, SSHORT, bool);
 static jrd_nod* par_rse(thread_db*, CompilerScratch*, SSHORT);
 static jrd_nod* par_sort(thread_db*, CompilerScratch*, bool);
+#ifdef NOT_USED_OR_REPLACED
 static jrd_nod* par_stream(thread_db*, CompilerScratch*);
-static jrd_nod* par_union(thread_db*, CompilerScratch*);
+#endif
+static jrd_nod* par_sys_function(thread_db*, CompilerScratch*);
+static jrd_nod* par_union(thread_db*, CompilerScratch*, bool);
 static USHORT par_word(CompilerScratch*);
 static jrd_nod* parse(thread_db*, CompilerScratch*, USHORT, USHORT expected_optional = 0);
 static void syntax_error(CompilerScratch*, const TEXT*);
@@ -208,7 +214,7 @@ jrd_nod* PAR_blr(thread_db*	tdbb,
 		error(csb, isc_metadata_corrupt,
 			  isc_arg_gds, isc_wroblrver,
 			  isc_arg_number, (SLONG) blr_version4,
-			  isc_arg_number, (SLONG) version, isc_arg_end);
+			  isc_arg_number, (SLONG) version, 0);
 	}
 
 	if (version == blr_version4)
@@ -232,7 +238,7 @@ jrd_nod* PAR_blr(thread_db*	tdbb,
 }
 
 
-USHORT PAR_desc(CompilerScratch* csb, DSC* desc)
+USHORT PAR_desc(thread_db* tdbb, CompilerScratch* csb, DSC* desc, ItemInfo* itemInfo)
 {
 /**************************************
  *
@@ -245,13 +251,28 @@ USHORT PAR_desc(CompilerScratch* csb, DSC* desc)
  *	of the datatype.
  *
  **************************************/
+	if (itemInfo)
+	{
+		itemInfo->nullable = true;
+		itemInfo->explicitCollation = false;
+		itemInfo->fullDomain = false;
+	}
+
 	desc->dsc_scale = 0;
 	desc->dsc_sub_type = 0;
 	desc->dsc_address = NULL;
 	desc->dsc_flags = 0;
 
 	const USHORT dtype = BLR_BYTE;
-	switch (dtype) {
+
+	switch (dtype)
+	{
+	case blr_not_nullable:
+		PAR_desc(tdbb, csb, desc, itemInfo);
+		if (itemInfo)
+			itemInfo->nullable = false;
+		break;
+
 	case blr_text:
 		desc->dsc_dtype = dtype_text;
 		desc->dsc_flags |= DSC_no_subtype;
@@ -362,8 +383,76 @@ USHORT PAR_desc(CompilerScratch* csb, DSC* desc)
 			break;
 		}
 
+	case blr_domain_name:
+	case blr_domain_name2:
+		{
+			bool fullDomain = (BLR_BYTE == blr_domain_full);
+			Firebird::MetaName* name = FB_NEW(csb->csb_pool) Firebird::MetaName(csb->csb_pool);
+			par_name(csb, *name);
+
+			FieldInfo fieldInfo;
+			bool exist = csb->csb_map_field_info.get(*name, fieldInfo);
+			MET_get_domain(tdbb, *name, desc, (exist ? NULL : &fieldInfo));
+
+			if (!exist)
+				csb->csb_map_field_info.put(*name, fieldInfo);
+
+			if (itemInfo)
+			{
+				itemInfo->field = *name;
+
+				if (fullDomain)
+				{
+					itemInfo->nullable = fieldInfo.nullable;
+					itemInfo->fullDomain = true;
+				}
+				else
+					itemInfo->nullable = true;
+			}
+
+			if (dtype == blr_domain_name2)
+			{
+				USHORT ttype = BLR_WORD;
+
+				switch (desc->dsc_dtype)
+				{
+					case dtype_cstring:
+					case dtype_text:
+					case dtype_varying:
+						INTL_ASSIGN_TTYPE(desc, ttype);
+						break;
+
+					case dtype_blob:
+						desc->dsc_scale = ttype & 0xFF;		// BLOB character set
+						desc->dsc_flags = ttype & 0xFF00;	// BLOB collation
+						break;
+
+					default:
+						error(csb, isc_collation_requires_text, 0);
+						break;
+				}
+			}
+
+			jrd_nod* dep_node = PAR_make_node(tdbb, e_dep_length);
+			dep_node->nod_type = nod_dependency;
+			dep_node->nod_arg[e_dep_object] = (jrd_nod*) name;
+			dep_node->nod_arg[e_dep_object_type] = (jrd_nod*)(IPTR) obj_field;
+			csb->csb_dependencies.push(dep_node);
+
+			break;
+		}
+
 	default:
-		error(csb, isc_datnotsup, isc_arg_end);
+		error(csb, isc_datnotsup, 0);
+	}
+
+	if (itemInfo)
+	{
+		if (dtype == blr_cstring2 || dtype == blr_text2 || dtype == blr_varying2 ||
+			dtype == blr_blob2 || dtype == blr_domain_name2)
+		{
+			itemInfo->explicitCollation = true;
+		}
 	}
 
 	return type_alignments[desc->dsc_dtype];
@@ -463,7 +552,7 @@ jrd_nod* PAR_make_field(thread_db* tdbb, CompilerScratch* csb,
 	if (relation)
 	{
 		if (!relation->rel_fields) {
-			ERR_post(isc_depend_on_uncommitted_rel, isc_arg_end);
+			ERR_post(isc_depend_on_uncommitted_rel, 0);
 		}
 		field = (*relation->rel_fields)[id];
 	}
@@ -543,7 +632,8 @@ jrd_nod* PAR_make_node(thread_db* tdbb, int size)
 }
 
 
-CompilerScratch* PAR_parse(thread_db* tdbb, const UCHAR* blr, USHORT internal_flag)
+CompilerScratch* PAR_parse(thread_db* tdbb, const UCHAR* blr, USHORT internal_flag,
+						   USHORT dbginfo_length, const UCHAR* dbginfo)
 {
 /**************************************
  *
@@ -567,7 +657,7 @@ CompilerScratch* PAR_parse(thread_db* tdbb, const UCHAR* blr, USHORT internal_fl
 	{
 		error(csb, isc_wroblrver,
 			  isc_arg_number, (SLONG) blr_version4,
-			  isc_arg_number, (SLONG) version, isc_arg_end);
+			  isc_arg_number, (SLONG) version, 0);
 	}
 
 	if (version == blr_version4)
@@ -575,6 +665,9 @@ CompilerScratch* PAR_parse(thread_db* tdbb, const UCHAR* blr, USHORT internal_fl
 		csb->csb_g_flags |= csb_blr_version4;
 	}
 
+	if (dbginfo_length > 0)
+		DBG_parse_debug_info(dbginfo_length, dbginfo, csb->csb_dbg_info);
+
 	jrd_nod* node = parse(tdbb, csb, OTHER);
 	csb->csb_node = node;
 
@@ -703,9 +796,6 @@ static SSHORT find_proc_field(const jrd_prc* procedure, const Firebird::MetaName
  *
  **************************************/
 	vec<Parameter*>* list = procedure->prc_output_fields;
-	if (!list)
-		return -1;
-
 	vec<Parameter*>::const_iterator ptr = list->begin();
 	for (const vec<Parameter*>::const_iterator end = list->end(); ptr < end; ++ptr)
 	{
@@ -768,11 +858,27 @@ static jrd_nod* par_cast(thread_db* tdbb, CompilerScratch* csb)
 	node->nod_arg[e_cast_fmt] = (jrd_nod*) format;
 
 	dsc* desc = &format->fmt_desc[0];
-	PAR_desc(csb, desc);
+	ItemInfo itemInfo;
+	PAR_desc(tdbb, csb, desc, &itemInfo);
 	format->fmt_length = desc->dsc_length;
 
 	node->nod_arg[e_cast_source] = parse(tdbb, csb, VALUE);
 
+	if (itemInfo.isSpecial())
+	{
+		ItemInfo* p = FB_NEW(*tdbb->getDefaultPool()) ItemInfo(*tdbb->getDefaultPool(), itemInfo);
+		node->nod_arg[e_cast_iteminfo] = (jrd_nod*) p;
+	}
+
+	if (itemInfo.explicitCollation)
+	{
+		jrd_nod* dep_node = PAR_make_node (tdbb, e_dep_length);
+		dep_node->nod_type = nod_dependency;
+		dep_node->nod_arg [e_dep_object] = (jrd_nod*)(IPTR) INTL_TEXT_TYPE(*desc);
+		dep_node->nod_arg [e_dep_object_type] = (jrd_nod*)(IPTR) obj_collation;
+		csb->csb_dependencies.push(dep_node);
+	}
+
 	return node;
 }
 
@@ -808,35 +914,34 @@ static PsqlException* par_condition(thread_db* tdbb, CompilerScratch* csb)
 
 	PsqlException* exception_list = FB_NEW_RPT(*tdbb->getDefaultPool(), 1) PsqlException();
 	exception_list->xcp_count = 1;
+	xcp_repeat& item = exception_list->xcp_rpt[0];
 	
 	switch (code_type) {
 	case blr_sql_code:
-		exception_list->xcp_rpt[0].xcp_type = xcp_sql_code;
-		exception_list->xcp_rpt[0].xcp_code = (SSHORT) BLR_WORD;
+		item.xcp_type = xcp_sql_code;
+		item.xcp_code = (SSHORT) BLR_WORD;
 		break;
 
 	case blr_gds_code:
-		exception_list->xcp_rpt[0].xcp_type = xcp_gds_code;
+		item.xcp_type = xcp_gds_code;
 		par_name(csb, name);
 		name.lower7();
 		code_number = PAR_symbol_to_gdscode(name);
 		if (code_number)
-			exception_list->xcp_rpt[0].xcp_code = code_number;
+			item.xcp_code = code_number;
 		else
-			error(csb, isc_codnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_codnotdef, isc_arg_string, ERR_cstring(name), 0);
 		break;
 
 	case blr_exception:
 	case blr_exception_msg:
-		exception_list->xcp_rpt[0].xcp_type = xcp_xcp_code;
+		item.xcp_type = xcp_xcp_code;
 		par_name(csb, name);
-		if (!(exception_list->xcp_rpt[0].xcp_code =
-			  MET_lookup_exception_number(tdbb, name)))
-			error(csb, isc_xcpnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+		if (!(item.xcp_code = MET_lookup_exception_number(tdbb, name)))
+			error(csb, isc_xcpnotdef, isc_arg_string, ERR_cstring(name), 0);
 		dep_node = PAR_make_node(tdbb, e_dep_length);
 		dep_node->nod_type = nod_dependency;
-		dep_node->nod_arg[e_dep_object] =
-			(jrd_nod*)(IPTR) exception_list->xcp_rpt[0].xcp_code;
+		dep_node->nod_arg[e_dep_object] = (jrd_nod*)(IPTR) item.xcp_code;
 		dep_node->nod_arg[e_dep_object_type] = (jrd_nod*)(IPTR) obj_exception;
 		csb->csb_dependencies.push(dep_node);
 		break;
@@ -875,42 +980,40 @@ static PsqlException* par_conditions(thread_db* tdbb, CompilerScratch* csb)
 	exception_list->xcp_count = n;
 	for (int i = 0; i < n; i++) {
 		const USHORT code_type = BLR_BYTE;
+		xcp_repeat& item = exception_list->xcp_rpt[i];
+		
 		switch (code_type) {
 		case blr_sql_code:
-			exception_list->xcp_rpt[i].xcp_type = xcp_sql_code;
-			exception_list->xcp_rpt[i].xcp_code = (SSHORT) BLR_WORD;
+			item.xcp_type = xcp_sql_code;
+			item.xcp_code = (SSHORT) BLR_WORD;
 			break;
 
 		case blr_gds_code:
-			exception_list->xcp_rpt[i].xcp_type = xcp_gds_code;
+			item.xcp_type = xcp_gds_code;
 			par_name(csb, name);
 			name.lower7();
 			code_number = PAR_symbol_to_gdscode(name);
 			if (code_number)
-				exception_list->xcp_rpt[i].xcp_code = code_number;
+				item.xcp_code = code_number;
 			else
-				error(csb, isc_codnotdef,
-					  isc_arg_string, ERR_cstring(name), isc_arg_end);
+				error(csb, isc_codnotdef, isc_arg_string, ERR_cstring(name), 0);
 			break;
 
 		case blr_exception:
-			exception_list->xcp_rpt[i].xcp_type = xcp_xcp_code;
+			item.xcp_type = xcp_xcp_code;
 			par_name(csb, name);
-			if (!(exception_list->xcp_rpt[i].xcp_code =
-				  MET_lookup_exception_number(tdbb, name)))
-				error(csb, isc_xcpnotdef,
-					  isc_arg_string, ERR_cstring(name), isc_arg_end);
+			if (!(item.xcp_code = MET_lookup_exception_number(tdbb, name)))
+				error(csb, isc_xcpnotdef, isc_arg_string, ERR_cstring(name), 0);
 			dep_node = PAR_make_node(tdbb, e_dep_length);
 			dep_node->nod_type = nod_dependency;
-			dep_node->nod_arg[e_dep_object] =
-				(jrd_nod*) (IPTR)exception_list->xcp_rpt[0].xcp_code;
+			dep_node->nod_arg[e_dep_object] = (jrd_nod*)(IPTR) item.xcp_code;
 			dep_node->nod_arg[e_dep_object_type] = (jrd_nod*)(IPTR) obj_exception;
 			csb->csb_dependencies.push(dep_node);
 			break;
 
 		case blr_default_code:
-			exception_list->xcp_rpt[i].xcp_type = xcp_default;
-			exception_list->xcp_rpt[i].xcp_code = 0;
+			item.xcp_type = xcp_default;
+			item.xcp_code = 0;
 			break;
 
 		default:
@@ -938,26 +1041,32 @@ static SSHORT par_context(CompilerScratch* csb, SSHORT* context_ptr)
  *
  **************************************/
 
+	const SSHORT context = (unsigned int) BLR_BYTE;
+
+	if (context_ptr)
+		*context_ptr = context;
+
+	CompilerScratch::csb_repeat* tail = CMP_csb_element(csb, context);
+
+	if (tail->csb_flags & csb_used) {
+		if (csb->csb_g_flags & csb_reuse_context) {
+			return tail->csb_stream;
+		}
+		else {
+			error(csb, isc_ctxinuse, 0);
+		}
+	}
+
 	const SSHORT stream = csb->nextStream(false);
 	if (stream >= MAX_STREAMS)
 	{
-		error(csb, isc_too_many_contexts, isc_arg_end);
-	}
-	const SSHORT context = (unsigned int) BLR_BYTE;
-	CMP_csb_element(csb, stream);
-	CompilerScratch::csb_repeat* tail = CMP_csb_element(csb, context);
-
-	if ((tail->csb_flags & csb_used) &&
-		!(csb->csb_g_flags & csb_no_context_check))
-	{
-		error(csb, isc_ctxinuse, isc_arg_end);
+		error(csb, isc_too_many_contexts, 0);
 	}
 
 	tail->csb_flags |= csb_used;
 	tail->csb_stream = (UCHAR) stream;
 
-	if (context_ptr)
-		*context_ptr = context;
+	CMP_csb_element(csb, stream);
 
 	return stream;
 }
@@ -1046,7 +1155,7 @@ static jrd_nod* par_exec_proc(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_
 			procedure = MET_lookup_procedure(tdbb, name, false);
 		}
 		if (!procedure)
-			error(csb, isc_prcnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_prcnotdef, isc_arg_string, ERR_cstring(name), 0);
 	}
 
 	jrd_nod* node = PAR_make_node(tdbb, e_esp_length);
@@ -1130,7 +1239,31 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
  **************************************/
 	SET_TDBB(tdbb);
 
-	const SSHORT context = (unsigned int) BLR_BYTE;
+	const USHORT context = (unsigned int) BLR_BYTE;
+
+	// check if this is a VALUE of domain's check constraint
+	if (!csb->csb_domain_validation.isEmpty() &&
+		(blr_operator == blr_fid || blr_operator == blr_field) &&
+		context == 0)
+	{
+		if (blr_operator == blr_fid) {
+			SSHORT id = BLR_WORD;
+			fb_assert(id == 0);
+		}
+		else {
+			Firebird::MetaName name;
+			par_name(csb, name);
+		}
+
+		jrd_nod* node = PAR_make_node(tdbb, e_domval_length);
+		node->nod_type = nod_domain_validation;
+		node->nod_count = 0;
+
+		dsc* desc = (dsc*) (node->nod_arg + e_domval_desc);
+		MET_get_domain(tdbb, csb->csb_domain_validation, desc, NULL);
+
+		return node;
+	}
 
 	if (context >= csb->csb_rpt.getCount())/* ||
 		!(csb->csb_rpt[context].csb_flags & csb_used) )
@@ -1141,7 +1274,7 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
 				but we must support legacy BLR.
 		*/
 	{
-		error(csb, isc_ctxnotdef, isc_arg_end);
+		error(csb, isc_ctxnotdef, 0);
 	}
 
 	Firebird::MetaName name;
@@ -1149,12 +1282,15 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
 	const SSHORT stream = csb->csb_rpt[context].csb_stream;
 	SSHORT flags = 0;
 	bool is_column = false;
-	if (blr_operator == blr_fid) {
+
+	if (blr_operator == blr_fid)
+	{
 		id = BLR_WORD;
 		flags = nod_id;
 		is_column = true;
 	}
-	else if (blr_operator == blr_field) {
+	else if (blr_operator == blr_field)
+	{
 		CompilerScratch::csb_repeat* tail = &csb->csb_rpt[stream];
 		const jrd_prc* procedure = tail->csb_procedure;
 
@@ -1173,14 +1309,14 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
 			par_name(csb, name);
 			if ((id = find_proc_field(procedure, name)) == -1)
 				error(csb,
-					  isc_fldnotdef,
+					  isc_fldnotdef2,
 					  isc_arg_string, ERR_cstring(name),
-					  isc_arg_string, procedure->prc_name.c_str(), isc_arg_end);
+					  isc_arg_string, procedure->prc_name.c_str(), 0);
 		}
 		else {
 			jrd_rel* relation = tail->csb_relation;
 			if (!relation)
-				error(csb, isc_ctxnotdef, isc_arg_end);
+				error(csb, isc_ctxnotdef, 0);
 
 			/* make sure relation has been scanned before using it */
 
@@ -1198,16 +1334,20 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
 					is_column = true;
 				}
 				else {
-					if (tdbb->tdbb_attachment->att_flags & ATT_gbak_attachment)
+					if (tdbb->getAttachment()->att_flags & ATT_gbak_attachment)
+					{
 						warning(csb, isc_fldnotdef, isc_arg_string,
 								ERR_cstring(name), isc_arg_string,
-								relation->rel_name.c_str(), isc_arg_end);
+								relation->rel_name.c_str(), 0);
+					}
 					else if (relation->rel_name.length() > 0)
+					{
 						error(csb, isc_fldnotdef, isc_arg_string,
 							  ERR_cstring(name), isc_arg_string,
-							  relation->rel_name.c_str(), isc_arg_end);
+							  relation->rel_name.c_str(), 0);
+					}
 					else
-						error(csb, isc_ctxnotdef, isc_arg_end);
+						error(csb, isc_ctxnotdef, 0);
 				}
 			}
 		}
@@ -1227,17 +1367,29 @@ static jrd_nod* par_field(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_oper
 	jrd_nod* node = PAR_gen_field(tdbb, stream, id);
 	node->nod_flags |= flags;
 
-	if (is_column) {
-		jrd_rel* temp_rel = csb->csb_rpt[stream].csb_relation;
-		if (temp_rel) {
-			jrd_fld* field = (*temp_rel->rel_fields)[id];
-			if (field) {
+	if (is_column)
+	{
+ 		jrd_rel* temp_rel = csb->csb_rpt[stream].csb_relation;
+
+		if (temp_rel)
+		{
+			jrd_fld* field;
+
+			if (id < (int) temp_rel->rel_fields->count() && (field = (*temp_rel->rel_fields)[id]))
+ 			{
 				if (field->fld_default_value && field->fld_not_null)
-					node->nod_arg[e_fld_default_value] =
-						field->fld_default_value;
-			}
-		}
-	}
+					node->nod_arg[e_fld_default_value] = field->fld_default_value;
+ 			}
+ 			else
+ 			{
+				if (temp_rel->rel_flags & REL_system)
+				{
+					node = PAR_make_node(tdbb, 0);
+					node->nod_type = nod_null;
+ 				}
+ 			}
+ 		}
+ 	}
 
 	return node;
 }
@@ -1261,7 +1413,7 @@ static jrd_nod* par_function(thread_db* tdbb, CompilerScratch* csb)
 	const USHORT count = par_name(csb, name);
 
 	UserFunction* function = FUN_lookup_function(name,
-					!(tdbb->tdbb_attachment->att_flags & ATT_gbak_attachment));
+					!(tdbb->getAttachment()->att_flags & ATT_gbak_attachment));
 	if (!function) {
 		if (tdbb->tdbb_flags & TDBB_prc_being_dropped) {
 			jrd_nod* anode = PAR_make_node(tdbb, e_fun_length);
@@ -1272,7 +1424,7 @@ static jrd_nod* par_function(thread_db* tdbb, CompilerScratch* csb)
 		}
 		else {
 			csb->csb_running -= count;
-			error(csb, isc_funnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_funnotdef, isc_arg_string, ERR_cstring(name), 0);
 		}
 	}
 
@@ -1283,17 +1435,19 @@ static jrd_nod* par_function(thread_db* tdbb, CompilerScratch* csb)
 	}
 
 	if (!homonyms)
-		if (tdbb->tdbb_attachment->att_flags & ATT_gbak_attachment)
+		if (tdbb->getAttachment()->att_flags & ATT_gbak_attachment)
+		{
 			warning(csb, isc_funnotdef,
 					isc_arg_string, ERR_cstring(name),
 					isc_arg_interpreted,
-					"module name or entrypoint could not be found", isc_arg_end);
+					"module name or entrypoint could not be found", 0);
+		}
 		else {
 			csb->csb_running -= count;
 			error(csb, isc_funnotdef,
 				  isc_arg_string, ERR_cstring(name),
 				  isc_arg_interpreted,
-				  "module name or entrypoint could not be found", isc_arg_end);
+				  "module name or entrypoint could not be found", 0);
 		}
 
 	jrd_nod* node = PAR_make_node(tdbb, e_fun_length);
@@ -1332,8 +1486,8 @@ static jrd_nod* par_literal(thread_db* tdbb, CompilerScratch* csb)
 	SET_TDBB(tdbb);
 
 	DSC desc;
-	PAR_desc(csb, &desc);
-	const SSHORT count = lit_delta +
+	PAR_desc(tdbb, csb, &desc);
+	const int count = lit_delta +
 		(desc.dsc_length + sizeof(jrd_nod*) - 1) / sizeof(jrd_nod*);
 	jrd_nod* node = PAR_make_node(tdbb, count);
 	Literal* literal = (Literal*) node;
@@ -1343,7 +1497,7 @@ static jrd_nod* par_literal(thread_db* tdbb, CompilerScratch* csb)
 	literal->lit_desc.dsc_address = p;
 	literal->lit_desc.dsc_flags = 0;
 	const UCHAR* q = csb->csb_running;
-	SSHORT l = desc.dsc_length;
+	USHORT l = desc.dsc_length;
 
 	switch (desc.dsc_dtype) {
 	case dtype_short:
@@ -1479,16 +1633,31 @@ static jrd_nod* par_message(thread_db* tdbb, CompilerScratch* csb)
 	ULONG offset = 0;
 
 	Format::fmt_desc_iterator desc, end;
-	for (desc = format->fmt_desc.begin(), end = desc + n; desc < end; ++desc) {
-		const USHORT alignment = PAR_desc(csb, &*desc);
+	USHORT index = 0;
+
+	for (desc = format->fmt_desc.begin(), end = desc + n; desc < end; ++desc, ++index)
+	{
+		ItemInfo itemInfo;
+		const USHORT alignment = PAR_desc(tdbb, csb, &*desc, &itemInfo);
 		if (alignment)
 			offset = FB_ALIGN(offset, alignment);
 		desc->dsc_address = (UCHAR *) (IPTR) offset;
 		offset += desc->dsc_length;
+
+		// ASF: Odd indexes are the nullable flag.
+		// So we only check even indexes, which is the actual parameter.
+		if (itemInfo.isSpecial() && index % 2 == 0)
+		{
+			csb->csb_dbg_info.argInfoToName.get(
+				Firebird::ArgumentInfo(csb->csb_msg_number, index / 2), itemInfo.name);
+
+			csb->csb_map_item_info.put(
+				Item(nod_argument, csb->csb_msg_number, index), itemInfo);
+		}
 	}
 
 	if (offset > MAX_FORMAT_SIZE)
-		error(csb, isc_imp_exc, isc_arg_gds, isc_blktoobig, isc_arg_end);
+		error(csb, isc_imp_exc, isc_arg_gds, isc_blktoobig, 0);
 
 	format->fmt_length = (USHORT) offset;
 
@@ -1496,7 +1665,7 @@ static jrd_nod* par_message(thread_db* tdbb, CompilerScratch* csb)
 }
 
 
-static jrd_nod* par_modify(thread_db* tdbb, CompilerScratch* csb)
+static jrd_nod* par_modify(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_operator)
 {
 /**************************************
  *
@@ -1512,17 +1681,17 @@ static jrd_nod* par_modify(thread_db* tdbb, CompilerScratch* csb)
 
 /* Parse the original and new contexts */
 
-	SSHORT context = (unsigned int) BLR_BYTE;
+	USHORT context = (unsigned int) BLR_BYTE;
 	if (context >= csb->csb_rpt.getCount() || 
 		!(csb->csb_rpt[context].csb_flags & csb_used))
 	{
-		error(csb, isc_ctxnotdef, isc_arg_end);
+		error(csb, isc_ctxnotdef, 0);
 	}
 	const SSHORT org_stream = csb->csb_rpt[context].csb_stream;
 	const SSHORT new_stream = csb->nextStream(false);
 	if (new_stream >= MAX_STREAMS)
 	{
-		error(csb, isc_too_many_contexts, isc_arg_end);
+		error(csb, isc_too_many_contexts, 0);
 	}
 	context = (unsigned int) BLR_BYTE;
 
@@ -1544,6 +1713,12 @@ static jrd_nod* par_modify(thread_db* tdbb, CompilerScratch* csb)
 	node->nod_arg[e_mod_new_stream] = (jrd_nod*) (IPTR) new_stream;
 	node->nod_arg[e_mod_statement] = parse(tdbb, csb, STATEMENT);
 
+	if (blr_operator == blr_modify2)
+	{
+		node->nod_count = 2;
+		node->nod_arg[e_mod_statement2] = parse(tdbb, csb, STATEMENT);
+	}
+
 	return node;
 }
 
@@ -1572,7 +1747,7 @@ static USHORT par_name(CompilerScratch* csb, Firebird::MetaName& name)
 			*s++ = BLR_BYTE;
 		}
 		*s = 0;
-		ERR_post(isc_identifier_too_long, isc_arg_string, ERR_cstring(st), isc_arg_end);
+		ERR_post(isc_identifier_too_long, isc_arg_string, ERR_cstring(st), 0);
 	}
 
 	char* s = name.getBuffer(l);
@@ -1596,13 +1771,11 @@ static size_t par_name(CompilerScratch* csb, Firebird::string& name)
  *
  * Functional description
  *	Parse a counted string of virtually unlimited size 
- *  (up to 64K), returning count.
+ *  (up to 64K, actually <= 255), returning count.
  *
  **************************************/
 	size_t l = BLR_BYTE;
-
-	name.assign(l, ' ');
-	char* s = name.begin();
+	char* s = name.getBuffer(l);
 
 	while (l--) 
 	{
@@ -1654,7 +1827,7 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 		   itself is redundant except in the case of a view,
 		   in which case the base relation (and alias) must be specified */
 
-		SSHORT n = BLR_BYTE;
+		USHORT n = (unsigned int) BLR_BYTE;
 		if (n != blr_relation && n != blr_relation2 &&
 			n != blr_rid && n != blr_rid2)
 		{
@@ -1671,7 +1844,7 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 
 		n = BLR_BYTE;
 		if (n >= csb->csb_rpt.getCount() || !(csb->csb_rpt[n].csb_flags & csb_used))
-			error(csb, isc_ctxnotdef, isc_arg_end);
+			error(csb, isc_ctxnotdef, 0);
 		const SSHORT stream = csb->csb_rpt[n].csb_stream;
 
 		relation_node->nod_arg[e_rel_stream] = (jrd_nod*) (IPTR) stream;
@@ -1707,14 +1880,18 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 				if (idx_status == MET_object_unknown ||
 					idx_status == MET_object_inactive)
 				{
-					if (tdbb->tdbb_attachment->att_flags & ATT_gbak_attachment)
+					if (tdbb->getAttachment()->att_flags & ATT_gbak_attachment)
+					{
 						warning(csb, isc_indexname, isc_arg_string,
 								ERR_cstring(name), isc_arg_string,
-								relation->rel_name.c_str(), isc_arg_end);
+								relation->rel_name.c_str(), 0);
+					}
 					else
+					{
 						error(csb, isc_indexname, isc_arg_string,
 							  ERR_cstring(name), isc_arg_string,
-							  relation->rel_name.c_str(), isc_arg_end);
+							  relation->rel_name.c_str(), 0);
+					}
 				}
 
 				/* save both the relation id and the index id, since
@@ -1723,7 +1900,7 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 
 				access_type->nod_arg[e_access_type_relation] = (jrd_nod*) (IPTR) relation_id;
 				access_type->nod_arg[e_access_type_index] = (jrd_nod*) (IPTR) index_id;
-				idx_name = ALL_cstring(tdbb->getDefaultPool(), name.c_str());
+				idx_name = stringDup(*tdbb->getDefaultPool(), name.c_str());
 				access_type->nod_arg[e_access_type_index_name] = (jrd_nod*) idx_name;
 
 				if (csb->csb_g_flags & csb_get_dependencies)
@@ -1772,14 +1949,18 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 					if (idx_status == MET_object_unknown ||
 						idx_status == MET_object_inactive)
 					{
-						if (tdbb->tdbb_attachment->att_flags & ATT_gbak_attachment)
+						if (tdbb->getAttachment()->att_flags & ATT_gbak_attachment)
+						{
 							warning(csb, isc_indexname, isc_arg_string,
 									ERR_cstring(name), isc_arg_string,
-									relation->rel_name.c_str(), isc_arg_end);
+									relation->rel_name.c_str(), 0);
+						}
 						else
+						{
 							error(csb, isc_indexname, isc_arg_string,
 								  ERR_cstring(name), isc_arg_string,
-								  relation->rel_name.c_str(), isc_arg_end);
+								  relation->rel_name.c_str(), 0);
+						}
 					}
 
 					/* save both the relation id and the index id, since
@@ -1788,7 +1969,7 @@ static jrd_nod* par_plan(thread_db* tdbb, CompilerScratch* csb)
 
 					*arg++ = (jrd_nod*) (IPTR) relation_id;
 					*arg++ = (jrd_nod*) (IPTR) index_id;
-					idx_name = ALL_cstring(tdbb->getDefaultPool(), name.c_str());
+					idx_name = stringDup(*tdbb->getDefaultPool(), name.c_str());
 					*arg++ = (jrd_nod*) idx_name;
 
 					if (csb->csb_g_flags & csb_get_dependencies)
@@ -1845,7 +2026,13 @@ static jrd_nod* par_procedure(thread_db* tdbb, CompilerScratch* csb, SSHORT blr_
 				name.printf("id %d", pid);
 		}
 		if (!procedure)
-			error(csb, isc_prcnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_prcnotdef, isc_arg_string, ERR_cstring(name), 0);
+	}
+
+	if (procedure->prc_type == prc_executable)
+	{
+		error(csb, isc_illegal_prc_type,
+			  isc_arg_string, ERR_string(procedure->prc_name), 0);
 	}
 
 	jrd_nod* node = PAR_make_node(tdbb, e_prc_length);
@@ -1900,7 +2087,7 @@ static void par_procedure_parms(
 				  input_flag ? isc_prcmismat : isc_prc_out_param_mismatch,
 				  isc_arg_string,
 				  ERR_cstring(procedure->prc_name.c_str()),
-				  isc_arg_end);
+				  0);
 		}
 		else
 			mismatch = true;
@@ -1971,8 +2158,9 @@ static void par_procedure_parms(
 
 			// default value for parameter 
 			if ((count <= 0) && input_flag) {
-				Parameter* parameter = (*procedure->prc_input_fields)[procedure->prc_inputs - n];
-				asgn->nod_arg[asgn_arg1] = CMP_clone_node(tdbb, csb, parameter->prm_default_value);
+				Parameter* parameter =
+					(*procedure->prc_input_fields)[procedure->prc_inputs - n];
+				asgn->nod_arg[asgn_arg1] = parameter->prm_default_value;
 			}
 			else {
 				asgn->nod_arg[asgn_arg1] = parse(tdbb, csb, VALUE);
@@ -1998,7 +2186,7 @@ static void par_procedure_parms(
 			  input_flag ? isc_prcmismat : isc_prc_out_param_mismatch,
 			  isc_arg_string,
 			  ERR_cstring(procedure->prc_name.c_str()),
-			  isc_arg_end);
+			  0);
 	}
 }
 
@@ -2037,7 +2225,7 @@ static jrd_nod* par_relation(
 		}
 		if (!(relation = MET_lookup_relation_id(tdbb, id, false))) {
 			name.printf("id %d", id);
-			error(csb, isc_relnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_relnotdef, isc_arg_string, ERR_cstring(name), 0);
 		}
 	}
 	else if (blr_operator == blr_relation || blr_operator == blr_relation2) {
@@ -2047,7 +2235,7 @@ static jrd_nod* par_relation(
 			par_name(csb, *alias_string);
 		}
 		if (!(relation = MET_lookup_relation(tdbb, name)))
-			error(csb, isc_relnotdef, isc_arg_string, ERR_cstring(name), isc_arg_end);
+			error(csb, isc_relnotdef, isc_arg_string, ERR_cstring(name), 0);
 	}
 
 /* if an alias was passed, store with the relation */
@@ -2269,6 +2457,7 @@ static jrd_nod* par_sort(thread_db* tdbb, CompilerScratch* csb, bool flag)
 }
 
 
+#ifdef NOT_USED_OR_REPLACED
 static jrd_nod* par_stream(thread_db* tdbb, CompilerScratch* csb)
 {
 /**************************************
@@ -2302,9 +2491,44 @@ static jrd_nod* par_stream(thread_db* tdbb, CompilerScratch* csb)
 		}
 	}
 }
+#endif
 
 
-static jrd_nod* par_union(thread_db* tdbb, CompilerScratch* csb)
+static jrd_nod* par_sys_function(thread_db* tdbb, CompilerScratch* csb)
+{
+/**************************************
+ *
+ *	p a r _ s y s _ f u n c t i o n
+ *
+ **************************************
+ *
+ * Functional description
+ *	Parse a system function reference.
+ *
+ **************************************/
+	SET_TDBB(tdbb);
+	
+	Firebird::MetaName name;
+	const USHORT count = par_name(csb, name);
+
+	const SysFunction* function = SysFunction::lookup(name);
+
+	if (!function)
+	{
+		csb->csb_running -= count;
+		error(csb, isc_funnotdef, isc_arg_string, ERR_cstring(name), 0);
+	}
+
+	jrd_nod* node = PAR_make_node(tdbb, e_sysfun_length);
+	node->nod_count = count_table[blr_sys_function];
+	node->nod_arg[e_sysfun_args] = par_args(tdbb, csb, VALUE);
+	node->nod_arg[e_sysfun_function] = (jrd_nod*) function;
+
+	return node;
+}
+
+
+static jrd_nod* par_union(thread_db* tdbb, CompilerScratch* csb, bool recursive)
 {
 /**************************************
  *
@@ -2322,9 +2546,19 @@ static jrd_nod* par_union(thread_db* tdbb, CompilerScratch* csb)
    and get the number of sub-RecordSelExpr's. */
 
 	jrd_nod* node = PAR_make_node(tdbb, e_uni_length);
-	node->nod_count = 2;
+	node->nod_count = 3;
 	const USHORT stream = par_context(csb, 0);
 	node->nod_arg[e_uni_stream] = (jrd_nod*) (IPTR) stream;
+
+	// assign separate context for mapped record if union is recursive
+	USHORT map_stream = stream;
+	if (recursive)
+	{
+		node->nod_flags |= nod_recurse;
+		map_stream = par_context(csb, 0);
+		node->nod_arg[e_uni_map_stream] = (jrd_nod*) (IPTR) map_stream;
+	}
+
 	SSHORT count = (unsigned int) BLR_BYTE;
 
 /* Pick up the sub-RecordSelExpr's and maps */
@@ -2333,7 +2567,7 @@ static jrd_nod* par_union(thread_db* tdbb, CompilerScratch* csb)
 
 	while (--count >= 0) {
 		clauses.push(parse(tdbb, csb, TYPE_RSE));
-		clauses.push(par_map(tdbb, csb, stream));
+		clauses.push(par_map(tdbb, csb, map_stream));
 	}
 
 	node->nod_arg[e_uni_clauses] = PAR_make_list(tdbb, clauses);
@@ -2376,6 +2610,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
  **************************************/
 	SET_TDBB(tdbb);
 
+	const USHORT blr_offset = csb->csb_running - csb->csb_blr;
 	const SSHORT blr_operator = BLR_BYTE;
 
 	if (blr_operator < 0 || blr_operator >= FB_NELEM(type_table)) {
@@ -2414,6 +2649,8 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 		arg = NULL;
 	}
 
+	bool set_type = true;
+
 /* Dispatch on operator type. */
 
 	switch (blr_operator) {
@@ -2495,12 +2732,18 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 		*arg++ = parse(tdbb, csb, sub_type);
 		break;
 
+	case blr_agg_list:
+	case blr_agg_list_distinct:
+		*arg++ = parse(tdbb, csb, sub_type);
+		*arg++ = parse(tdbb, csb, VALUE);
+		break;
+
 	case blr_exec_sql:
 		*arg++ = parse(tdbb, csb, sub_type);
 		break;
 
 	case blr_exec_into:
-		n = BLR_WORD + 2 /*e_exec_into_count - 1*/ ;
+		n = BLR_WORD + 2; // e_exec_into_count - 1
 		node = PAR_make_node(tdbb, n);
 		arg = node->nod_arg;
 		*arg++ = parse(tdbb, csb, VALUE);
@@ -2539,7 +2782,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 		n = BLR_BYTE;
 		if (n > MAX_TIME_PRECISION) {
 			ERR_post(isc_invalid_time_precision,
-					 isc_arg_number, MAX_TIME_PRECISION, isc_arg_end);
+					 isc_arg_number, MAX_TIME_PRECISION, 0);
 		}
 		node->nod_arg[0] = (jrd_nod*) (IPTR) n;
 		break;
@@ -2549,7 +2792,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 			*arg++ = (jrd_nod*) (IPTR) BLR_BYTE;
 			Firebird::MetaName name;
 			par_name(csb, name);
-			*arg++ = (jrd_nod*) ALL_cstring(tdbb->getDefaultPool(), name.c_str());
+			*arg++ = (jrd_nod*) stringDup(*tdbb->getDefaultPool(), name.c_str());
 			break;
 		}
 
@@ -2581,13 +2824,14 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 	case blr_erase:
 		n = BLR_BYTE;
 		if (n >= csb->csb_rpt.getCount() || !(csb->csb_rpt[n].csb_flags & csb_used))
-			error(csb, isc_ctxnotdef, isc_arg_end);
+			error(csb, isc_ctxnotdef, 0);
 		node->nod_arg[e_erase_stream] =
 			(jrd_nod*) (IPTR) csb->csb_rpt[n].csb_stream;
 		break;
 	
 	case blr_modify:
-		node = par_modify(tdbb, csb);
+	case blr_modify2:
+		node = par_modify(tdbb, csb, blr_operator);
 		break;
 
 	case blr_exec_proc:
@@ -2648,9 +2892,9 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 			if (BLR_PEEK == blr_seek)
 				node->nod_arg[e_cursor_stmt_seek] = parse(tdbb, csb, STATEMENT);
 #endif
-			csb->csb_g_flags |= csb_no_context_check;
+			csb->csb_g_flags |= csb_reuse_context;
 			node->nod_arg[e_cursor_stmt_into] = parse(tdbb, csb, STATEMENT);
-			csb->csb_g_flags &= ~csb_no_context_check;
+			csb->csb_g_flags &= ~csb_reuse_context;
 			break;
 		default:
 			syntax_error(csb, "cursor operation clause");
@@ -2675,7 +2919,11 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 		break;
 
 	case blr_union:
-		node = par_union(tdbb, csb);
+		node = par_union(tdbb, csb, false);
+		break;
+
+	case blr_recurse:
+		node = par_union(tdbb, csb, true);
 		break;
 
 	case blr_aggregate:
@@ -2694,6 +2942,8 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 	case blr_field:
 	case blr_fid:
 		node = par_field(tdbb, csb, blr_operator);
+		if (node->nod_type == nod_domain_validation || node->nod_type == nod_null)
+			set_type = false;	// to not change nod->nod_type to nod_field
 		break;
 
 	case blr_gen_id:
@@ -2705,7 +2955,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 			const SLONG tmp = MET_lookup_generator(tdbb, name.c_str());
 			if (tmp < 0) {
 				error(csb, isc_gennotdef,
-					  isc_arg_string, ERR_cstring(name), isc_arg_end);
+					  isc_arg_string, ERR_cstring(name), 0);
 			}
 			node->nod_arg[e_gen_relation] = (jrd_nod*) (IPTR) tmp;
 			node->nod_arg[e_gen_value] = parse(tdbb, csb, VALUE);
@@ -2730,7 +2980,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 	case blr_dbkey:
 		n = BLR_BYTE;
 		if (n >= csb->csb_rpt.getCount() || !(csb->csb_rpt[n].csb_flags & csb_used))
-			error(csb, isc_ctxnotdef, isc_arg_end);
+			error(csb, isc_ctxnotdef, 0);
 		node->nod_arg[0] = (jrd_nod*) (IPTR) csb->csb_rpt[n].csb_stream;
 		break;
 
@@ -2773,12 +3023,31 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 
 	case blr_dcl_variable:
 		{
+			dsc* desc = (dsc*) (node->nod_arg + e_dcl_desc);
+
+			ItemInfo itemInfo;
+
 			n = BLR_WORD;
 			node->nod_arg[e_dcl_id] = (jrd_nod*) (IPTR) n;
-			PAR_desc(csb, (DSC *) (node->nod_arg + e_dcl_desc));
+			PAR_desc(tdbb, csb, desc, &itemInfo);
 			vec<jrd_nod*>* vector = csb->csb_variables =
 				vec<jrd_nod*>::newVector(*tdbb->getDefaultPool(), csb->csb_variables, n + 1);
 			(*vector)[n] = node;
+
+			if (itemInfo.isSpecial())
+			{
+				csb->csb_dbg_info.varIndexToName.get(n, itemInfo.name);
+				csb->csb_map_item_info.put(Item(nod_variable, n), itemInfo);
+			}
+
+			if (itemInfo.explicitCollation)
+			{
+				jrd_nod* dep_node = PAR_make_node (tdbb, e_dep_length);
+				dep_node->nod_type = nod_dependency;
+				dep_node->nod_arg [e_dep_object] = (jrd_nod*)(IPTR) INTL_TEXT_TYPE(*desc);
+				dep_node->nod_arg [e_dep_object_type] = (jrd_nod*)(IPTR) obj_collation;
+				csb->csb_dependencies.push(dep_node);
+			}
 		}
 		break;
 
@@ -2804,14 +3073,14 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 			if (n >= csb->csb_rpt.getCount() ||
 				!(message = csb->csb_rpt[n].csb_message))
 			{
-				error(csb, isc_badmsgnum, isc_arg_end);
+				error(csb, isc_badmsgnum, 0);
 			}
 			node->nod_arg[e_arg_message] = message;
 			n = BLR_WORD;
 			node->nod_arg[e_arg_number] = (jrd_nod*) (IPTR) n;
 			const Format* format = (Format*) message->nod_arg[e_msg_format];
 			if (n >= format->fmt_count)
-				error(csb, isc_badparnum, isc_arg_end);
+				error(csb, isc_badparnum, 0);
 			if (blr_operator != blr_parameter) {
 				jrd_nod* temp = PAR_make_node(tdbb, e_arg_length);
 				node->nod_arg[e_arg_flag] = temp;
@@ -2822,7 +3091,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 				n = BLR_WORD;
 				temp->nod_arg[e_arg_number] = (jrd_nod*) (IPTR) n;
 				if (n >= format->fmt_count)
-					error(csb, isc_badparnum, isc_arg_end);
+					error(csb, isc_badparnum, 0);
 			}
 			if (blr_operator == blr_parameter3) {
 				jrd_nod* temp = PAR_make_node(tdbb, e_arg_length);
@@ -2834,7 +3103,7 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 				n = BLR_WORD;
 				temp->nod_arg[e_arg_number] = (jrd_nod*) (IPTR) n;
 				if (n >= format->fmt_count)
-					error(csb, isc_badparnum, isc_arg_end);
+					error(csb, isc_badparnum, 0);
 			}
 		}
 		break;
@@ -2932,14 +3201,48 @@ static jrd_nod* parse(thread_db* tdbb, CompilerScratch* csb, USHORT expected,
 		break;
 #endif
 
+	case blr_init_variable:
+		{
+			n = BLR_WORD;
+			node->nod_arg[e_init_var_id] = (jrd_nod*)(U_IPTR) n;
+			vec<jrd_nod*>* vector = csb->csb_variables;
+			if (!vector || n >= vector->count() ||
+				!(node->nod_arg[e_init_var_variable] = (*vector)[n]))
+			{
+				syntax_error(csb, "variable identifier");
+			}
+		}
+		break;
+
+	case blr_sys_function:
+		node = par_sys_function(tdbb, csb);
+		break;
+
 	default:
 		syntax_error(csb, elements[expected]);
 	}
 
-	if (csb->csb_g_flags & csb_blr_version4)
-		node->nod_type = (NOD_T) (USHORT) blr_table4[(int) blr_operator];
-	else
-		node->nod_type = (NOD_T) (USHORT) blr_table[(int) blr_operator];
+	if (set_type)
+	{
+		if (csb->csb_g_flags & csb_blr_version4)
+			node->nod_type = (NOD_T) (USHORT) blr_table4[(int) blr_operator];
+		else
+			node->nod_type = (NOD_T) (USHORT) blr_table[(int) blr_operator];
+	}
+
+	size_t pos = 0;
+	if (csb->csb_dbg_info.blrToSrc.find(blr_offset, pos))
+	{
+		Firebird::MapBlrToSrcItem& i = csb->csb_dbg_info.blrToSrc[pos];
+		jrd_nod* node_src = PAR_make_node(tdbb, e_src_info_length);
+
+		node_src->nod_type = nod_src_info;
+		node_src->nod_arg[e_src_info_line] = (jrd_nod*) (IPTR) i.mbs_src_line;
+		node_src->nod_arg[e_src_info_col] = (jrd_nod*) (IPTR) i.mbs_src_col;
+		node_src->nod_arg[e_src_info_node] = node;
+
+		return node_src;
+	}
 
 	return node;
 }
@@ -2961,7 +3264,7 @@ static void syntax_error(CompilerScratch* csb, const TEXT* string)
 	error(csb, isc_syntaxerr,
 		  isc_arg_string, string,
 		  isc_arg_number, (SLONG) (csb->csb_running - csb->csb_blr - 1),
-		  isc_arg_number, (SLONG) csb->csb_running[-1], isc_arg_end);
+		  isc_arg_number, (SLONG) csb->csb_running[-1], 0);
 }
 
 
