@@ -19,8 +19,6 @@
  *
  * All Rights Reserved.
  * Contributor(s): ______________________________________.
- * Added TCP_NO_DELAY option for superserver on Linux
- * FSG 16.03.2001
  *
  * 2002.02.15 Sean Leyne - Code Cleanup, removed obsolete "EPSON" port
  * 2002.02.15 Sean Leyne - Code Cleanup, removed obsolete "XENIX" port
@@ -86,6 +84,7 @@
 
 #if !(defined VMS || defined WIN_NT)
 #include <netinet/tcp.h>
+#include <netinet/in.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #endif // !(defined VMS || defined WIN_NT)
@@ -132,15 +131,11 @@ const char* GDS_HOSTS_FILE = "";
 const char* PROXY_FILE	= "/etc/gds_proxy";
 const char* HOSTS_FILE	= "/etc/hosts.equiv";
 #ifdef MULTI_THREAD
-const USHORT MAX_PTYPE	= ptype_batch_send;
+const USHORT MAX_PTYPE	= ptype_lazy_send;
 #else
 const USHORT MAX_PTYPE	= ptype_out_of_band;
 #endif
-#ifdef SMALL_FILE_NAMES
-const char* GDS_HOSTS_FILE	= "/etc/gdshosts.eqv";
-#else
 const char* GDS_HOSTS_FILE	= "/etc/gds_hosts.equiv";
-#endif
 #endif // VMS
 
 #ifdef WIN_NT
@@ -185,6 +180,14 @@ const int NOTASOCKET = EBADF;
 #define ENOBUFS	0
 #endif
 
+#ifndef FB_SEND_FLAGS
+#define FB_SEND_FLAGS 0
+#endif
+
+#ifndef FB_SETOPT_FLAGS
+#define FB_SETOPT_FLAGS 0
+#endif
+
 SLONG INET_remote_buffer;
 SLONG INET_max_data;
 static bool first_time = true;
@@ -285,7 +288,7 @@ static int fork(SOCKET, USHORT);
 #endif
 
 static in_addr get_bind_address();
-static in_addr get_host_address(const TEXT *);
+static in_addr get_host_address(const Firebird::string&);
 
 static void copy_p_cnct_repeat_array(	p_cnct::p_cnct_repeat*			pDest,
 										const p_cnct::p_cnct_repeat*	pSource,
@@ -294,7 +297,9 @@ static void copy_p_cnct_repeat_array(	p_cnct::p_cnct_repeat*			pDest,
 static void		inet_copy(const void*, UCHAR*, int);
 static int		inet_destroy(XDR *);
 static void		inet_gen_error(rem_port*, ISC_STATUS, ...);
+#if !defined(SUPERSERVER) || defined(EMBEDDED)
 static bool_t	inet_getbytes(XDR *, SCHAR *, u_int);
+#endif
 static bool_t	inet_getlong(XDR *, SLONG *);
 static u_int	inet_getpostn(XDR *);
 #if !(defined WIN_NT)
@@ -311,8 +316,8 @@ static rem_port*		inet_try_connect(	PACKET*,
 									Firebird::PathName&,
 									const TEXT*,
 									ISC_STATUS*,
-									const SCHAR*,
-									SSHORT);
+									const UCHAR*,
+									USHORT);
 static bool_t	inet_write(XDR *, int);
 static void		inet_zero(sockaddr_in*, int);
 #if !(defined WIN_NT)
@@ -330,7 +335,7 @@ static rem_port*		receive(rem_port*, PACKET *);
 static rem_port*		select_accept(rem_port*);
 
 static rem_port*		select_port(rem_port*, SLCT *);
-static bool		select_multi(rem_port*, UCHAR* buffer, SSHORT bufsize, SSHORT* length, rem_port*& port);
+static rem_port*        select_multi(rem_port*, UCHAR* buffer, SSHORT bufsize, SSHORT* length);
 static int		select_wait(rem_port*, SLCT *);
 static int		send_full(rem_port*, PACKET *);
 static int		send_partial(rem_port*, PACKET *);
@@ -348,10 +353,10 @@ static XDR::xdr_ops inet_ops =
 {
 	inet_getlong,
 	inet_putlong,
-#if defined(SUPERSERVER) && !defined(EMBEDDED)
-	REMOTE_getbytes,
-#else
+#if !defined(SUPERSERVER) || defined(EMBEDDED)
 	inet_getbytes,
+#else
+	REMOTE_getbytes,
 #endif
 	inet_putbytes,
 	inet_getpostn,
@@ -397,31 +402,27 @@ static WSADATA INET_wsadata;
 
 #ifdef  SUPERSERVER
 
-static MUTX_T	port_mutex;
-static bool		port_mutex_inited = false;
+static Firebird::Mutex	port_mutex;
 
 #define DEFER_PORT_CLEANUP
 
-inline void START_PORT_CRITICAL() {
-	if (!port_mutex_inited) {
-		port_mutex_inited = true;
-	}
+inline void START_PORT_CRITICAL() 
+{
 	THREAD_EXIT();
-	THD_mutex_lock (&port_mutex);
+	port_mutex.enter();
 	THREAD_ENTER();
 }
 
-inline void STOP_PORT_CRITICAL() {
+inline void STOP_PORT_CRITICAL()
+{
 	THREAD_EXIT();
-	THD_mutex_unlock (&port_mutex);
+	port_mutex.leave();
 	THREAD_ENTER();
 }
 
 #else
-inline void START_PORT_CRITICAL() {
-}
-inline void STOP_PORT_CRITICAL() {
-}
+inline void START_PORT_CRITICAL() { }
+inline void STOP_PORT_CRITICAL() { }
 #endif
 
 
@@ -430,8 +431,8 @@ rem_port* INET_analyze(Firebird::PathName& file_name,
 					const TEXT*	node_name,
 					const TEXT*	user_string,
 					bool	uv_flag,
-					const SCHAR*	dpb,
-					SSHORT	dpb_length)
+					const UCHAR*	dpb,
+					USHORT	dpb_length)
 {
 /**************************************
  *
@@ -498,10 +499,11 @@ rem_port* INET_analyze(Firebird::PathName& file_name,
 	static const p_cnct::p_cnct_repeat protocols_to_try1[] =
 	{
 		REMOTE_PROTOCOL(PROTOCOL_VERSION8, ptype_rpc, MAX_PTYPE, 1),
-		REMOTE_PROTOCOL(PROTOCOL_VERSION10, ptype_rpc, MAX_PTYPE, 2)
+		REMOTE_PROTOCOL(PROTOCOL_VERSION10, ptype_rpc, MAX_PTYPE, 2),
+		REMOTE_PROTOCOL(PROTOCOL_VERSION11, ptype_rpc, MAX_PTYPE, 3),
 #ifdef SCROLLABLE_CURSORS
 		,
-		REMOTE_PROTOCOL(PROTOCOL_SCROLLABLE_CURSORS, ptype_rpc, MAX_PTYPE, 3)
+		REMOTE_PROTOCOL(PROTOCOL_SCROLLABLE_CURSORS, ptype_rpc, MAX_PTYPE, 99)
 #endif
 	};
 
@@ -588,9 +590,9 @@ rem_port* INET_analyze(Firebird::PathName& file_name,
 
 /* once we've decided on a protocol, concatenate the version
    string to reflect it...  */
-
 	Firebird::string temp;
-	temp.printf("%s/P%d", port->port_version->str_data, port->port_protocol);
+	temp.printf("%s/P%d", port->port_version->str_data, 
+						  port->port_protocol & FB_PROTOCOL_MASK);
 	ALLR_free(port->port_version);
 	port->port_version = REMOTE_make_string(temp.c_str());
 
@@ -606,13 +608,17 @@ rem_port* INET_analyze(Firebird::PathName& file_name,
 		port->port_flags |= PORT_no_oob;
 	}
 
+	if (packet->p_acpt.p_acpt_type == ptype_lazy_send) {
+		port->port_flags |= PORT_lazy;
+	}
+
 	return port;
 }
 
 rem_port* INET_connect(const TEXT* name,
 							 PACKET* packet,
 							 ISC_STATUS* status_vector,
-							 USHORT flag, const SCHAR* dpb, SSHORT dpb_length)
+							 USHORT flag, const UCHAR* dpb, USHORT dpb_length)
 {
 /**************************************
  *
@@ -643,8 +649,7 @@ rem_port* INET_connect(const TEXT* name,
 
 	rem_port* port = alloc_port(0);
 	port->port_status_vector = status_vector;
-	REMOTE_get_timeout_params(port, reinterpret_cast<const UCHAR*>(dpb),
-							  dpb_length);
+	REMOTE_get_timeout_params(port, dpb, dpb_length);
 	status_vector[0] = isc_arg_gds;
 	status_vector[1] = 0;
 	status_vector[2] = isc_arg_end;
@@ -708,6 +713,8 @@ rem_port* INET_connect(const TEXT* name,
 	}
 #else
 
+	TEXT msg[BUFFER_SMALL];
+
 /* U N I X style sockets */
 
 	address.sin_family = AF_INET;
@@ -716,18 +723,20 @@ rem_port* INET_connect(const TEXT* name,
 
 	if (packet) {
 		// client connection
-		host_addr = get_host_address(host.c_str());
+		host_addr = get_host_address(host);
 
 		if (host_addr.s_addr == INADDR_NONE)
 		{
-			gds__log("INET/INET_connect: gethostbyname (%s) failed, error code = %d",
-					 host.c_str(), H_ERRNO);
+			SNPRINTF(msg, FB_NELEM(msg),
+					"INET/INET_connect: gethostbyname (%s) failed, error code = %d",
+					host.c_str(), H_ERRNO);
+			gds__log(msg, 0);
 			inet_gen_error(port,
 						   isc_network_error,
 						   isc_arg_string,
 						   port->port_connection->str_data,
 						   isc_arg_gds,
-						   isc_net_lookup_err, isc_arg_gds, isc_host_unknown, isc_arg_end);
+						   isc_net_lookup_err, isc_arg_gds, isc_host_unknown, 0);
 
 			disconnect(port);
 			return NULL;
@@ -790,8 +799,10 @@ rem_port* INET_connect(const TEXT* name,
 		{
 			/* end of modification by FSG */
 			/* this is the original code */
-			gds__log("INET/INET_connect: getservbyname failed, error code = %d",
-					 H_ERRNO);
+			SNPRINTF(msg, FB_NELEM(msg),
+					"INET/INET_connect: getservbyname failed, error code = %d",
+					H_ERRNO);
+			gds__log(msg, 0);
 			inet_gen_error(port,
 						   isc_network_error,
 						   isc_arg_string,
@@ -801,7 +812,8 @@ rem_port* INET_connect(const TEXT* name,
 						   isc_arg_gds,
 						   isc_service_unknown,
 						   isc_arg_string,
-						   protocol.c_str(), isc_arg_string, "tcp", isc_arg_end);
+						   protocol.c_str(),
+						   isc_arg_string, "tcp", 0);
 			return NULL;
 		}						/* else / not hardwired gds_db translation */
 	}
@@ -906,12 +918,11 @@ rem_port* INET_connect(const TEXT* name,
 					break;
 			}
 		}
-	}
-
-	if (n == -1) {
-		inet_error(port, "bind", isc_net_connect_listen_err, INET_ERRNO);
-		disconnect(port);
-		return NULL;
+		else {
+			inet_error(port, "bind", isc_net_connect_listen_err, INET_ERRNO);
+			disconnect(port);
+			return NULL;
+		}
 	}
 
 	n = listen((SOCKET) port->port_handle, 5);
@@ -984,8 +995,8 @@ rem_port* INET_reconnect(HANDLE handle, ISC_STATUS* status_vector)
 	port->port_handle = handle;
 	port->port_server_flags |= SRVR_server;
 
-	int optval = TRUE;
-	int n = setsockopt((SOCKET) port->port_handle, SOL_SOCKET,
+	int n = 0, optval = TRUE;
+	n = setsockopt((SOCKET) port->port_handle, SOL_SOCKET,
 					SO_KEEPALIVE, (SCHAR*) &optval, sizeof(optval));
 	if (n == -1) {
 		gds__log("inet server err: setting KEEPALIVE socket option \n");
@@ -1011,6 +1022,7 @@ rem_port* INET_server(int sock)
  *	established.  Set up port block with the appropriate socket.
  *
  **************************************/
+	int n = 0;
 #ifdef VMS
 	ISC_tcp_setup(ISC_wait, gds__completion_ast);
 #endif
@@ -1019,7 +1031,7 @@ rem_port* INET_server(int sock)
 	port->port_handle = (HANDLE) sock;
 
 	int optval = 1;
-	int n = setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
+	n = setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
 			   (SCHAR *) & optval, sizeof(optval));
 	if (n == -1) {
 		gds__log("inet server err: setting KEEPALIVE socket option \n");
@@ -1281,8 +1293,12 @@ static rem_port* alloc_port( rem_port* parent)
 		if (WSAStartup(version, &INET_wsadata)) {
 			if (parent)
 				inet_error(parent, "WSAStartup", isc_net_init_error, INET_ERRNO);
-			else
-				gds__log("INET/alloc_port: WSAStartup failed, error code = %d", INET_ERRNO);
+			else {
+				SNPRINTF(buffer, FB_NELEM(buffer),
+						 "INET/alloc_port: WSAStartup failed, error code = %d",
+						 INET_ERRNO);
+				gds__log(buffer, 0);
+			}
 			return NULL;
 		}
 		gds__register_cleanup(exit_handler, 0);
@@ -1300,7 +1316,12 @@ static rem_port* alloc_port( rem_port* parent)
 		}
 		INET_max_data = INET_remote_buffer;
 #ifdef DEBUG
-		gds__log(" Info: Remote Buffer Size set to %ld", INET_remote_buffer);
+		{
+			char msg[BUFFER_SMALL];
+			SNPRINTF(msg, FB_NELEM(msg), " Info: Remote Buffer Size set to %ld",
+					 INET_remote_buffer);
+			gds__log(msg, 0);
+		}
 #endif
 		first_time = false;
 	}
@@ -1498,7 +1519,8 @@ static rem_port* aux_request( rem_port* port, PACKET* packet)
 
 	int optval = TRUE;
 	int ret = setsockopt(n, SOL_SOCKET, SO_REUSEADDR,
-			   (SCHAR *) &optval, sizeof(optval));
+						 (SCHAR *) &optval, sizeof(optval));
+
 	if (ret == -1) {
 		inet_error(port, "setsockopt REUSE", isc_net_event_listen_err, INET_ERRNO);
 		return NULL;
@@ -1541,9 +1563,7 @@ static rem_port* aux_request( rem_port* port, PACKET* packet)
 				(UCHAR*) &address.sin_addr,
 				sizeof(address.sin_addr));
 
-	response->p_resp_data.cstr_address = 
-			reinterpret_cast<UCHAR*>(& response->p_resp_blob_id);
-	response->p_resp_data.cstr_length = sizeof(response->p_resp_blob_id);
+	response->p_resp_data.cstr_length = sizeof(address);
 	inet_copy(&address,
 			  response->p_resp_data.cstr_address,
 			  response->p_resp_data.cstr_length);
@@ -1600,7 +1620,7 @@ static int check_host(
  **************************************
  *
  * Functional description
- *	Check the host on the other end of the socket to see it
+ *	Check the host on the other end of the socket to see if
  *	it's an equivalent host.
  * NB.: First check the ~/.rhosts then the HOSTS_FILE - both have
  *	the same line formats (see parse_line)
@@ -1613,6 +1633,10 @@ static int check_host(
 	if (getpeername((int) port->port_handle, (struct sockaddr*)&address, &length) == -1)
 		return 0;
 
+	// If source address is in the loopback net - trust it
+	if((ntohl(address.sin_addr.s_addr) >> IN_CLASSA_NSHIFT) == IN_LOOPBACKNET)
+		return 1;
+
 	const struct hostent* host = gethostbyaddr((SCHAR *) & address.sin_addr,
 							   sizeof(address.sin_addr), address.sin_family);
 	if (!host)
@@ -1872,6 +1896,10 @@ static void cleanup_port( rem_port* port)
 	delete port->port_queue;
 #endif
 
+#ifdef TRUSTED_AUTH
+	delete port->port_trusted_auth;
+#endif
+
 	ALLR_release((UCHAR *) port);
 	return;
 }
@@ -1948,7 +1976,7 @@ static int fork( SOCKET old_handle, USHORT flag)
 					DUPLICATE_SAME_ACCESS);
 
 	Firebird::string cmdLine;
-	cmdLine.printf("%s -s -i -h %"SLONGFORMAT, name, (SLONG) new_handle);
+	cmdLine.printf("%s -i -h %"SLONGFORMAT, name, (SLONG) new_handle);
 
 	STARTUPINFO start_crud;
 	start_crud.cb = sizeof(STARTUPINFO);
@@ -1961,10 +1989,10 @@ static int fork( SOCKET old_handle, USHORT flag)
 	
 	PROCESS_INFORMATION pi;
 	if (CreateProcess(NULL, cmdLine.begin(), NULL, NULL, TRUE,
-							(flag & SRVR_high_priority ?
-							 HIGH_PRIORITY_CLASS | DETACHED_PROCESS :
-							 NORMAL_PRIORITY_CLASS | DETACHED_PROCESS),
-							NULL, NULL, &start_crud, &pi))
+					  (flag & SRVR_high_priority ?
+						 HIGH_PRIORITY_CLASS | DETACHED_PROCESS :
+						 NORMAL_PRIORITY_CLASS | DETACHED_PROCESS),
+					  NULL, NULL, &start_crud, &pi))
 	{
 		CloseHandle(pi.hThread);
 		CloseHandle(pi.hProcess);
@@ -1998,7 +2026,7 @@ static in_addr get_bind_address()
 	return config_address;
 }
 
-static in_addr get_host_address(const TEXT* name)
+static in_addr get_host_address(const Firebird::string& name)
 {
 /**************************************
  *
@@ -2014,11 +2042,11 @@ static in_addr get_host_address(const TEXT* name)
 
 	THREAD_EXIT();
 
-	address.s_addr = inet_addr(name);
+	address.s_addr = inet_addr(name.c_str());
 
 	if (address.s_addr == INADDR_NONE) {
 
-		const hostent* host = gethostbyname(name);
+		const hostent* host = gethostbyname(name.c_str());
 
 		/* On Windows NT/9x, gethostbyname can only accomodate
 		 * 1 call at a time.  In this case it returns the error
@@ -2031,7 +2059,7 @@ static in_addr get_host_address(const TEXT* name)
 		if (!host) {
 			if (H_ERRNO == INET_RETRY_ERRNO) {
 				for (int retry = 0; retry < INET_RETRY_CALL; retry++) {
-					if ( (host = gethostbyname(name)) )
+					if ( (host = gethostbyname(name.c_str())) )
 						break;
 				}
 			}
@@ -2346,7 +2374,7 @@ static rem_port* receive( rem_port* main_port, PACKET * packet)
 #endif //SUPERSERVER
 }
 
-static bool select_multi(rem_port* main_port, UCHAR* buffer, SSHORT bufsize, SSHORT* length, rem_port*& port)
+static rem_port* select_multi(rem_port* main_port, UCHAR* buffer, SSHORT bufsize, SSHORT* length)
 {
 /**************************************
  *
@@ -2362,21 +2390,21 @@ static bool select_multi(rem_port* main_port, UCHAR* buffer, SSHORT bufsize, SSH
  *
  **************************************/
 	fb_assert(main_port->port_server_flags & SRVR_multi_client);
-
+	
 	for (;;) 
 	{
-		port = select_port(main_port, &INET_select);
+		rem_port* port = select_port(main_port, &INET_select);
 		if (port == main_port) 
 		{
-			if ( (port = select_accept(main_port)) )
+			if (port = select_accept(main_port)) 
 			{
 				if (!packet_receive(port, buffer, bufsize, length))
 				{
 					*length = 0;
 				}
-				return (*length) ? true : false;
+				return port;
 			}
-
+			
 			continue;
 		}
 		if (port) 
@@ -2385,24 +2413,20 @@ static bool select_multi(rem_port* main_port, UCHAR* buffer, SSHORT bufsize, SSH
 			{
 				port->port_dummy_timeout = port->port_dummy_packet_interval;
 				if (port->port_flags & PORT_async ||
-					port->port_protocol < PROTOCOL_VERSION8)
-				{
-					continue;
-				}
+					port->port_protocol < PROTOCOL_VERSION8) continue;
 				*length = 0;
-				return true;
+				return port;
 			}
-
+			
 			if (!packet_receive(port, buffer, bufsize, length))
 			{
 				*length = 0;
 			}
-			return (*length) ? true : false;
+			return port;
 		}
 		if (!select_wait(main_port, &INET_select))
 		{
-			port = NULL;
-			return false;
+			return NULL;
 		}
 	}
 }
@@ -2446,7 +2470,10 @@ static rem_port* select_accept( rem_port* main_port)
 	if (n >= INET_max_clients) {
 		main_port->port_state = state_closed;
 		SOCLOSE((int) main_port->port_handle);
-		gds__log("INET/select_accept: exec new server at client limit: %d", n);
+		TEXT msg[BUFFER_SMALL];
+		SNPRINTF(msg, FB_NELEM(msg),
+				"INET/select_accept: exec new server at client limit: %d", n);
+		gds__log(msg, 0);
 
 		setreuid(0, 0);
 		kill(getppid(), SIGUSR1);
@@ -2549,6 +2576,7 @@ static int select_wait( rem_port* main_port, SLCT * selct)
  *	to read from them.
  *
  **************************************/
+	TEXT msg[BUFFER_SMALL];
 	struct timeval timeout;
 	bool checkPorts = false;
 
@@ -2647,7 +2675,8 @@ static int select_wait( rem_port* main_port, SLCT * selct)
 
 		if (!found)
 		{
-			gds__log("INET/select_wait: client rundown complete, server exiting");
+			gds__log("INET/select_wait: client rundown complete, server exiting",
+				 0);
 			return FALSE;
 		}
 
@@ -2699,7 +2728,10 @@ static int select_wait( rem_port* main_port, SLCT * selct)
 			}
 			else {
 				THREAD_ENTER();
-				gds__log("INET/select_wait: select failed, errno = %d", inetErrNo);
+				SNPRINTF(msg, FB_NELEM(msg),
+						 "INET/select_wait: select failed, errno = %d",
+						 inetErrNo);
+				gds__log(msg, 0);
 				return FALSE;
 			}
 		}	// for (;;)
@@ -2857,6 +2889,7 @@ static void inet_gen_error( rem_port* port, ISC_STATUS status, ...)
 	}
 }
 
+#if !defined(SUPERSERVER) || defined(EMBEDDED)
 static bool_t inet_getbytes( XDR * xdrs, SCHAR * buff, u_int count)
 {
 /**************************************
@@ -2916,6 +2949,7 @@ static bool_t inet_getbytes( XDR * xdrs, SCHAR * buff, u_int count)
 
 	return TRUE;
 }
+#endif
 
 static bool_t inet_getlong( XDR * xdrs, SLONG * lp)
 {
@@ -3021,6 +3055,8 @@ static void inet_error(
  *	to format the status vector if any.
  *
  **************************************/
+	TEXT msg[BUFFER_SMALL];
+
 	if (status) {
 #ifdef VMS
 		if ((status >= sys_nerr || !sys_errlist[status]) &&
@@ -3030,16 +3066,18 @@ static void inet_error(
 						   isc_arg_string,
 						   (ISC_STATUS) port->port_connection->str_data,
 						   isc_arg_gds, operation, isc_arg_string,
-						   (ISC_STATUS) win_errlist[status], isc_arg_end);
+						   (ISC_STATUS) win_errlist[status], 0);
 		}
 		else
 #endif
 			inet_gen_error(port, isc_network_error,
 						   isc_arg_string,
 						   (ISC_STATUS) port->port_connection->str_data,
-						   isc_arg_gds, operation, SYS_ERR, status, isc_arg_end);
+						   isc_arg_gds, operation, SYS_ERR, status, 0);
 
-		gds__log("INET/inet_error: %s errno = %d", function, status);
+		SNPRINTF(msg, FB_NELEM(msg),
+				 "INET/inet_error: %s errno = %d", function, status);
+		gds__log(msg, 0);
 	}
 	else {
 		/* No status value, just format the basic arguments. */
@@ -3047,7 +3085,7 @@ static void inet_error(
 		inet_gen_error(port, isc_network_error,
 					   isc_arg_string,
 					   (ISC_STATUS) port->port_connection->str_data, isc_arg_gds,
-					   operation, isc_arg_end);
+					   operation, 0);
 	}
 }
 
@@ -3219,8 +3257,8 @@ static rem_port* inet_try_connect(
 							 Firebird::PathName& file_name,
 							 const TEXT* node_name, 
 							 ISC_STATUS* status_vector,
-							 const SCHAR* dpb, 
-							 SSHORT dpb_length)
+							 const UCHAR* dpb,
+							 USHORT dpb_length)
 {
 /**************************************
  *
@@ -3550,7 +3588,7 @@ static int packet_receive(
 	}
 
 	if (!n) {
-		inet_error(port, "read end_of_file", isc_net_read_err, isc_arg_end);
+		inet_error(port, "read end_of_file", isc_net_read_err, 0);
 		return FALSE;
 	}
 
@@ -3565,7 +3603,7 @@ static int packet_receive(
 		if (INET_force_error == 0) {
 			INET_force_error = 1;
 			inet_error(port, "simulated error - read",
-					   isc_net_read_err, isc_arg_end);
+					   isc_net_read_err, 0);
 			return FALSE;
 		}
 	}
@@ -3577,6 +3615,7 @@ static int packet_receive(
 	return TRUE;
 }
 
+
 static bool_t packet_send( rem_port* port, const SCHAR* buffer, SSHORT buffer_length)
 {
 /**************************************
@@ -3602,7 +3641,7 @@ static bool_t packet_send( rem_port* port, const SCHAR* buffer, SSHORT buffer_le
 		}
 #endif
 		SSHORT n = -1;
-		n = send((SOCKET) port->port_handle, data, length, 0);
+		n = send((SOCKET) port->port_handle, data, length, FB_SEND_FLAGS);
 #ifdef DEBUG
 		if (INET_trace & TRACE_operations) {
 			fprintf(stdout, "After Send n is %d\n", n);
@@ -3641,7 +3680,7 @@ static bool_t packet_send( rem_port* port, const SCHAR* buffer, SSHORT buffer_le
 		int inetErrNo;
 		const char* b = buffer;
 
-		while ((n = send((SOCKET) port->port_handle, b, 1, MSG_OOB)) == -1 &&
+		while ((n = send((SOCKET) port->port_handle, b, 1, MSG_OOB | FB_SEND_FLAGS)) == -1 &&
 				(INET_ERRNO == ENOBUFS || INTERRUPT_ERROR(INET_ERRNO)))
 		{
 			inetErrNo = INET_ERRNO;
@@ -3709,7 +3748,7 @@ static bool_t packet_send( rem_port* port, const SCHAR* buffer, SSHORT buffer_le
 		if (INET_force_error == 0) {
 			INET_force_error = 1;
 			inet_error(port, "simulated error - send",
-					   isc_net_write_err, isc_arg_end);
+					   isc_net_write_err, 0);
 			return FALSE;
 		}
 	}
