@@ -19,8 +19,6 @@
  *
  * All Rights Reserved.
  * Contributor(s): ______________________________________.
- * Added TCP_NO_DELAY option for superserver on Linux
- * FSG 16.03.2001
  *
  * 2002.02.15 Sean Leyne - This module needs to be cleanedup to remove obsolete ports/defines:
  *                            - "EPSON", "XENIX" +++
@@ -81,6 +79,20 @@ const USHORT PROTOCOL_VERSION9	= 9;
 
 const USHORT PROTOCOL_VERSION10	= 10;
 
+// Since protocol 11 we must be separated from Borland Interbase.
+// Therefore always set highmost bit in protocol version to 1.
+// For unsigned protocol version this does not break version's compare.
+
+const USHORT FB_PROTOCOL_FLAG = 0x8000;
+const USHORT FB_PROTOCOL_MASK = ~FB_PROTOCOL_FLAG;
+
+// Protocol 11 has support for user authentication related 
+// operations (op_update_account_info, op_authenticate_user and 
+// op_trusted_auth). When specific operation is not supported,
+// we say "sorry".
+
+const USHORT PROTOCOL_VERSION11	= (FB_PROTOCOL_FLAG | 11);
+
 #ifdef SCROLLABLE_CURSORS
 This Protocol includes support for scrollable cursors
 and is purposely being undefined so that changes can be made
@@ -103,7 +115,7 @@ typedef enum
 	arch_alliant	= 6,
 	arch_msdos		= 7,
 	arch_sun4		= 8,
-	arch_sunx86		= 9,
+	arch_sun386		= 9,
 	arch_hpux		= 10,
 	arch_hpmpexl	= 11,
 	arch_mac		= 12,
@@ -134,8 +146,9 @@ typedef enum
 	arch_freebsd	= 37,
 	arch_netbsd		= 38,
 	arch_darwin_ppc		= 39,
-
-	arch_max		= 40	/* Keep this at the end */
+	arch_winnt_64	= 40,
+	arch_darwin_64		= 41,
+	arch_max		= 42	/* Keep this at the end */
 } P_ARCH;
 
 /* Protocol Types */
@@ -144,6 +157,7 @@ const USHORT ptype_page			= 1;	/* Page server protocol */
 const USHORT ptype_rpc			= 2;	/* Simple remote procedure call */
 const USHORT ptype_batch_send	= 3;	/* Batch sends, no asynchrony */
 const USHORT ptype_out_of_band	= 4;	/* Batch sends w/ out of band notification */
+const USHORT ptype_lazy_send	= 5;	/* Deferred packets delivery */
 
 /* Generic object id */
 
@@ -151,6 +165,12 @@ typedef USHORT OBJCT;
 const int MAX_OBJCT_HANDLES	= 65000;
 const int INVALID_OBJECT = MAX_USHORT;
 
+/* Statement flags */
+
+const USHORT STMT_BLOB			= 1;
+const USHORT STMT_NO_BATCH		= 2;
+const USHORT STMT_DEFER_EXECUTE	= 4;
+
 /* Operation (packet) types */
 
 typedef enum
@@ -263,12 +283,15 @@ typedef enum
 	op_service_start		= 85,
 
 	op_rollback_retaining	= 86,
-// Two following opcode are used in vulcan.
-// No plans to implement protocol 11, where they are used,
-// place here only for informational reasons and to have common op-space.
-//	op_update_account_info	= 87,
-//	op_authenticate_user	= 88,
+
+	// Two following opcode are used in vulcan.
+	// No plans to implement them completely for a while, but to 
+	// support protocol 11, where they are used, have them here.
+	op_update_account_info	= 87,
+	op_authenticate_user	= 88,
+
 	op_partial				= 89,	// packet is not complete - delay processing
+	op_trusted_auth			= 90,
 
 	op_max
 } P_OP;
@@ -366,8 +389,8 @@ const UCHAR CNCT_user_verification	= 6;	/* Attach/create using this connection
 
 typedef struct bid	/* BLOB ID */
 {
-	ULONG	bid_relation_id;	/* Relation id (or null) */
-	ULONG	bid_number;			/* Record number */
+	ULONG	bid_quad_high;
+	ULONG	bid_quad_low;
 } *BID;
 
 
@@ -574,6 +597,26 @@ typedef struct p_sqlcur {
     USHORT	p_sqlcur_type;		/* type of cursor */
 } P_SQLCUR;
 
+typedef struct p_trau
+{
+	CSTRING	p_trau_data;					// Context
+} P_TRAU;
+
+struct p_update_account {
+    OBJCT			p_account_database;		// Database object id
+    CSTRING_CONST	p_account_apb;			// Account parameter block (apb)
+};
+
+struct p_authenticate {
+    OBJCT			p_auth_database;		// Database object id
+    CSTRING_CONST	p_auth_dpb;				// Database parameter block w/ user credentials
+	CSTRING			p_auth_items;			// Information
+	CSTRING			p_auth_recv_items;		// Receive information
+	USHORT			p_auth_buffer_length;	// Target buffer length
+};
+
+
+
 /* Generalize packet (sic!) */
 
 typedef struct packet {
@@ -608,6 +651,9 @@ typedef struct packet {
     P_SQLCUR	p_sqlcur;	/* DSQL Set cursor name */
     P_SQLFREE	p_sqlfree;	/* DSQL Free statement */
     P_TRRQ	p_trrq;		/* Transact request packet */
+	P_TRAU	p_trau;		/* Trusted authentication */
+	p_update_account p_account_update;
+	p_authenticate p_authenticate_user;
 } PACKET;
 
 #endif // REMOTE_PROTOCOL_H
