@@ -46,7 +46,8 @@ int xdrmem_create(XDR *, SCHAR *, u_int, enum xdr_op);
 const SLONG DUMMY_INTERVAL		= 60;	/* seconds */
 const int ATTACH_FAILURE_SPACE	= 2048;	/* bytes */
 
-static Firebird::StringsBuffer* attachFailures = NULL;
+static TEXT* attach_failures = NULL;
+static TEXT* attach_failures_ptr;
 
 static void cleanup_memory(void*);
 static SLONG get_parameter(const UCHAR**);
@@ -282,7 +283,7 @@ rrq* REMOTE_find_request(rrq* request, USHORT level)
 }
 
 
-void REMOTE_free_packet( rem_port* port, PACKET * packet)
+void REMOTE_free_packet( rem_port* port, PACKET * packet, bool partial)
 {
 /**************************************
  *
@@ -291,8 +292,8 @@ void REMOTE_free_packet( rem_port* port, PACKET * packet)
  **************************************
  *
  * Functional description
- *	Zero out a packet block.
- *
+ *	Zero out a full packet block (partial == false) or
+ *	part of packet used in last operation (partial == true)
  **************************************/
 	XDR xdr;
 	USHORT n;
@@ -302,12 +303,18 @@ void REMOTE_free_packet( rem_port* port, PACKET * packet)
 					  sizeof(PACKET), XDR_FREE);
 		xdr.x_public = (caddr_t) port;
 
-		for (n = (USHORT) op_connect; n < (USHORT) op_max; n++) {
-			packet->p_operation = (P_OP) n;
+		if (partial) {
 			xdr_protocol(&xdr, packet);
 		}
+		else {
+			for (n = (USHORT) op_connect; n < (USHORT) op_max; n++) {
+				packet->p_operation = (P_OP) n;
+				xdr_protocol(&xdr, packet);
+			}
+		}
 #ifdef DEBUG_XDR_MEMORY
-		/* All packet memory allocations should now be voided. */
+		// All packet memory allocations should now be voided. 
+		// note: this code will may work properly if partial == true
 
 		for (n = 0; n < P_MALLOC_SIZE; n++)
 			fb_assert(packet->p_malloc[n].p_operation == op_void);
@@ -642,22 +649,55 @@ void REMOTE_save_status_strings( ISC_STATUS* vector)
  *	strings to a special buffer.
  *
  **************************************/
-	if (!attachFailures)
+	if (!attach_failures)
 	{
-		try 
-		{
-			attachFailures = FB_NEW(*getDefaultMemoryPool()) Firebird::CircularStringsBuffer<ATTACH_FAILURE_SPACE>;
-			/* FREE: freed by exit handler cleanup_memory() */
-		}
-		catch (const std::bad_alloc&)	/* NOMEM: don't bother trying to copy */
-		{
+		attach_failures =
+			(TEXT*) ALLOC_LIB_MEMORY((SLONG) ATTACH_FAILURE_SPACE);
+		/* FREE: freed by exit handler cleanup_memory() */
+		if (!attach_failures)	/* NOMEM: don't bother trying to copy */
 			return;
-		}
-
+#ifdef DEBUG_GDS_ALLOC
+		/* This buffer is freed by the exit handler - but some of the
+		 * reporting mechanisms will report it anyway, so flag it as
+		 * "known unfreed" to get the reports quiet.
+		 */
+		gds_alloc_flag_unfreed((void*) attach_failures);
+#endif	/* DEBUG_GDS_ALLOC */
+		attach_failures_ptr = attach_failures;
 		gds__register_cleanup(cleanup_memory, 0);
 	}
 
-	attachFailures->makePermanentVector(vector, vector);
+	TEXT* p;
+	USHORT l = 0; // silence non initialized warning
+	while (*vector)
+	{
+		const ISC_STATUS status = *vector++;
+		switch (status)
+		{
+		case isc_arg_cstring:
+			l = (USHORT) *vector++;
+
+		case isc_arg_interpreted:
+		case isc_arg_string:
+			p = (TEXT*) *vector;
+			if (status != isc_arg_cstring)
+				l = strlen(p) + 1;
+
+			/* If there isn't any more room in the buffer,
+			   start at the beginning again */
+
+			if (attach_failures_ptr + l > attach_failures + ATTACH_FAILURE_SPACE)
+				attach_failures_ptr = attach_failures;
+			*vector++ = (ISC_STATUS) attach_failures_ptr;
+			memcpy(attach_failures_ptr, p, l);
+			attach_failures_ptr += l;
+			break;
+
+		default:
+			++vector;
+			break;
+		}
+	}
 }
 
 
@@ -714,7 +754,7 @@ OBJCT REMOTE_set_object(rem_port* port, BLK object, OBJCT slot)
 }
 
 
-static void cleanup_memory(void *)
+static void cleanup_memory( void *block)
 {
 /**************************************
  *
@@ -727,10 +767,11 @@ static void cleanup_memory(void *)
  *
  **************************************/
 
-	delete attachFailures;
-	attachFailures = NULL;
+	if (attach_failures)
+		FREE_LIB_MEMORY(attach_failures);
 
 	gds__unregister_cleanup(cleanup_memory, 0);
+	attach_failures = NULL;
 }
 
 
@@ -774,9 +815,9 @@ rem_port* rem_port::receive(PACKET* pckt)
 	return (*this->port_receive_packet)(this, pckt);
 }
 
-bool rem_port::select_multi(UCHAR* buffer, SSHORT bufsize, SSHORT* length, rem_port*& port)
+rem_port* rem_port::select_multi(UCHAR* buffer, SSHORT bufsize, SSHORT* length)
 {
-	return (*this->port_select_multi)(this, buffer, bufsize, length, port);
+	return (*this->port_select_multi)(this, buffer, bufsize, length);
 }
 
 XDR_INT rem_port::send(PACKET* pckt)
@@ -847,3 +888,24 @@ bool_t REMOTE_getbytes (XDR * xdrs, SCHAR * buff, u_int count)
 	return TRUE;
 }
 #endif //SUPERSERVER
+
+#ifdef TRUSTED_AUTH
+ServerAuth::ServerAuth(const char* fName, int fLen, const Firebird::ClumpletWriter& pb, 
+					   ServerAuth::Part2* p2, P_OP op)
+: fileName(*getDefaultMemoryPool()), clumplet(*getDefaultMemoryPool()), 
+  part2(p2), operation(op)
+{
+	fileName.assign(fName, fLen);
+	size_t pbLen = pb.getBufferLength();
+	if (pbLen)
+	{
+		memcpy(clumplet.getBuffer(pbLen), pb.getBuffer(), pbLen);
+	}
+	authSspi = FB_NEW(*getDefaultMemoryPool()) AuthSspi;
+}
+
+ServerAuth::~ServerAuth()
+{
+	delete authSspi;
+}
+#endif // TRUSTED_AUTH
