@@ -33,6 +33,10 @@
 #include "../remote/remote_def.h"
 #include "../jrd/thd.h"
 #include "../common/classes/objects_array.h"
+#include "../auth/trusted/AuthSspi.h"
+#include "../common/classes/fb_string.h"
+#include "../common/classes/ClumpletWriter.h"
+#include "../common/StatusHolder.h"
 
 /* Include some apollo include files for tasking */
 
@@ -281,7 +285,7 @@ typedef struct rsr
 	rem_fmt*		rsr_format;				/* Format of current message */
 	message*		rsr_message;			/* Next message to process */
 	message*		rsr_buffer;				/* Next buffer to use */
-	ISC_STATUS_ARRAY	rsr_status_vector;	/* saved status for buffered errors */
+	Firebird::StatusHolder* rsr_status;		/* saved status for buffered errors */
 	USHORT			rsr_id;
 	USHORT			rsr_flags;
 	USHORT			rsr_fmt_length;
@@ -298,7 +302,47 @@ const USHORT RSR_eof		= 2;		/* End-of-stream encountered */
 const USHORT RSR_blob		= 4;		/* Statement relates to blob op */
 const USHORT RSR_no_batch	= 8;		/* Do not batch fetch rows */
 const USHORT RSR_stream_err	= 16;		/* There is an error pending in the batched rows */
+const USHORT RSR_lazy		= 32;		/* To be allocated at the first reference */
+const USHORT RSR_defer_execute	= 64;	// op_execute can be deferred
+const USHORT RSR_past_eof	= 128;		// EOF was returned by fetch from this statement
 
+// will be methods of remote statement class
+inline void stmt_save_exception(RSR statement, const ISC_STATUS* status, bool overwrite)
+{
+	if (!statement->rsr_status) {
+		statement->rsr_status = new Firebird::StatusHolder();
+	}
+	if (overwrite || !statement->rsr_status->getError()) {
+		statement->rsr_status->save(status);
+	}
+}
+
+inline void stmt_clear_exception(RSR statement)
+{
+	if (statement->rsr_status)
+		statement->rsr_status->clear();
+}
+
+inline ISC_STATUS stmt_have_exception(RSR statement)
+{
+	return (statement->rsr_status ?	
+		statement->rsr_status->getError() : 0);
+}
+
+inline void stmt_raise_exception(RSR statement)
+{
+	if (statement->rsr_status)
+		statement->rsr_status->raise();
+}
+
+inline void stmt_release_exception(RSR statement)
+{
+	if (statement->rsr_status) 
+	{
+		delete statement->rsr_status;
+		statement->rsr_status = NULL;
+	}
+}
 
 enum blk_t
 {
@@ -353,6 +397,35 @@ typedef int HANDLE;
 struct p_cnct;
 struct rmtque;
 
+/* Queue of deferred packets */
+
+struct rem_que_packet
+{
+	PACKET packet;
+	bool sent;
+};
+
+typedef Firebird::Array<rem_que_packet> PacketQueue;
+
+#ifdef TRUSTED_AUTH
+// delayed authentication block for trusted auth callback
+class ServerAuth
+{
+public:
+	typedef void Part2(rem_port*, P_OP, const char* fName, int fLen, const UCHAR* pb, int pbLen, PACKET*);
+	Firebird::PathName fileName;
+	Firebird::HalfStaticArray<UCHAR, 128> clumplet;
+	AuthSspi* authSspi;
+	Part2* part2;
+	P_OP operation;
+
+	ServerAuth(const char* fName, int fLen, const Firebird::ClumpletWriter& pb, Part2* p2, P_OP op);
+	~ServerAuth();
+};
+#endif // TRUSTED_AUTH
+
+/* Port itself */
+
 class port_interface
 {
 public:
@@ -396,7 +469,7 @@ struct rem_port
 	XDR_INT			(*port_send_partial)(rem_port*, PACKET*);
 	t_port_connect	port_connect;		/* Establish secondary connection */
 	rem_port*		(*port_request)(rem_port*, PACKET*);	/* Request to establish secondary connection */
-	bool			(*port_select_multi)(rem_port*, UCHAR*, SSHORT, SSHORT*, rem_port*&);	// get packet from active port
+	rem_port*		(*port_select_multi)(rem_port*, UCHAR*, SSHORT, SSHORT*);	// get packet from active port
 									
 	rdb*			port_context;
 	t_event_ast		port_ast;		/* AST for events */
@@ -421,10 +494,15 @@ struct rem_port
 #ifdef VMS
 	USHORT			port_iosb[4];
 #endif
-	void*			port_xcc;              /* interprocess structure */
+	void*			port_xcc;				/* interprocess structure */
+	PacketQueue*	port_deferred_packets;	/* queue of deferred packets */
+	OBJCT			port_last_object_id;	/* cached last id */
 #ifdef SUPERSERVER
 	Firebird::ObjectsArray< Firebird::Array< char > >*	port_queue;
 	size_t			port_qoffset;			// current packet in the queue
+#endif
+#ifdef TRUSTED_AUTH
+	ServerAuth*		port_trusted_auth;
 #endif
 	UCHAR			port_buffer[1];
 
@@ -437,7 +515,65 @@ struct rem_port
 	XDR_INT	send_partial(PACKET* pckt);
 	rem_port*	connect(PACKET* pckt, t_event_ast);
 	rem_port*	request(PACKET* pckt);
-	bool   select_multi(UCHAR* buffer, SSHORT bufsize, SSHORT* length, rem_port*& port);
+	rem_port*   select_multi(UCHAR* buffer, SSHORT bufsize, SSHORT* length);
+
+	bool haveRecvData() const
+	{
+		return (port_receive.x_handy > 0 
+#ifdef SUPERSERVER
+			|| port_queue && (port_qoffset < port_queue->getCount())
+#endif
+			);
+	}
+
+	void clearRecvQue()
+	{
+#ifdef SUPERSERVER
+		if (port_queue)
+			port_queue->clear();
+		port_qoffset = 0; 
+#endif
+		port_receive.x_private = port_receive.x_base;
+	}
+
+	class RecvQueState
+	{
+	public:
+		int save_handy;
+		size_t save_private;
+#ifdef SUPERSERVER
+		size_t save_qoffset;
+#endif
+
+		RecvQueState(rem_port* port)
+		{ 
+			save_handy = port->port_receive.x_handy;
+			save_private = port->port_receive.x_private - port->port_receive.x_base;
+#ifdef SUPERSERVER
+			save_qoffset = port->port_qoffset;
+#endif
+		}
+	};
+
+	RecvQueState getRecvState()
+	{
+		return RecvQueState(this);
+	}
+
+	void setRecvState(const RecvQueState& rs)
+	{
+#ifdef SUPERSERVER
+		if (rs.save_qoffset > 0 && (rs.save_qoffset != port_qoffset))
+		{
+			Firebird::Array<char>& q = (*port_queue)[rs.save_qoffset - 1];
+			memcpy(port_receive.x_base, q.begin(), q.getCount());
+		}
+		port_qoffset = rs.save_qoffset;
+#endif
+		port_receive.x_private = port_receive.x_base + rs.save_private;
+		port_receive.x_handy = rs.save_handy;
+	}
+
 
 	/* TMN: The following member functions are conceptually private
 	 *      to server.cpp and should be _made_ private in due time!
@@ -472,8 +608,8 @@ struct rem_port
 	ISC_STATUS	receive_msg(P_DATA*, PACKET*);
 	ISC_STATUS	seek_blob(P_SEEK*, PACKET*);
 	ISC_STATUS	send_msg(P_DATA*, PACKET*);
-	ISC_STATUS	send_response(PACKET*, OBJCT, USHORT, const ISC_STATUS*);
-	ISC_STATUS	service_attach(P_ATCH*, PACKET*);
+	ISC_STATUS	send_response(PACKET*, OBJCT, USHORT, const ISC_STATUS*, bool);
+	ISC_STATUS	service_attach(const char*, const USHORT, Firebird::ClumpletWriter&, PACKET*);
 	ISC_STATUS	service_end(P_RLSE*, PACKET*);
 	ISC_STATUS	service_start(P_INFO*, PACKET*);
 	ISC_STATUS	set_cursor(P_SQLCUR*, PACKET*);
@@ -481,7 +617,6 @@ struct rem_port
 	ISC_STATUS	start_and_send(P_OP, P_DATA*, PACKET*);
 	ISC_STATUS	start_transaction(P_OP, P_STTR*, PACKET*);
 	ISC_STATUS	transact_request(P_TRRQ *, PACKET*);
-
 };
 
 // port_flags
@@ -498,7 +633,9 @@ const USHORT PORT_not_trusted	= 256;	/* Connection is from an untrusted node */
 // This is tested only in wnet.cpp but never set.
 //const USHORT PORT_impersonate	= 512;	// A remote user is being impersonated
 const USHORT PORT_dummy_pckt_set= 1024;	/* A dummy packet interval is set  */
-const USHORT PORT_partial_data  = 2048;	/* Physical packet doesn't contain all API packet */
+const USHORT PORT_partial_data	= 2048;	/* Physical packet doesn't contain all API packet */
+const USHORT PORT_lazy			= 4096;	/* Deferred operations are allowed */
+const USHORT PORT_busy			= 8192;	// disable receive - port is busy now
 
 
 /* Misc declarations */
