@@ -46,6 +46,10 @@
 
 #include "../utilities/gsec/call_service.h"
 #include "../common/utils_proto.h"
+#include "../common/classes/MsgPrint.h"
+
+using MsgFormat::SafeArg;
+
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -69,24 +73,28 @@ const int MAXSTUFF	= 1000;		/* longest interactive command line */
 class tsec *gdsec;
 #endif
 
-static int common_main(int, char**, Jrd::pfn_svc_output, Jrd::Service*);
+static int common_main(int, fb_utils::arg_string*, Jrd::pfn_svc_output, Jrd::Service*);
 static void util_output(const SCHAR*, ...);
+static int util_print(const SCHAR*, ...);
+static int vutil_print(const SCHAR*, va_list);
 
 static void data_print(void*, const internal_user_data*, bool);
 static bool get_line(int*, SCHAR**, TEXT*, size_t);
-static bool get_switches(int, const TEXT* const*, const in_sw_tab_t*, tsec*, bool*);
-static SSHORT parse_cmd_line(int, const TEXT* const*, tsec*);
+static bool get_switches(int, fb_utils::arg_string*, const in_sw_tab_t*, tsec*, bool*);
+static SSHORT parse_cmd_line(int, fb_utils::arg_string*, tsec*);
 static void printhelp(void);
 #ifndef SERVICE_THREAD
 static int output_main(Jrd::Service*, const UCHAR*);
 #endif
-inline void msg_get(USHORT number, TEXT* msg);
+inline void msg_get(USHORT, TEXT*);
+static void get_security_error(ISC_STATUS*, int);
+static void insert_error(ISC_STATUS*, ISC_STATUS);
 
 void inline gsec_exit(int code, tsec* tdsec)
 {
 	tdsec->tsec_exit_code = code;
 	if (tdsec->tsec_throw)
-		Firebird::status_exception::raise();
+		Firebird::LongJump::raise();
 }
 
 #ifdef SERVICE_THREAD
@@ -101,7 +109,7 @@ THREAD_ENTRY_DECLARE GSEC_main(THREAD_ENTRY_PARAM arg)
  *   Entrypoint for GSEC via the services manager
  **********************************************/
 	Jrd::Service* service = (Jrd::Service*)arg;
-	const int exit_code = common_main(service->svc_argc, service->svc_argv,
+	const int exit_code = common_main(service->svc_argc, service->svc_argv.begin(),
 						  SVC_output, service);
 
 /* Mark service thread as finished. */
@@ -160,7 +168,7 @@ inline void envPick(TEXT* dest, size_t size, const TEXT* var)
 #endif /* SERVICE_THREAD */
 
 int common_main(int argc,
-				char* argv[],
+				fb_utils::arg_string argv[],
 				Jrd::pfn_svc_output output_proc,
 				Jrd::Service* output_data)
 {
@@ -176,7 +184,9 @@ int common_main(int argc,
  *	the specified argc/argv to SECURITY_exec_line (see below).
  *
  **************************************/
+#ifndef SERVICE_THREAD
 	TEXT stuff[MAXSTUFF];		/* a place to put stuff */
+#endif
 
 #ifdef VMS
 	argc = VMS_parse(&argv, argc);
@@ -190,9 +200,9 @@ int common_main(int argc,
 	tdsec->tsec_user_data = &u;
 
 	try {
-/* Perform some special handling when run as an Interbase service.  The
-   first switch can be "-svc" (lower case!) or it can be "-svc_re" followed
-   by 3 file descriptors to use in re-directing stdin, stdout, and stderr. */
+	// Perform some special handling when run as a Firebird service.  The
+	// first switch can be "-svc" (lower case!) or it can be "-svc_re" followed
+	// by 3 file descriptors to use in re-directing stdin, stdout and stderr.
 
 	tdsec->tsec_throw = true;
 	tdsec->tsec_interactive = true;
@@ -236,8 +246,8 @@ int common_main(int argc,
 		argc -= 4;
 	}
 
+	ISC_STATUS* status = tdsec->tsec_status;
 	SSHORT ret = parse_cmd_line(argc, argv, tdsec);
-
 	Firebird::PathName databaseName;
 	bool databaseNameEntered = user_data->database_name_entered;
 	if (user_data->database_name_entered)
@@ -271,7 +281,6 @@ int common_main(int argc,
 	}
 	databaseName.copyTo(user_data->database_name, sizeof(user_data->database_name));
 	
-	ISC_STATUS* status = tdsec->tsec_status;
 #ifdef SUPERCLIENT
 	useServices = true;
 #else //SUPERCLIENT
@@ -282,15 +291,33 @@ int common_main(int argc,
 		Firebird::ClumpletWriter dpb(Firebird::ClumpletReader::Tagged, MAX_DPB_SIZE, isc_dpb_version1);
 		dpb.insertByte(isc_dpb_gsec_attach, 1); // not 0 - yes, I'm gsec
 
-		if (user_data->dba_user_name_entered) {
-			dpb.insertString(isc_dpb_user_name, 
-				user_data->dba_user_name, strlen(user_data->dba_user_name));
+#ifdef TRUSTED_SERVICES
+		if (user_data->dba_trust_user_name_entered) 
+		{
+			dpb.insertString(isc_dpb_trusted_auth, 
+				user_data->dba_trust_user_name, strlen(user_data->dba_trust_user_name));
 		}
+		else
+#endif
+		{
+#ifdef TRUSTED_AUTH
+			if (user_data->trusted_auth) 
+			{
+				dpb.insertTag(isc_dpb_trusted_auth);
+			}
+#endif
+			if (user_data->dba_user_name_entered) 
+			{
+				dpb.insertString(isc_dpb_user_name, 
+					user_data->dba_user_name, strlen(user_data->dba_user_name));
+			}
 
-		if (user_data->dba_password_entered) {
-			dpb.insertString(tdsec->tsec_service_gsec ? 
-							isc_dpb_password_enc : isc_dpb_password, 
-				user_data->dba_password, strlen(user_data->dba_password));
+			if (user_data->dba_password_entered) 
+			{
+				dpb.insertString(tdsec->tsec_service_gsec ? 
+								isc_dpb_password_enc : isc_dpb_password, 
+					user_data->dba_password, strlen(user_data->dba_password));
+			}
 		}
 
 		if (user_data->sql_role_name_entered) {
@@ -302,7 +329,7 @@ int common_main(int argc,
 				dpb.getBufferLength(), 
 				reinterpret_cast<const char*>(dpb.getBuffer())))
 		{
-			GSEC_error_redirect(status, GsecMsg15, NULL, NULL);
+			GSEC_error_redirect(status, GsecMsg15);
 		}
 	}
 #endif //SUPERCLIENT
@@ -311,17 +338,27 @@ int common_main(int argc,
 	if (useServices)
 	{
 #ifndef SERVICE_THREAD
-		envPick(user_data->dba_user_name, sizeof user_data->dba_user_name, "ISC_USER");
-		envPick(user_data->dba_password, sizeof user_data->dba_password, "ISC_PASSWORD");
+#ifdef TRUSTED_AUTH
+		if (!(user_data->trusted_auth))
+#endif
+		{
+			envPick(user_data->dba_user_name, sizeof user_data->dba_user_name, "ISC_USER");
+			envPick(user_data->dba_password, sizeof user_data->dba_password, "ISC_PASSWORD");
+		}
 #endif //SERVICE_THREAD
 		sHandle = attachRemoteServiceManager(
 					status,
 					user_data->dba_user_name,
 					user_data->dba_password,
+#ifdef TRUSTED_AUTH
+					user_data->trusted_auth,
+#else
+					false,
+#endif
 					serverName.c_str());
 		if (! sHandle)
 		{
-			GSEC_error_redirect(status, GsecMsg15, NULL, NULL);
+			GSEC_error_redirect(status, GsecMsg15);
 		}
 	}
 
@@ -340,12 +377,14 @@ int common_main(int argc,
 			{
 				ret = SECURITY_exec_line(status, db_handle, 
 							user_data, data_print, NULL);
-				if (ret) {
-					GSEC_print(ret, user_data->user_name, NULL, NULL, NULL, NULL);
+				if (ret) 
+				{
+					GSEC_print(ret, user_data->user_name);
 					if (status[1])
 					{
 						GSEC_print_status(status);
 					}
+					get_security_error(status, ret);
 				}
 			}
 			else
@@ -360,26 +399,51 @@ int common_main(int argc,
 			}
 		}
 	}
+#ifndef SERVICE_THREAD
 	else {
 		int local_argc;
-		SCHAR* local_argv[MAXARGS];
+		char* local_argv[MAXARGS];
 		for (;;) {
+			MOVE_CLEAR(status, sizeof(ISC_STATUS_ARRAY));
 			/* Clear out user data each time through this loop. */
-			MOVE_CLEAR(tdsec->tsec_user_data, sizeof(internal_user_data));
+			MOVE_CLEAR(user_data, sizeof(internal_user_data));
 			if (get_line(&local_argc, local_argv, stuff, sizeof(stuff)))
 				break;
 			if (local_argc > 1) {
 				ret = parse_cmd_line(local_argc, local_argv, tdsec);
+				if (ret == 1)
+				{
+					// quit command
+					break;
+				}
+				if (user_data->dba_user_name_entered || 
+					user_data->dba_password_entered ||
+					user_data->database_name_entered
+#ifdef TRUSTED_AUTH
+					|| user_data->trusted_auth
+#endif
+					)
+				{
+					GSEC_print(GsecMsg92);
+					continue;
+				}
+
 				databaseName.copyTo(user_data->database_name, sizeof(user_data->database_name));
 				user_data->database_name_entered = databaseNameEntered;
-				if (ret == 1)
-					break;
 				if (ret == 0) {
 #ifndef SUPERCLIENT
 					if (!useServices)
 					{
 						ret = SECURITY_exec_line(status, db_handle, 
 							user_data, data_print, NULL);
+						if (ret) 
+						{
+							GSEC_print(ret, user_data->user_name);
+							if (status[1]) {
+								GSEC_print_status(status);
+							}
+							continue;
+						}
 					}
 					else
 #endif //SUPERCLIENT
@@ -388,28 +452,20 @@ int common_main(int argc,
 						if (status[1])
 						{
 							GSEC_print_status(status);
-							ret = GsecMsg75;
-							break;
+							continue;
 						}
 					}
-					if (ret) {
-						GSEC_print(ret, user_data->user_name, NULL, NULL,
-								   NULL, NULL);
-						if (status[1]) {
-							GSEC_print_status(status);
-						}
-						break;
-					}
 				}
 			}
 		}
 	}
+#endif
 
 #ifndef SUPERCLIENT
 	if (db_handle) {
 		ISC_STATUS_ARRAY loc_status;
 		if (isc_detach_database(loc_status, &db_handle)) {
-			GSEC_error_redirect(loc_status, 0, NULL, NULL);
+			GSEC_error_redirect(loc_status, GsecMsg93);
 		}
 	}
 #endif //SUPERCLIENT
@@ -429,30 +485,27 @@ int common_main(int argc,
 	return ret;					// silence compiler warning
 
 	}	// try
-	catch (const std::exception& e) {
-		/* 
-		 * All exceptions and calls to gsec_exit(), 
-		 * normal and error exits, wind up here 
-		 */
-		int exit_code = tdsec->tsec_exit_code;
-
-		ISC_STATUS_ARRAY status;
-		memset(status, 0, sizeof status);
-		Firebird::stuff_exception(status, e);
+	catch (const Firebird::LongJump&) {
+		/* All calls to gsec_exit(), normal and error exits, wind up here */
+		const int exit_code = tdsec->tsec_exit_code;
 
 		tdsec->tsec_service_blk->svc_started();
 		tdsec->tsec_throw = false;
-		
-		if (status[0]) {
-			// We have real exception, gsec_exit() was not called
-			GSEC_print_status(status);
-			memcpy(tdsec->tsec_status, status, sizeof status);
-			exit_code = 127;
-		}
 
-		/* All returns occur from this point - even normal returns */
 		return exit_code;
 	}
+	catch (const Firebird::Exception& e) {
+		// Real exceptions are coming here
+		ISC_STATUS *status = tdsec->tsec_status;
+		e.stuff_exception(status);
+
+		tdsec->tsec_service_blk->svc_started();
+		tdsec->tsec_throw = false;
+
+		GSEC_print_status(status, false);
+
+		return 127;
+	}
 }
 
 
@@ -526,8 +579,8 @@ static void data_print(void* arg, const internal_user_data* data, bool first)
 	}
 	else {
 		if (first) {
-			GSEC_print(GsecMsg26, NULL, NULL, NULL, NULL, NULL);
-			GSEC_print(GsecMsg27, NULL, NULL, NULL, NULL, NULL);
+			GSEC_print(GsecMsg26);
+			GSEC_print(GsecMsg27);
 /* msg26: "    user name                    uid   gid     full name" */
 /* msg27: "-------------------------------------------------------------------------------------------" */
 		}
@@ -553,7 +606,7 @@ static bool get_line(int* argc, SCHAR** argv, TEXT* stuff, size_t maxstuff)
  *	unused), and a max of MAXSTUFF characters, at which point
  *
  **************************************/
-	GSEC_print_partial(GsecMsg1, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print_partial(GsecMsg1);
 	*argc = 1;
 	TEXT* cursor = stuff;
 	int count = (int) maxstuff - 1;
@@ -604,7 +657,7 @@ static bool get_line(int* argc, SCHAR** argv, TEXT* stuff, size_t maxstuff)
 
 static bool get_switches(
 							int argc,
-							const TEXT* const* argv,
+							fb_utils::arg_string* argv,
 							const in_sw_tab_t* in_sw_table,
 							tsec* tdsec, bool* quitflag)
 {
@@ -634,35 +687,25 @@ static bool get_switches(
 	USHORT last_sw = IN_SW_GSEC_0;
 	tdsec->tsec_sw_version = false;
 	for (--argc; argc > 0; argc--) {
-		const TEXT* string = *++argv;
+		fb_utils::arg_string string = *++argv;
 		if (*string == '?')
 			user_data->operation = HELP_OPER;
 		else if (*string != '-') {
 			/* this is not a switch, so it must be a parameter for
 			   the previous switch, if any */
-			char quote;
 
 			switch (last_sw) {
 			case IN_SW_GSEC_ADD:
 			case IN_SW_GSEC_DEL:
 			case IN_SW_GSEC_DIS:
 			case IN_SW_GSEC_MOD:
-				quote = ' ';
-				for (l = 0; l < 32 && string[l] && string[l] != quote; )
-				{
-					if (l == 0 && (*string == '\'' || *string == '"'))
-					{
-						quote = *string++;
-						continue;
-					}
+				for (l = 0; l < 32 && string[l] && string[l] != ' '; l++)
 					user_data->user_name[l] = UPPER(string[l]);
-					++l;
-				}
 				if (l == 32) {
 #ifdef SERVICE_THREAD
-					GSEC_error(GsecMsg76, NULL, NULL, NULL, NULL, NULL);
+					GSEC_error(GsecMsg76);
 #else
-					GSEC_print(GsecMsg76, NULL, NULL, NULL, NULL, NULL);
+					GSEC_print(GsecMsg76);
 #endif
 					/* invalid user name (maximum 31 bytes allowed) */
 					return false;
@@ -674,7 +717,7 @@ static bool get_switches(
 				for (l = 0; l < 9 && string[l] && string[l] != ' '; l++)
 					user_data->password[l] = string[l];
 				if ((l == 9) && !(tdsec->tsec_service_gsec)) {
-					GSEC_print(GsecMsg77, NULL, NULL, NULL, NULL, NULL);
+					GSEC_print(GsecMsg77);
 					// warning password maximum 8 significant bytes used
 				}
 				user_data->password[l] = '\0';
@@ -717,19 +760,25 @@ static bool get_switches(
 				user_data->dba_user_name_entered = true;
 				break;
 			case IN_SW_GSEC_DBA_PASSWORD:
-				strncpy(user_data->dba_password, string, sizeof(user_data->dba_password));
+				strncpy(user_data->dba_password, fb_utils::get_passwd(string), sizeof(user_data->dba_password));
 				user_data->dba_password_entered = true;
 				break;
 			case IN_SW_GSEC_SQL_ROLE_NAME:
 				strncpy(user_data->sql_role_name, string, sizeof(user_data->sql_role_name));
 				user_data->sql_role_name_entered = true;
 				break;
+#ifdef TRUSTED_SERVICES
+			case IN_SW_GSEC_DBA_TRUST_USER:
+				strncpy(user_data->dba_trust_user_name, string, sizeof(user_data->dba_trust_user_name));
+				user_data->dba_trust_user_name_entered = true;
+				break;
+#endif
 			case IN_SW_GSEC_Z:
 			case IN_SW_GSEC_0:
 #ifdef SERVICE_THREAD
-				GSEC_error(GsecMsg29, NULL, NULL, NULL, NULL, NULL);
+				GSEC_error(GsecMsg29);
 #else
-				GSEC_print(GsecMsg29, NULL, NULL, NULL, NULL, NULL);
+				GSEC_print(GsecMsg29);
 #endif
 				/* gsec - invalid parameter, no switch defined */
 				return false;
@@ -788,7 +837,7 @@ static bool get_switches(
 			case IN_SW_GSEC_QUIT:
 			case IN_SW_GSEC_HELP:
 				if (user_data->operation) {
-					GSEC_error(GsecMsg30, NULL, NULL, NULL, NULL, NULL);
+					GSEC_error(GsecMsg30);
 					/* gsec - operation already specified */
 					return false;
 				}
@@ -825,6 +874,9 @@ static bool get_switches(
 			case IN_SW_GSEC_MNAME:
 			case IN_SW_GSEC_LNAME:
 			case IN_SW_GSEC_DATABASE:
+#ifdef TRUSTED_SERVICES
+			case IN_SW_GSEC_DBA_TRUST_USER:
+#endif
 			case IN_SW_GSEC_DBA_USER_NAME:
 			case IN_SW_GSEC_DBA_PASSWORD:
 			case IN_SW_GSEC_SQL_ROLE_NAME:
@@ -910,6 +962,16 @@ static bool get_switches(
 					user_data->dba_user_name_specified = true;
 					user_data->dba_user_name[0] = '\0';
 					break;
+#ifdef TRUSTED_SERVICES
+				case IN_SW_GSEC_DBA_TRUST_USER:
+					if (user_data->dba_trust_user_name_specified) {
+						err_msg_no = GsecMsg79;
+						break;
+					}
+					user_data->dba_trust_user_name_specified = true;
+					user_data->dba_trust_user_name[0] = '\0';
+					break;
+#endif
 				case IN_SW_GSEC_DBA_PASSWORD:
 					if (user_data->dba_password_specified) {
 						err_msg_no = GsecMsg80;
@@ -928,7 +990,7 @@ static bool get_switches(
 					break;
 				}
 				if (err_msg_no) {
-					GSEC_error(err_msg_no, NULL, NULL, NULL, NULL, NULL);
+					GSEC_error(err_msg_no);
 					return false;
 				}
 				break;
@@ -939,19 +1001,24 @@ static bool get_switches(
 				}
 				tdsec->tsec_sw_version = true;
 				break;
+#ifdef TRUSTED_AUTH
+			case IN_SW_GSEC_TRUSTED_AUTH:
+				user_data->trusted_auth = true;
+				break;
+#endif
 			case IN_SW_GSEC_0:
 #ifdef SERVICE_THREAD
-				GSEC_error(GsecMsg40, NULL, NULL, NULL, NULL, NULL);
+				GSEC_error(GsecMsg40);
 #else
-				GSEC_print(GsecMsg40, NULL, NULL, NULL, NULL, NULL);
+				GSEC_print(GsecMsg40);
 #endif
 				/* gsec - invalid switch specified */
 				return false;
 			case IN_SW_GSEC_AMBIG:
 #ifdef SERVICE_THREAD
-				GSEC_error(GsecMsg41, NULL, NULL, NULL, NULL, NULL);
+				GSEC_error(GsecMsg41);
 #else
-				GSEC_print(GsecMsg41, NULL, NULL, NULL, NULL, NULL);
+				GSEC_print(GsecMsg41);
 #endif
 				/* gsec - ambiguous switch specified */
 				return false;
@@ -969,7 +1036,7 @@ static bool get_switches(
 		{
 			switch (user_data->operation) {
 			case 0:
-				GSEC_error(GsecMsg42, NULL, NULL, NULL, NULL, NULL);
+				GSEC_error(GsecMsg42);
 				/* gsec - no operation specified for parameters */
 				return false;
 			case ADD_OPER:
@@ -980,7 +1047,7 @@ static bool get_switches(
 			case DIS_OPER:
 			case QUIT_OPER:
 			case HELP_OPER:
-				GSEC_error(GsecMsg43, NULL, NULL, NULL, NULL, NULL);
+				GSEC_error(GsecMsg43);
 				/* gsec - no parameters allowed for this operation */
 				return false;
 			}
@@ -1007,160 +1074,164 @@ static void printhelp(void)
  **************************************/
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg45, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg45);
 /* gsec utility - maintains user password database */
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg46, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg46);
 	util_output("%s", "     ");
-	GSEC_print_partial(GsecMsg2, NULL, NULL, NULL, NULL, NULL);
-	GSEC_print_partial(GsecMsg82, NULL, NULL, NULL, NULL, NULL);
-	GSEC_print(GsecMsg47, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print_partial(GsecMsg2);
+	GSEC_print_partial(GsecMsg82);
+	GSEC_print(GsecMsg47);
 /* gsec [ <options> ... ] -<command> [ <parameter> ... ] */
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg48, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg48);
 /* interactive usage: */
 
 	util_output("%s", "     ");
-	GSEC_print_partial(GsecMsg2, NULL, NULL, NULL, NULL, NULL);
-	GSEC_print(GsecMsg82, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print_partial(GsecMsg2);
+	GSEC_print(GsecMsg82);
 /* gsec [ <options> ... ] */
 
 	util_output("%s", "     ");
-	GSEC_print_partial(GsecMsg1, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print_partial(GsecMsg1);
 	util_output("\n%s", "     ");
-	GSEC_print(GsecMsg47, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg47);
 /* GSEC> <command> [ <parameter> ... ] */
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg83, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg83);
 /* available options: */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg84, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg84);
 /* -user <database administrator name> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg85, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg85);
 /* -password <database administrator password> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg86, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg86);
 /* -role <database administrator SQL role name> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg87, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg91);
+/* -trusted (use trusted authentication) */
+
+	util_output("%s", "     ");
+	GSEC_print(GsecMsg87);
 /* -database <database to manage> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg88, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg88);
 /* -z */
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg49, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg49);
 /* available commands: */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg50, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg50);
 /* adding a new user: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg51, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg51);
 /* add <name> [ <parameter> ... ] */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg52, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg52);
 /* deleting a current user: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg53, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg53);
 /* delete <name> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg54, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg54);
 /* displaying all users: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg55, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg55);
 /* display */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg56, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg56);
 /* displaying one user: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg57, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg57);
 /* display <name> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg58, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg58);
 /* modifying a user's parameters: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg59, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg59);
 /* modify <name> <parameter> [ <parameter> ... ] */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg60, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg60);
 /* help: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg61, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg61);
 /* ? (interactive only) */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg62, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg62);
 /* help */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg89, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg89);
 /* displaying version number: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg90, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg90);
 /* z (interactive only) */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg63, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg63);
 /* quit interactive session: */
 
 	util_output("%s", "       ");
-	GSEC_print(GsecMsg64, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg64);
 /* quit (interactive only) */
 
 	util_output("\n%s", "   ");
-	GSEC_print(GsecMsg65, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg65);
 /* available parameters: */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg66, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg66);
 /* -pw <password> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg67, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg67);
 /* -uid <uid> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg68, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg68);
 /* -gid <gid> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg71, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg71);
 /* -fname <firstname> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg72, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg72);
 /* -mname <middlename> */
 
 	util_output("%s", "     ");
-	GSEC_print(GsecMsg73, NULL, NULL, NULL, NULL, NULL);
+	GSEC_print(GsecMsg73);
 /* -lname <lastname> */
 	util_output("\n", NULL);
 }
 
 
-static SSHORT parse_cmd_line(int argc, const TEXT* const* argv, tsec* tdsec)
+static SSHORT parse_cmd_line(int argc, fb_utils::arg_string* argv, tsec* tdsec)
 {
 /**************************************
  *
@@ -1184,9 +1255,9 @@ static SSHORT parse_cmd_line(int argc, const TEXT* const* argv, tsec* tdsec)
 	SSHORT ret = 0;
 	if (!get_switches(argc, argv, gsec_in_sw_table, tdsec, &quitflag)) {
 #ifdef SERVICE_THREAD
-		GSEC_error(GsecMsg16, NULL, NULL, NULL, NULL, NULL);
+		GSEC_error(GsecMsg16);
 #else
-		GSEC_print(GsecMsg16, NULL, NULL, NULL, NULL, NULL);
+		GSEC_print(GsecMsg16);
 #endif
 		/* gsec - error in switch specifications */
 		ret = -1;
@@ -1200,7 +1271,7 @@ static SSHORT parse_cmd_line(int argc, const TEXT* const* argv, tsec* tdsec)
 				 user_data->operation != QUIT_OPER &&
 				 !user_data->user_name_entered)
 		{
-			GSEC_error(GsecMsg18, NULL, NULL, NULL, NULL, NULL);
+			GSEC_error(GsecMsg18);
 			/* gsec - no user name specified */
 			ret = -1;
 		}
@@ -1214,7 +1285,7 @@ static SSHORT parse_cmd_line(int argc, const TEXT* const* argv, tsec* tdsec)
 	return ret;
 }
 
-void GSEC_print_status(const ISC_STATUS* status_vector)
+void GSEC_print_status(const ISC_STATUS* status_vector, bool exitOnError)
 {
 /**************************************
  *
@@ -1239,22 +1310,68 @@ void GSEC_print_status(const ISC_STATUS* status_vector)
 			for (int j = 0; status_vector[j] && (i < ISC_STATUS_LENGTH); j++, i++)
 				*status++ = status_vector[j];
 		}
-#endif
-
+#else
 		SCHAR s[1024];
-		const char* nl = vector[0] == isc_arg_interpreted ? "" : "\n";
-		if (fb_interpret(s, sizeof(s), &vector)) {
+		while (fb_interpret(s, sizeof(s), &vector)) 
+		{
 			TRANSLATE_CP(s);
-			util_output("%s%s", s, nl);
-			while (fb_interpret(s, sizeof(s), &vector)) {
-				TRANSLATE_CP(s);
-				util_output("%s%s", s, nl);
+			const char* nl = (s[0] ? s[strlen(s) - 1] != '\n' : true) ? "\n" : "";
+			int exitCode = util_print("%s%s", s, nl);
+			if (exitOnError && exitCode != 0) 
+			{
+				gsec_exit(exitCode, tsec::getSpecific());
 			}
 		}
+#endif
 	}
 }
 
-static void util_output( const SCHAR* format, ...)
+static int vutil_print(const SCHAR* format, va_list arglist)
+{
+/**************************************
+ *
+ *	v u t i l _ p r i n t
+ *
+ **************************************
+ *
+ * Functional description
+ *	Platform independent output routine.
+ *  Varargs function.
+ *
+ **************************************/
+	tsec* tdsec = tsec::getSpecific();
+
+	Firebird::string buf;
+	buf.vprintf(format, arglist);
+	
+	return tdsec->tsec_output_proc(tdsec->tsec_output_data, (UCHAR*)(buf.c_str()));
+}
+
+static void util_output(const SCHAR* format, ...)
+{
+/**************************************
+ *
+ *	u t i l _ o u t p u t
+ *
+ **************************************
+ *
+ * Functional description
+ *	Platform independent output routine.
+ *  Exit on output error
+ *
+ **************************************/
+	va_list arglist;
+	va_start(arglist, format);
+	int exit_code = vutil_print(format, arglist);
+	va_end(arglist);
+
+	if (exit_code != 0) 
+	{
+		gsec_exit(exit_code, tsec::getSpecific());
+	}
+}
+
+static int util_print(const SCHAR* format, ...)
 {
 /**************************************
  *
@@ -1266,29 +1383,16 @@ static void util_output( const SCHAR* format, ...)
  *	Platform independent output routine.
  *
  **************************************/
-	int exit_code;
+	va_list arglist;
+	va_start(arglist, format);
+	int exit_code = vutil_print(format, arglist);
+	va_end(arglist);
 
-	tsec* tdsec = tsec::getSpecific();
-
-	if (format[0] == '\0') {
-		exit_code = tdsec->tsec_output_proc(tdsec->tsec_output_data,
-									(UCHAR * )(""));
-	}
-	else {
-		UCHAR buf[1000];
-		va_list arglist;
-		va_start(arglist, format);
-		VSNPRINTF((char *) buf, sizeof(buf), format, arglist);
-		va_end(arglist);
-		exit_code = tdsec->tsec_output_proc(tdsec->tsec_output_data, buf);
-	}
-
-	if (exit_code != 0)
-		gsec_exit(exit_code, tdsec);
+	return exit_code;
 }
 
 void GSEC_error_redirect(const ISC_STATUS* status_vector,
-						 USHORT errcode, const TEXT* arg1, const TEXT* arg2)
+						 USHORT errcode)
 {
 /**************************************
  *
@@ -1302,14 +1406,10 @@ void GSEC_error_redirect(const ISC_STATUS* status_vector,
  **************************************/
 
 	GSEC_print_status(status_vector);
-	GSEC_error(errcode, arg1, arg2, NULL, NULL, NULL);
+	GSEC_error(errcode);
 }
 
-void GSEC_error(
-				USHORT errcode,
-				const TEXT* arg1,
-				const TEXT* arg2, const TEXT* arg3,
-				const TEXT* arg4, const TEXT* arg5)
+void GSEC_error(USHORT errcode)
 {
 /**************************************
  *
@@ -1321,33 +1421,26 @@ void GSEC_error(
  *	Format and print an error message, then punt.
  *
  **************************************/
-#ifdef SERVICE_THREAD
+	static const SafeArg dummy;
+
 	tsec* tdsec = tsec::getSpecific();
+#ifdef SERVICE_THREAD
 	ISC_STATUS* status = tdsec->tsec_service_blk->svc_status;
 
-	CMD_UTIL_put_svc_status(status, GSEC_MSG_FAC, errcode,
-							isc_arg_string, arg1,
-							isc_arg_string, arg2,
-							isc_arg_string, arg3,
-							isc_arg_string, arg4, isc_arg_string, arg5);
+	CMD_UTIL_put_svc_status(status, GSEC_MSG_FAC, errcode, dummy);
 	tdsec->tsec_service_blk->svc_started();
-#else
-	tsec* tdsec = tsec::getSpecific();
 #endif
 
-	GSEC_print(errcode, arg1, arg2, arg3, arg4, arg5);
+	GSEC_print(errcode);
 	gsec_exit(FINI_ERROR, tdsec);
 }
 
-void GSEC_print(
-				USHORT number,
-				const TEXT* arg1,
-				const TEXT* arg2, const TEXT* arg3,
-				const TEXT* arg4, const TEXT* arg5)
+void GSEC_print(USHORT number,
+				const char* str)
 {
 /**************************************
  *
- *	U T I L _ p r i n t
+ *	G S E C _ p r i n t
  *
  **************************************
  *
@@ -1357,21 +1450,20 @@ void GSEC_print(
  **************************************/
 	TEXT buffer[256];
 
-	gds__msg_format(0, GSEC_MSG_FAC, number, sizeof(buffer), buffer, arg1,
-					arg2, arg3, arg4, arg5);
+	SafeArg arg;
+	if (str)
+		arg << str;
+
+	fb_msg_format(0, GSEC_MSG_FAC, number, sizeof(buffer), buffer, arg);
 	TRANSLATE_CP(buffer);
 	util_output("%s\n", buffer);
 }
 
-void GSEC_print_partial(
-						USHORT number,
-						const TEXT* arg1,
-						const TEXT* arg2, const TEXT* arg3,
-						const TEXT* arg4, const TEXT* arg5)
+void GSEC_print_partial(USHORT number)
 {
 /**************************************
  *
- *	U T I L _ p r i n t _ p a r t i a l
+ *	G S E C _ p r i n t _ p a r t i a l
  *
  **************************************
  *
@@ -1379,10 +1471,10 @@ void GSEC_print_partial(
  *	Retrieve a message from the error file, format it, and print it.
  *
  **************************************/
+	static const SafeArg dummy;
 	TEXT buffer[256];
 
-	gds__msg_format(0, GSEC_MSG_FAC, number, sizeof(buffer), buffer, arg1,
-					arg2, arg3, arg4, arg5);
+	fb_msg_format(0, GSEC_MSG_FAC, number, sizeof(buffer), buffer, dummy);
 	TRANSLATE_CP(buffer);
 	util_output("%s ", buffer);
 }
@@ -1401,6 +1493,81 @@ inline void msg_get(USHORT number, TEXT* msg)
  *
  **************************************/
 
-	gds__msg_format(NULL, GSEC_MSG_FAC, number, MSG_LENGTH, msg,
-					NULL, NULL, NULL, NULL, NULL);
+	static const SafeArg dummy;
+	fb_msg_format(NULL, GSEC_MSG_FAC, number, MSG_LENGTH, msg, dummy);
+}
+
+
+static void insert_error(ISC_STATUS* status, ISC_STATUS isc_err)
+{
+/**************************************
+ *
+ *      i n s e r t _ e r r o r
+ *
+ **************************************
+ *
+ * Functional description
+ *
+ *	Adds isc error code to status vector
+ **************************************/
+	if (status[1])
+	{
+		memmove(&status[2], &status[0], sizeof(ISC_STATUS_ARRAY) - 2 * sizeof(ISC_STATUS));
+	}
+	else
+	{
+		status[2] = isc_arg_end;
+	}
+	status[0] = isc_arg_gds;
+	status[1] = isc_err;
+}
+
+
+static void get_security_error(ISC_STATUS* status, int gsec_err)
+{
+/**************************************
+ *
+ *      g e t _ s e c u r i t y _ e r r o r
+ *
+ **************************************
+ *
+ * Functional description
+ *
+ *    Converts the gsec error code to an isc
+ *    error code and adds it to the status vector
+ **************************************/
+
+
+	switch (gsec_err) {
+	case GsecMsg19:			/* gsec - add record error */
+		insert_error(status, isc_error_adding_sec_record);
+		return;
+
+	case GsecMsg20:			/* gsec - modify record error */
+		insert_error(status, isc_error_modifying_sec_record);
+		return;
+
+	case GsecMsg21:			/* gsec - find/modify record error */
+		insert_error(status, isc_error_modifying_sec_record);
+		return;
+
+	case GsecMsg22:			/* gsec - record not found for user: */
+		insert_error(status, isc_usrname_not_found);
+		return;
+
+	case GsecMsg23:			/* gsec - delete record error */
+		insert_error(status, isc_error_deleting_sec_record);
+		return;
+
+	case GsecMsg24:			/* gsec - find/delete record error */
+		insert_error(status, isc_error_deleting_sec_record);
+		return;
+
+	case GsecMsg75:			/* gsec error */
+		insert_error(status, isc_error_updating_sec_db);
+		return;
+
+	default:
+		return;
+	}
 }
